/**
 * MacHandler.
 *
 * @author     Gifford Cheung
 * @version    $Id: MacHandler.java,v 1.12 2002/09/09 05:04:15 srhea Exp $
 */  

package ostore.mac;
import java.util.*;
import java.io.*;
import java.math.*;
import java.net.*;
import ostore.util.*;
import ostore.dispatch.*;
import ostore.tapestry.*;
import ostore.network.*;
import seda.sandStorm.api.*;
import seda.sandStorm.internal.SinkProxy;
/**
 *  MacHandler. Responsible for checking message authenticity <i>and</i> 
 *  initializing 'first contact' with another OceanStore instance. The result
 *  being <br></br>(1) a MAC relationship established between both parties, 
 *  <br></br>(2) a GUID exchange, and <br></br>
 *  (3) the source of <i>your</i> machine's GUID. <br></br>
 *  Additional details are implementation specific.<p>
 *  <b>Note:</b>There is to be only one MacHandler per instance of sandstorm.
 *
 */
public abstract class MacHandler 
implements SingleThreadedEventHandlerIF, EventHandlerIF, ProfilableIF {

    public static final boolean DEBUG = true;
    
    public abstract void handleEvent(QueueElementIF item)
        throws EventHandlerException;
    

    public void handleEvents(QueueElementIF item[])
        throws EventHandlerException {

        for (int i = 0; i < item.length; i += 1)
            handleEvent(item[i]);
    }

    /**
     * Need these to happen before processing subsequent messages, so we
     * call it from the Router directly now.
     */
    public abstract void handle_dh_message (DHMessage dhmsg);

    /**
     * Id String
     */
    public abstract String idString();

    /**
     * My GUID
     */
    public abstract SecureHash myGuid();

    /**
     * Establish a connection between the guid and this instance of sandstorm
     * <br></br> 
     * Once a connection is established, the MacHandler will broadcast
     * a {@link KeyEstablished} announcement. If a connection is pending
     * this will be ignored. If a connection to this guid is already
     * established, a {@link KeyEstablished} will be broadcasted.
     */
    public abstract void establish(SecureHash guid);

    /**
     * @return <code>true</code> iff this message is authentic, i.e.
     * verified by the internal MAC
     */
    public abstract boolean authenticate(QuickSerializable user_data, 
        QSByteArray mac, SecureHash guid) throws GuidNotEstablished;

    public abstract void init(ConfigDataIF config) throws Exception ;
    
    public abstract void destroy() throws Exception ;

    /* Specified by seda.sandStorm.api.ProfilableIF */
    public abstract int profileSize();

    /**
     * Thrown when a request for action (such as {@link #authenticate} or
     * is called on a guid that doesn't have an
     * established connection yet.
     */
    public class GuidNotEstablished extends Exception {
        public GuidNotEstablished() { }
        public GuidNotEstablished (String msg) {super(msg);}
        public String toString() {
            return "GuidNotEstablished " + super.toString() + ".";
        }
    }

    /**
     * Announcement that we have established a connection with 
     * the <tt>guid</tt> and are ready to call {@link #authenticate}
     * on that guid.
     */
    public class KeyEstablished implements QueueElementIF {
        public SecureHash guid;
        public KeyEstablished(SecureHash guid) {
            this.guid = guid;
        }
	public String toString () {
	    return "(KeyEstablished with " + guid + ")";
	}
    }
        
    /**
     * Announcement that after multiple retries, we were still unable
     * to successfully establish a connection with the <tt>guid</tt>.
     */
    public class HandshakeFailed implements QueueElementIF {
        public SecureHash guid;
        public HandshakeFailed(SecureHash guid) {
            this.guid = guid;
        }
	public String toString () {
	    return "(HandshakeFailed with " + guid + ")";
	}
    }
}

                
		
