/**
 * LossPatch.java
 *
 * @author	Jeremy Stribling
 * @version	$Id: LossPatch.java,v 1.5 2003/04/29 21:10:16 strib Exp $
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;

import java.util.HashMap;
import ostore.util.InvalidConfigFileException;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.util.NodeId;
import ostore.util.Clock;
import ostore.util.Pair;
import ostore.util.PriorityQueue;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.UnknownEventException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SingleThreadedEventHandlerIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkFullException;   
import seda.sandStorm.api.StagesInitializedSignal;   

/**
 * A <code>Patch</code> that measures loss rate in percent, from 0 to 100, 
 * where 100 percent means all packets are lost.
 *
 * @author Jeremy Stribling
 * @version    $Id: LossPatch.java,v 1.5 2003/04/29 21:10:16 strib Exp $
 */
public class LossPatch extends Patch {

    public static final long VALUE_NON_LOSS = 0;
    public static final long VALUE_LOSS = 100;

    private int _loss_respond_every;
    
    // the number of probes seen since the last response
    // TODO: It's possible to have very small memory leaks in this map.
    // If a node goes away forever and stops sending you probes for good,
    // you will still have a reference to a NumSeen for it. I can't think
    // of a good way to clean the state up though, because even if you remove
    // a node from this patch, it doesn't mean that node won't still be 
    // expecting responses from you.
    private HashMap _num_seens;

    private int _max_seq_no;

    public LossPatch( NodeId self_node_id, Classifier classifier, 
		      PatchworkTable table, FaultAdmin fault_admin, 
		      Period period, int lossRespondEvery ) {
	super( self_node_id, classifier, table, fault_admin, period );
	class_tag = "LossPatch[" + _self_node_id + "]";
	_type = TYPE_LOSS;
	DEBUG = DebugFlags.NETWORK_PATCH_LOSS;
	_loss_respond_every = lossRespondEvery;
	_num_seens = new HashMap();

	// determine what the highest seqNo can be.  It should fall exactly
	// on a lossRespondEvery number
	int i;
	for( i = Integer.MAX_VALUE - _loss_respond_every + 1;
	     i % _loss_respond_every != 0 && i < Integer.MAX_VALUE; i++ );
	_max_seq_no = i;
	
    }

    protected void handleProbeResponse( ProbeMsg response, long timeRecvd ) {

	final String tag = class_tag + ".handleProbeResponse";
	
	// once we get a latency response, we should just record the latency
	long timeSent = registerResponse( response );

	// if it's -1, that means we have since timed out or never sent a
	// probe to this guy
	if( timeSent == -1 ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Ignoring response: " + response );
	    }
	    return;
	}

	// otherwise, we have our num received
	int numRecvd = response.data;

	// if the response indicates the node saw more than lossRespondEvery
	// probes, then it must have missed a force-response probe.
	// eventually, that missed probe will time out and all of
	// its probes will be counted as losses, so count all the extras
	// as 2 non-losses now to make up for it.
	//
	// TODO:  This is not the best thing to do, since the handling of the
	// timeout could happen after the next recalculation, making the next
	// stats look overly good but the stats after that look overly bad.
	// But I believe it will all pretty much balance itself out in the end.
	for( int i = 0; i < numRecvd-_loss_respond_every; i++ ) {
	    if( !response.quick_stat ) {
		makeObservation( response.peer, VALUE_NON_LOSS );
		makeObservation( response.peer, VALUE_NON_LOSS );
	    } else {
		handleQuickStatProbeResp( response.peer, VALUE_NON_LOSS );
		handleQuickStatProbeResp( response.peer, VALUE_NON_LOSS );
	    }
	}

	for( int i = 0; i < _loss_respond_every; i++ ) {
	    long value = VALUE_LOSS;
	    // record 0 numRecvd time, 100 lossRespondEvery-numLost times
	    if( i < numRecvd ) {
		value = VALUE_NON_LOSS;
	    }

	    // record it
	    if( !response.quick_stat ) {
		makeObservation( response.peer, value );
	    } else {
		handleQuickStatProbeResp( response.peer, value );
	    }
	}

    }

    protected void handleProbe( ProbeMsg inProbe ) {

	NumSeen ns = (NumSeen) _num_seens.get( inProbe.peer );
	if( ns == null ) {
	    ns = new NumSeen();
	    _num_seens.put( inProbe.peer, ns );
	}
	ns._num_seen++;

	// quickly respond if prompted, with the number seen
	if( inProbe.do_respond ) {
	    ProbeMsg pm = new ProbeMsg( inProbe.peer, ProbeMsg.MINIMUM_SIZE,
					inProbe.type, inProbe.seq_no, true, 
					inProbe.quick_stat );
	    pm.data = ns._num_seen;
	    ns._num_seen = 0;
	    dispatch( pm );
	}
    }

    protected void handleAlarm( AlarmMsg alarm ) {

	final String tag = class_tag + ".handleAlarm";

	if( DEBUG ) {
	    Debug.printtagln( tag, "Got alarm: " + alarm );
	}

	// time to send a probe to someone.
	int seqNo = getAndUpdateSeqNo( alarm.peer );

	// only send a probe that needs a response every lossRespondEvery
	// packets
	ProbeMsg pm = new ProbeMsg( alarm.peer, ProbeMsg.MINIMUM_SIZE,
				    _type, seqNo, false );
	if( seqNo % _loss_respond_every == 0 ) {
	    sendProbe( pm );
	} else {
	    dispatch( pm );
	}

    }

    protected void sendQuickStatProbe( NodeId peer ) {
	int seqNo = getAndUpdateSeqNo( peer );
	ProbeMsg pm = new ProbeMsg( peer, ProbeMsg.MINIMUM_SIZE,
				    _type, seqNo, false, true );
	if( seqNo % _loss_respond_every == 0 ) {
	    sendProbe( pm );
	} else {
	    dispatch( pm );
	}
    }

    protected void handleTimeout( ProbeTimeout timeout ) {

	final String tag = class_tag + ".handleTimout";

	if( registerTimeout( timeout ) ) {
	    // actually a valid timeout, so make a note and observations
	    if( DEBUG ) {
		Debug.printtagln( tag, "Timed out." );
	    }
	    for( int i = 0; i < _loss_respond_every; i++ ) {
		makeObservation( timeout.peer, VALUE_LOSS );
	    }

	    // back off on the timeout
	    // if we got a timeout, better pump up the timeout value
	    // remember that this sets the timeout for all patches of this node
	    long curr_timeout = _period.getTimeout( timeout.peer );
	    _period.setTimeout( timeout.peer, (long) 1.2*curr_timeout );
	    if( DEBUG ) {
		Debug.printtagln( tag, "Increasing timeout to " + 
				  1.2*curr_timeout + " for node " + 
				  timeout.peer );
	    }
	}

    }

    protected int getAndUpdateSeqNo( NodeId node ) {
	SeqNo sn = (SeqNo) _seq_nos.get( node );
	if( sn == null ) {
	    sn = new SeqNo();
	    // start sequence numbers at 1, so that the right number 
	    // of probes are sent before the first response.
	    sn.seq_no = 1;
	    _seq_nos.put( node, sn ); 
	}
	int tmp = sn.seq_no;
	if( sn.seq_no == _max_seq_no ) {
	    sn.seq_no = 1;
	} else {
	    sn.seq_no++;
	}
	return tmp;
    }

    private class NumSeen {
	int _num_seen = 0;
    }

}
