/**
 * Patch.java
 *
 * @author	Jeremy Stribling
 * @version	$Id: Patch.java,v 1.19 2003/06/19 21:38:06 srhea Exp $
 *
 * Copyright (c) 2003 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;

import java.net.DatagramPacket;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Vector;
import ostore.util.InvalidConfigFileException;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.network.RecordBytesMsg;
import ostore.util.ByteArrayOutputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.Clock;
import ostore.util.Pair;
import ostore.util.PriorityQueue;
import ostore.util.TypeTable;
import seda.nbio.NonblockingDatagramSocket;
import seda.nbio.Selectable;
import seda.nbio.SelectItem;
import seda.nbio.SelectSet;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.UnknownEventException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SingleThreadedEventHandlerIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkFullException;   
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.StagesInitializedSignal;   
import seda.util.MDWUtil;

/**
 * Abstract class from which all measurements modules (Patches) extend.
 * Contains a large amount of functionality that all Patches share.  Also
 * defines the types used to identify different types of measurements to
 * the outside world (and the inside world, as well).
 *
 * @author Jeremy Stribling
 * @version    $Id: Patch.java,v 1.19 2003/06/19 21:38:06 srhea Exp $
 */
public abstract class Patch {

    protected String class_tag;
    protected boolean DEBUG;

    protected static boolean sim_running;
    protected static boolean use_tcp;

    /**
     * Measurement type for loss rate
     */
    public static final byte TYPE_LOSS = 0;
    /**
     * Measurement type for latency
     */
    public static final byte TYPE_LATENCY = 1;
    /**
     * Measurement type for bandwidth
     */
    public static final byte TYPE_BANDWIDTH = 2;
    /**
     * Measurement type for recalculation (internal use only)
     */
    protected static final byte TYPE_RECALC = 3;
    /**
     * The total number of measurement types
     */
    public static final int NUM_TYPES = 4;

    // my node id
    protected NodeId _self_node_id;

    // my classifier
    protected Classifier _classifier;

    // my model
    protected PatchworkTable _table;

    // my model
    protected FaultAdmin _fault_admin;

    // my identifying type
    protected byte _type;

    // current sequence numbers (keyed off nodeIds)
    protected HashMap _seq_nos;

    // a map of registered probes times, keyed off a Pair( peer, seqNo )
    protected HashMap _probe_times;

    // a map of temporary "quick stat" nodes to their PatchworkStats
    protected HashMap _quick_stats;

    // a map of a Vector of userDatas and sinks for each node
    protected HashMap _userdata_sinks;

    // udp socket stuff
    
    protected static NonblockingDatagramSocket _udp_socket;
    protected static SelectSet _write_selset;
    protected static SelectItem _write_selitem;
    protected static Object _udp_lock = new Object();

    // the line of nodes waiting to be probed
    // priority = absolute time of probe
    protected PriorityQueue _queue;

    // my periodic fucnction
    protected Period _period;
    
    // next alarm data
    private long _next_alarm_time;
    private Object _next_alarm_token;
    private NodeId _next_alarm_node;

    /**
     * Create a new Patch.
     *
     * @param self_node_id  this node's id
     * @param classifier   this node's classifier
     * @param table    the collection of measurements from this node to 
     *                 other nodes
     * @param period  defines the periodic function governing the probes of
     *                this Patch
     */
    public Patch( NodeId self_node_id, Classifier classifier, 
		  PatchworkTable table, FaultAdmin fault_admin, 
		  Period period ) {
	_self_node_id = self_node_id;
	_classifier = classifier;
	_table = table;
	_fault_admin = fault_admin;
	_probe_times = new HashMap();
	_seq_nos = new HashMap();
	_userdata_sinks = new HashMap();
	_period = period;
	// arbitrary initial queue size:
	_queue = new PriorityQueue(50);
	resetAlarmState();
    }

    /**
     * Tell this Patch to begin monitoring a node at some probing level
     *
     * @param node   the node to begin monitoring
     * @param level  the probing level for that node
     * @param userData  an object that will identify this monitoring instance
     *                  to the calling application
     * @param sink   the sink on which to enqueue status responses
     */
    public void addNode( NodeId node, int level, Object userData, 
			 SinkIF sink ) {
	final String tag = class_tag + ".addNode";
	if( DEBUG ) {
	    Debug.printtagln( tag, "Adding node " + node + "at level " 
			      + level + "." );
	}
	// if it was actually changed, remove from and add to the queue
	if( _period.changePeriod( node, level ) ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Period for " + node + " is now " +
				  _period.getPeriod( node ) );
	    }
	    _queue.remove( node );
	    addToQueue( node );
	}

	// catalogue the userData and sink
	LinkedList l = (LinkedList) _userdata_sinks.get( node );
	if( l == null ) {
	    l = new LinkedList();
	    _userdata_sinks.put( node, l );
	}
	Pair p = new Pair( userData, sink );
	if( !l.contains( p ) ) {
	    l.add( p );
	}

    }

    /**
     * Tells this patch that a node wants to hear about status messages 
     * from everyone.
     *
     * @param userData  an object that will identify this monitoring instance
     *                  to the calling application
     * @param sink   the sink on which to enqueue status responses
     */
    public void addMonitorAll( Object userData, SinkIF sink ) {
	// do nothing
    }

    /**
     * Tell this Patch to stop monitoring a node
     *
     * @param node   the node to stop monitoring
     * @param level  the probing level for that node
     * @param userData  an object that will identify this monitoring instance
     *                  to the calling application
     * @param sink   the sink on which to enqueue status responses
     */
    public void removeNode( NodeId node, Object userData, SinkIF sink ) {
	final String tag = class_tag + ".removeNode";
	if( DEBUG ) {
	    Debug.printtagln( tag, "Removing node " + node + "." );
	}
	LinkedList l = (LinkedList) _userdata_sinks.get( node );
	if( l != null ) {
	    Pair p = new Pair( userData, sink );
	    l.remove( p );
	}
	if( l == null || l.isEmpty() ) {
	    // remove from all sets (Recalc will remove it from _table)
	    _period.removeNode( node );
	    _queue.remove( node );
	    // if this node already has an alarm set, cancel it
	    if( _next_alarm_node != null && _next_alarm_node.equals( node ) ) {
		_classifier.cancel( _next_alarm_token );
		setNextAlarm();
	    }
	}
    }

    /**
     * Add this node back on the queue.  If it is supposed to be probed
     * before the node that currently has the alarm set, interrupt that alarm
     * and set it for the new guy.
     *
     * @param node the node to add
     */
    protected void addToQueue( NodeId node ) {
	final String tag = class_tag + ".addToQueue";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called with " + node );
	}
	long relTime = _period.getPeriod( node );
	if( relTime < 0 ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Ignoring removed node: " + node );
		Debug.printtagln( tag, "Period for " + node + " is now " +
				  _period.getPeriod( node ) );
	    }
	    // this node has been removed, so no more alarms
	    return;
	}
	long alarmTime = relTime + Clock.current_date(_self_node_id).getTime();
	// if the new alarm has a higher priority than the old alarm
	// set it instead
	if( alarmTime < _next_alarm_time && _next_alarm_node != null &&
	    !_next_alarm_node.equals( node ) ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "cancelling alarm for " + node );
	    }
	    _classifier.cancel( _next_alarm_token );
	    _queue.insert( _next_alarm_node, _next_alarm_time, false );
	    resetAlarmState();
	    setNextAlarm( node, alarmTime );
	} else if(_next_alarm_node == null || !_next_alarm_node.equals(node)) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "inserting into queue: " + node );
	    }
	    // otherwise just insert it if it's not the current alarm node
	    _queue.insert( node, alarmTime, false );
	}
	// if there is no current alarm, you can set one now
	if( _next_alarm_node == null ) {
	    setNextAlarm();
	}
    }
    
    private void resetAlarmState() {
	_next_alarm_node = null;
	_next_alarm_time = -1;
	_next_alarm_token = null;
    }

    /**
     * Pick the node from the queue that needs to be probed the soonest,
     * and set the alarm for it.
     */
    protected void setNextAlarm() {
	// get the next one from the queue
	long time = _queue.topPriority( false );
	NodeId node = (NodeId) _queue.extract( false );
	if( node != null ) {
	    setNextAlarm( node, time );
	} else {
	    // no alarm, so reset state
	    resetAlarmState();
	}
    }

    /**
     * Set an alarm for the given node.
     *
     * @param node  the node for which to set the alarm
     * @param time  the absolute time at which to set the alarm (not relative!)
     */
    private void setNextAlarm( NodeId node, long time ) {
	final String tag = class_tag + ".setNextAlarm";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called with " + node + ", at " 
			      + time + "ms" );
	}
	long relTime = time - Clock.current_date(_self_node_id).getTime();
	AlarmMsg alarm = new AlarmMsg( node, _type );
	// if this time has already passed (dispatch_later isn't exact),
	// dispatch it now
	if( relTime <= 0 ) {
	    handleEvent( alarm, Clock.current_date(_self_node_id).getTime() );
	} else {
	    _next_alarm_node = node;
	    _next_alarm_time = time;
	    // set alarm
	    if( DEBUG ) {
		Debug.printtagln( tag, "setting alarm " + alarm + " for " 
				  + relTime + " ms" );
	    }
	    _next_alarm_token = _classifier.dispatch_later( alarm, relTime );
	}
    }

    /**
     * Handle events given the exact time they were received by Patchwork.
     *
     * @param msg   the msg to handle
     * @param timeReceived  the absolute time (in microseconds) msg 
     *                      was received
     */
    public void handleEvent( PatchMsg msg, long timeReceived ) {
	if( DEBUG ) {
	    Debug.printtagln( class_tag + ".handleEvent", "Got event: " +
			      msg );
	}
	if( msg instanceof ProbeMsg ) {
	    ProbeMsg pm = (ProbeMsg) msg;
	    // should we drop this?
	    if( _fault_admin.dropMessage( pm ) ) {
		if( DEBUG ) {
		    Debug.printtagln( class_tag + ".handleEvent",
				      "Dropping message due to fault injection:"
				      + msg );
		}
		return;
	    }
	    // if not a response, then you need to respond:
	    if( !pm.response ) {		
		handleProbe( pm );
	    } else {
		handleProbeResponse( pm, timeReceived ); 
	    }
	} else if( msg instanceof AlarmMsg ) {
	    resetAlarmState();
	    AlarmMsg alarm = (AlarmMsg) msg;
	    // make sure this node hasn't been removed
	    if( _period.getPeriod( alarm.peer ) >= 0 ) {
		handleAlarm( alarm );
		// now add this guy back into the queue and set the next alarm
		addToQueue( alarm.peer );
	    }
	    setNextAlarm();
	} else if( msg instanceof QuickStatReq ) {
	    handleQuickStatReq( (QuickStatReq) msg );
	} else if( msg instanceof QuickStatTimeout ) {
	    handleQuickStatTimeout( (QuickStatTimeout) msg );
	} else if( msg instanceof ProbeTimeout ) {
	    handleTimeout( (ProbeTimeout) msg );
	} else {
	    // unknown message
	    handleUnknown( msg );
	}
    }

    /**
     * Got a message we weren't expecting, did we?
     */
    protected void handleUnknown( PatchMsg msg ) {
	Carp.error( "Dropping unexpected message: " + msg );
    }

    protected synchronized void handleQuickStatReq( QuickStatReq msg ) {

	final String tag = class_tag + ".handleQuickStatReq";

	// make the quick stats map
	if( _quick_stats == null ) {
	    _quick_stats = new HashMap();
	}

	QuickStatTimeout timeout = new QuickStatTimeout( msg, _type );

        // put a new stats in for everyone
	for( int i = 0; i < msg.peers.length; i++ ) {
	    Object obj;
	    // make a new one only if one doesn't yet exist for that node
	    if( (obj = _quick_stats.get( msg.peers[i] )) == null ) {
		Vector vect = new Vector(1);
		if( DEBUG ) {
		    Debug.printtagln( tag, "Adding first message for " +
				      msg.peers[i] + ": " + msg ); 
		}
		vect.add( msg );
		_quick_stats.put( msg.peers[i], 
				  new Pair( new PatchworkStats( msg.peers[i], 
								_type ), 
					    vect ) );
		// send the first probe
		sendQuickStatProbe( msg.peers[i] );
	    } else {
		// someone else is already monitoring this node, so just
		// add yourself as another timeout 
		// since the probes are already happening, no need to start 
		// more
		if( DEBUG ) {
		    Debug.printtagln( tag, "Adding message for " +
				      msg.peers[i] + ": " + msg ); 
		}
		Pair pair = (Pair) obj;
		Vector vect = (Vector) pair.second;
		if( DEBUG ) {
		    Debug.printtagln( tag, "The vector had size " + vect.size()
				      + " before the addition." );
		}
		// we must protect all accesses to vect now that there's
		// another thread that can touch it.
		vect.add( msg );
	    }
	}

	// now set the timeout
	_classifier.dispatch_later( timeout, msg.time_to_resp );
	
    }

    /**
     * Got a response to a quick stat probe, with the given observed 
     * measurement.
     */
    protected synchronized void handleQuickStatProbeResp( NodeId peer, 
							  long value ) {

	final String tag = class_tag + ".handleQuickStatProbeResp";

	if( DEBUG ) {
	    Debug.printtagln( tag, "Got a quick stat response for " + 
			      peer + " of value " + value );
	}

	// look up the stats for this node
	Pair pair = (Pair) _quick_stats.get( peer );
	if( pair == null ) {
	    // either the timeout period is over or this is a mistake
	    if( DEBUG ) {
		Debug.printtagln( tag, "Got a QSProbe for " + peer + 
				  ", who needs no more stats.  Ignoring." );
	    }
	    return;
	}

	((PatchworkStats) pair.first).addStat( value );

	// if any of the requests were "just_once", then see if we're
	// all done
	Vector vect = (Vector) pair.second;
	boolean finish = false;
	boolean dont_set_finish = false;
	for( int i = 0; i < vect.size(); i++ ) {
	    QuickStatReq qsr = (QuickStatReq) vect.get(i);
	    if( qsr.just_once ) {
		if( !dont_set_finish ) {
		    finish = true;
		}
		// no need to remove if we're keeping track of which
		// nodes in qsr.peers have finished and which haven't
		//vect.remove( qsr );
		//i--;
		synchronized( qsr ) {
		    // find where you are in the peers
		    int j;
		    for( j = 0; j < qsr.peers.length; j++ ) {
			if( qsr.peers[j].equals( peer ) ) {
			    break;
			}
		    }
		    if( j == qsr.peers.length ) {
			Carp.die( "Couldn't find " + peer + " in " + qsr );
		    }
		    // now increment how many have finished and set your 
		    // status to done
		    if( !qsr.which_done[j] ) {
			qsr.how_many_done++;
			qsr.which_done[j] = true;
		    }
		    if( DEBUG ) {
			Debug.printtagln( tag, "how_many_done: " +
					  qsr.how_many_done +
					  ", peers.length: " + 
					  qsr.peers.length );
		    }
		    if( qsr.how_many_done == qsr.peers.length ) {
			handleQuickStatTimeout( new QuickStatTimeout( qsr, 
								      _type ));
		    }
		}
	    } else {
		// if we find one QSR that wasn't just once, 
		// then don't finish
		finish = false;
		dont_set_finish = true;
	    }
	}
	
	if( !finish ) {
	    // start the next probe
	    sendQuickStatProbe( peer );
	}
    }

    protected synchronized void handleQuickStatTimeout(QuickStatTimeout msg) {
	
	final String tag = class_tag + ".handleQuickStatTimeout";

	if( DEBUG ) {
	    Debug.printtagln( tag, "called with " + msg );
	}

	NodeId[] peers = msg.qs_req.peers;
	PatchworkStats[] stats = new PatchworkStats[ peers.length ];

	for( int i = 0; i < peers.length; i++ ) {

	    // retrieve the stats, delete entry if this is the last 
	    // relevant timeout
	    Pair pair = (Pair) _quick_stats.get( peers[i] );
	    if( pair == null ) {
		if( DEBUG && msg.qs_req.just_once ) {
		    Debug.printtagln( tag, "Got QuickStatTimeout for " + 
				      msg.qs_req + " but just_once already " +
				      "finished the job [" + peers[i] + "]." );
		} else if( !msg.qs_req.just_once ) {
		    Carp.die( "No quick_stats found for " + peers[i] + 
			      ", and just_once not set!" );
		}
		return;
	    }
	    stats[i] = (PatchworkStats) pair.first;
	    stats[i].recalculate(_table.getAlpha(), _table.getThreshold());
	    Vector vect = (Vector) pair.second;
	    vect.remove( msg.qs_req );
	    if( DEBUG ) {
		Debug.printtagln( tag, "Vector size: " + vect.size() );
	    }
	    if( vect.isEmpty() ) {
		// make sure this peer doesn't appear again in the 
		// peers list
		boolean again = false;
		for( int j = i+1; j < peers.length; j++ ) {
		    if( peers[i].equals(peers[j]) ) {
			again = true;
		    }
		}
		if( !again ) {
		    if( DEBUG ) {
			Debug.printtagln( tag, "Removing " + peers[i] +
					  " from quick_stats" );
		    }
		    _quick_stats.remove( peers[i] );
		}
	    }
	}

	// make the response
	QuickStatResp resp = new QuickStatResp( peers, _type, stats, 
						msg.qs_req.userData );
	
	// mark any as failed that had no stats
	for( int i = 0; i < peers.length; i++ ) {
	    if( stats[i].getCount() == 0 ) {
		resp.failures[i] = true;
	    }
	}

	enqueue( resp, msg.qs_req.sink );

    }

    /**
     * Send a standard quick probe to the given node.
     */
    protected void sendQuickStatProbe( NodeId peer ) {
	int seqNo = getAndUpdateSeqNo( peer );
	sendProbe( new ProbeMsg( peer, ProbeMsg.MINIMUM_SIZE,
				 _type, seqNo, false, true ) );	
    }

    /**
     * Record an observation (of the Patch's type) in the table
     *
     * @param peer  the node being observed
     * @param value   the value of the observed measurement
     */
    protected void makeObservation( NodeId peer, long value ) {
	makeObservation( peer, value, _type );
    }

    /**
     * Record an observation (of arbitrary type) in the table
     *
     * @param peer  the node being observed
     * @param value   the value of the observed measurement
     * @param type   the type of the observation
     */
    protected void makeObservation( NodeId peer, long value, byte type ) {
	final String tag = class_tag + ".makeObservation";
	if( DEBUG ) {
	    Debug.printtagln( tag, "Observing: peer=" + peer + ", value="
			      + value + ", type=" + type );
	}
	// however, if a node has been removed, we shouldn't observe anything
	if( _period.getPeriod(peer) < 0 ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Ignoring observation because node " +
				  peer + " has been removed." );
	    }
	    return;
	}
	_table.record( new PatchworkTable.Observation( peer, value, type ) );
    }

    /**
     * Handle the response to a probe.
     *
     * @param response  the response
     * @param timeRecvd  the absolute time the repsonse was received 
     *                   (in microseconds)
     */
    protected abstract void handleProbeResponse( ProbeMsg response, 
						 long timeRecvd );
    protected abstract void handleAlarm( AlarmMsg alarm );
    protected abstract void handleTimeout( ProbeTimeout timeout );

    protected void handleProbe( ProbeMsg inProbe ) {
	// quickly respond to a Probe msg
	if( inProbe.do_respond ) {
	    dispatch( new ProbeMsg( inProbe.peer, ProbeMsg.MINIMUM_SIZE,
				    inProbe.type, inProbe.seq_no, true, 
				    inProbe.quick_stat ) );
	}
    }

    /**
     * Send a probe to another node, expecting a response back.
     */
    protected void sendProbe( ProbeMsg probe ) {

	// since we're registering it, we should set the do_respond bit
	probe.do_respond = true;

	// set a timeout
	ProbeTimeout pt = new ProbeTimeout( probe.peer, probe.type, 
					    probe.seq_no );
	_classifier.dispatch_later( pt, _period.getTimeout(probe.peer) );

	// insert the time into a hashmap, keyed off a pair of nodeId and seqNo
	Pair key = new Pair( probe.peer, new Integer(probe.seq_no) );
	Long time_sent;	
	if( !sim_running ) {
	    time_sent = new Long(MDWUtil.currentTimeUsec());
	} else {
	    // TODO: if I'm in the simulator, I can directly access the
	    // latency to another node.  No need to send probes.
	    time_sent = 
		new Long(1000*Clock.current_date(_self_node_id).getTime());
	}
	_probe_times.put( key, time_sent );
	dispatch( probe );
    }

    /**
     * We received a response, so figure out state
     *
     * @return the time the probe was sent to which this was a response
     *         (in microseconds)
     */
    protected long registerResponse( ProbeMsg response ) {
	
	// look up the pair
	Pair key = new Pair( response.peer, new Integer(response.seq_no) );
	Long time_sent = (Long) _probe_times.remove( key );

	if( time_sent == null ) {
	    return -1;
	} else {
	    return time_sent.longValue();
	}

    }

    /**
     * Got a timeout for a probe.
     *
     * @return true iff a response was not yet received for the probe connected
     *         with this timeout
     */
    protected boolean registerTimeout( ProbeTimeout pt ) {
	
	Pair key = new Pair( pt.peer, new Integer(pt.seq_no) );

	// if we've got no response, this timeout is valid
	// return true if the timeout is valid
	return (_probe_times.remove( key ) != null );

    }

    /**
     * Get the next sequence number for this node, and update it for next time.
     *
     * @return the sequence number to use for the next probe.
     */
    protected int getAndUpdateSeqNo( NodeId node ) {
	SeqNo sn = (SeqNo) _seq_nos.get( node );
	if( sn == null ) {
	    sn = new SeqNo();
	    _seq_nos.put( node, sn ); 
	}
	int tmp = sn.seq_no;
	if( sn.seq_no == Integer.MAX_VALUE ) {
	    sn.seq_no = Integer.MIN_VALUE;
	} else {
	    sn.seq_no++;
	}
	return tmp;
    }

    /**
     * An internal method used for dispatching events
     */
    protected void dispatch( QueueElementIF event ) {

	final String tag = class_tag+".dispatch";
	if( DEBUG ) Debug.printtagln( tag, "called with event " + event );
	
	try {
	    _classifier.dispatch( event );
	} catch ( SinkFullException sfe ) {
	    Carp.die( "SinkFull: " + sfe ); 
	}                                         
	
    }

    /**
     * An internal method used for dispatching probe messages (UDP)
     */
    protected void dispatch( ProbeMsg event ) {

	final String tag = class_tag+".dispatch(ProbeMsg)";
	if( DEBUG ) Debug.printtagln( tag, "called with event " + event );

	// if this is the simulator, don't use UDP!
	if( sim_running || use_tcp ) {
	    dispatch( (QueueElementIF) event );
	    return;
	}
	
	// make the socket stuff completely static, because it doesn't 
	// matter how you send it out, and if you have too many
	// select sets the nbio jni code will throw an OutOfMemory exception
	synchronized( _udp_lock ) {
	    // if we haven't made a socket yet, do it
	    if( _udp_socket == null ) {
		try {
		    // bind to any port
		    _udp_socket = new NonblockingDatagramSocket();
		} catch( java.io.IOException e ) {
		    Carp.die( "Couldn't open udp socket: " + e );
		}
		_write_selset = new SelectSet();
		_write_selitem = 
		    new SelectItem( _udp_socket, 
				    (short)Selectable.WRITE_READY);
		_write_selset.add(_write_selitem);
	    }
	    
	    // serialize the message
	    ProbeMsg clone_msg = null;
	    try {
		clone_msg = (ProbeMsg) event.clone();
	    } catch( CloneNotSupportedException e ) {
		Carp.die( "WAAAAAA???????? " + e );
	    }
	    // set the peer (because the other end won't know how
	    clone_msg.peer = _self_node_id;
	    clone_msg.sender = _self_node_id; // this is dumb	
	    
	    // now serialize the message
	    byte[] data = new byte[Patchwork.MESSAGE_SIZE];
	    ByteArrayOutputBuffer ob = 
		new ByteArrayOutputBuffer( data, 0, 
					   Patchwork.MESSAGE_SIZE );
            clone_msg.peer.serialize (ob);
            clone_msg.sender.serialize (ob);
	    ob.add( clone_msg );
	    // data has been filled by the OutputBuffer
	    // ********** ASSUMPTION ********************
	    // The "Patchwork" port is always one above the real port
	    DatagramPacket dp = new DatagramPacket( data, 
						    ob.size(),
						    event.peer.address(),
						    event.peer.port()+1 );
	    // tell the network about these bytes
	    dispatch( new RecordBytesMsg( clone_msg, ob.size() ) );
	    // now wait until we can write, and then go for it
	    while ((_write_selitem.revents & Selectable.WRITE_READY) == 0) {
		_write_selset.select(0); // don't block while waiting
	    }
	    if (DEBUG) Debug.printtagln( tag, "A write is ready." );
	    // reset it after getting an event
	    _write_selitem.revents = 0;
	    try {
		_udp_socket.nbSend( dp );
	    } catch( java.io.IOException e ) {
		if( DebugFlags.LOG_BAD_NETWORK_EVENTS ) {
		    Carp.out( "Could not send probe " + event + ", got " + e +
			      ". Dropping it." );
		}
	    }
	}
    }

    protected void enqueue( QueueElementIF event, SinkIF sink ) {
	final String tag = class_tag+".enqueue";
	if (DEBUG) Debug.printtagln (tag, "enqueueing " + event);
	try {
	    sink.enqueue (event);
	}
	catch (SinkException e) {
	    Carp.die ("Where is the completion q?  " + e);
	}
    } 

    /**
     * Internal class for wrapping sequence numbers
     */
    protected class SeqNo {
	int seq_no = Integer.MIN_VALUE;
    }
    
}
