/**
 * Matrix.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil;

import ostore.util.Carp;
import ostore.util.Debug;
import java.util.Iterator;
import java.util.HashMap;
import java.text.NumberFormat;
import ostore.util.QuickSerializable;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QSException;
import ostore.util.ByteUtils;
import ostore.util.TypeTable;

/**
 * A sparse array of <code>Array</code>s.  Elements are indexed by
 * arbitrary <code>QuickSerializable</code> objects.
 * <p>
 *
 * This class is synchronized.
 *
 * @author	Dennis Geels
 * @version	$Id: Matrix.java,v 1.15 2002/07/27 08:20:19 strib Exp $
 */

public class Matrix implements Cloneable, QuickSerializable
{
    ////////////////////       CONSTANTS	////////////////////

    private static final boolean DEBUG = ostore.util.DebugFlags.OIL_MATRIX;

    ////////////////////    STATIC VARIABLES 	////////////////////

    /**
     * Used in <code>toString</code> for uniform number lengths.
     */
    private static final double ln_10 = Math.log(10);
    
    ////////////////////       VARIABLES	////////////////////

    /**
     * The internal storage mechanism.
     */
    protected HashMap map;
    
    ////////////////////  STATIC CONSTRUCTORS	////////////////////
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Construct a new <code>Matrix</code>.
     */
    public Matrix()
    {
	this.map = new HashMap();
    }

    /**
     * Implied from ostore.util.QuickSerializable interface
     * Construct an <code>Matrix</code> from its
     * <code>QuickSerializable</code> form.
     */
    public Matrix( InputBuffer buffer ) throws QSException
    {
	int size = buffer.nextInt();
	this.map = new HashMap( size );

	QuickSerializable id;
	Array row;
	for( int i=0; i<size; i++ ) {
	    // Deserialize the id
	    id = (QuickSerializable)buffer.nextObject();
	
	    // Now deserialize the row
	    row = (Array)buffer.nextObject();

	    // Finally, add the mapping to the <code>Matrix</code>.
	    put( id, row );
	}
    }

    /**
     * Construct an <code>Matrix</code> from its
     * <code>QuickSerializable</code> form.
     * See {@link #serialize(OutputBuffer, HashMap, HashMap)} for more details.
     * 
     * @param decompression_map	An <code>Integer->QuickSerializable</code> mapping
     * 			for decompressing IDs.
     * @param row_decompression_map  An <code>Integer->QuickSerializable</code>
     * 			mapping	for decompressing IDs within rows.
     */
    public Matrix( InputBuffer buffer, HashMap decompression_map,
		   HashMap row_decompression_map ) throws QSException
    {
	throw new NoSuchMethodError();
	// FIXME - InputBuffer doesn't support decompression
	
//  	int size = buffer.nextInt();
//  	this.map = new HashMap( size );

//  	int mapped_id;
//  	QuickSerializable id;
//  	Array row;
//  	for( int i=0; i<size; i++ ) {
//  	    // Deserialize the id, using the decompression mapping
//  	    mapped_id = buffer.nextInt();
//  	    id = (QuickSerializable)
//  		decompression_map.get( new Integer(mapped_id) );
//  	    if( id == null ) {
//  		throw new QSException("Matrix: corrupted cecompression map");
//  	    }

//  	    // Now deserialize the row
//  	    row = new Array( data, offset, row_decompression_map );

//  	    // Finally, add the mapping to the <code>Matrix</code>.
//  	    put( id, row );
//  	}
    }
    
    ////////////////////     STATIC METHODS 	////////////////////

    /**
     * Returns a human-readable representation of the specified matrix.
     *
     * @param matrix	the matrix to print
     *
     * @return a human-readable representation of the specified matrix.
     */
    public static String toString( double[][] matrix )
    {
	// Estimate the buffer length (ok to miss).
	// Assume average of at most two-digit ints for values.
	int num_frac_digits = 4; // Should == Array.num_frac_digits.
	int element_len = ((matrix[0].length * (num_frac_digits+4)) + 2);
	int len = (matrix.length * element_len) ;
	StringBuffer sbuf = new StringBuffer( len );
	
	sbuf.append('{');
	for( int i=0; i<matrix.length; i++ ) {
	    sbuf.append( Array.toString( matrix[i] ) );
	    sbuf.append(",\n");
	}
	sbuf.setCharAt( sbuf.length()-2, '}' );
	sbuf.setLength( sbuf.length()-1 );

	return sbuf.toString();
    }

    /**
     * Returns a human-readable representation of the specified matrix, with
     * indices.  This method should be used to print sub-matrices, or simply
     * because the indices are helpful.
     * This method calls {@link Array#toString(double[], int[])} on
     * each row of the matrix.
     *
     * @param matrix	the matrix to print
     * @param indices	the indices, in order, of the elements to print
     *
     * @exception IllegalArgumentException if width of <code>indices<code>
     * greater than width of corresponding row of <code>matrix</code>.
     *
     * @see #max
     *
     * @return a human-readable representation of the specified matrix.
     */
    public static String toString( double[][] matrix, int[][] indices )
    {
	NumberFormat nf = NumberFormat.getInstance();
	// Tell the formatter to print indices at the same length.
	int num_int_digits =
	    (int)(Math.ceil(Math.log(matrix.length)/ln_10));
		
	// Estimate the buffer length (ok to miss).
	// Assume average of at most two-digit ints for values.
	int array_index_len =
	    (int)(Math.ceil(Math.log(matrix[0].length)/ln_10));
	int num_frac_digits = 4; // Should == Array.num_frac_digits.
	int element_len = num_int_digits + 5 +
	    (indices[0].length * (num_frac_digits+array_index_len+7));
	int len = (indices.length * element_len);
	
	StringBuffer sbuf = new StringBuffer( len );
	
	sbuf.append('{');
	for( int i=0; i<indices.length; i++ ) {
	    sbuf.append( '<' );
	    // Set the formatter for indices.
	    nf.setMinimumIntegerDigits( num_int_digits );
	    nf.setMaximumFractionDigits( 0 );
	    nf.setMinimumFractionDigits( 0 );
	    sbuf.append( nf.format( i ) );
	    sbuf.append( ':' );
	    sbuf.append( Array.toString( matrix[i], indices[i] ) );
	    sbuf.append( ">" );
	    sbuf.append(",\n");
	}
	sbuf.setCharAt( sbuf.length()-2, '}' );
	sbuf.setLength( sbuf.length()-1 );

	return sbuf.toString();
    }

    /**
     * Returns the transpose of the specified matrix.
     * The width of the new matrix is the length of the specified matrix,
     * and vice versa. <b>WARNING</b> This method is in some ways even more
     * expensive than <code>toString</code>.
     *
     * @param matrix	the matrix to flip
     *
     * @return the transpose of the specified matrix.
     */
    public final static double[][] transpose( double[][] matrix )
    {
	int M = matrix.length;
	int N = matrix[0].length;
	double[][] t = new double[N][M];
	for( int i=0; i<M; i++ ) {
	    for( int j=0; j<N; j++ ) {
		t[j][i] = matrix[i][j];
	    }
	}
	return t;
    }

     /**
     * Finds the indices of the largest elements of the specified Matrix.
     * This method calls {@link Array#max} on each row of the matrix.  
     * Thus the width of <code>indices</code> determines the number of top
     * elements of each row to find.  To find the top elements of each
     * column (instead of by row) use <code>transpose</code> to flip the
     * matrix first.
     *
     * @param matrix	the matrix to check
     * @param indices	the matrix into which to put the (sorted) indices of
     * the largest elements of <code>matrix</code>.
     *
     * @exception IllegalArgumentException if width of <code>indices<code>
     * greater than width of corresponding row of <code>matrix</code>.
     */
    public static void max( double[][] matrix, int[][] indices )
    {
	// Find the top elements on a per-row basis.
	for( int i=0; i<indices.length; i++ ) {
	    Array.max( matrix[i], indices[i] );
	}
    }


    //////////////////// 	METHODS		////////////////////

    /**
     * Returns the size of the <code>Matrix</code>.
     * 
     * @return the number of <code>Array</code>s stored in this
     * <code>Matrix</code>. 
     */
    public synchronized int size()
    {
	return map.size();
    }

    /**
     * Returns the full size of the <code>Matrix</code>.
     * 
     * @return the total number of elements explicitly stored in the
     * <code>Array</code>s stored in this <code>Matrix</code>. 
     */
    public synchronized int full_size()
    {
	int size = 0;
	Iterator rows = ids();
	while( rows.hasNext() ) {
	    size += get((QuickSerializable)rows.next()).size();
	}
	return size;
    }
    
    /**
     * Returns an <code>Iterator</code> over the IDs stored explicitly in
     * this <code>Matrix</code>.
     *
     * @return a <code>HashIterator</code> over the backing map's keys.
     */
    public synchronized Iterator ids(){
	return map.keySet().iterator();
    }

    /**
     * Shows whether the specified ID is explicitly stored in this
     * <code>Matrix</code>.
     *
     * @param id	The unique ID for the row to check.
     * @return <code>true</code> iff <code>id</code> is explicitly mapped
     * to a row.
     */
    public synchronized boolean contains( QuickSerializable id )
    {
	return map.containsKey( id );
    }

    /**
     * Returns the <code>Array</code> indexed by the specified ID.
     *
     * @param id	The unique ID for a row of this
     * 			<code>Matrix</code>.
     * @return the <code>Array</code> stored under <code>id</code>, or
     * <code>null</code> if the ID is unknown.
     */
    public synchronized Array get( QuickSerializable id )
    {
	return (Array)map.get( id );
    }

    /**
     * Places the specified mapping into this <code>Matrix</code>.  If
     * <code>id</code> already maps to an <code>Array</code>, the old one
     * is replaced and returned.
     *
     * @param id	The index under which to store this row.
     * @param row	The <code>Array</code> to store for this index.
     * @return the <code>Array</code> previously mapped by this index, or
     * <code>null</code> if the ID is unknown.
     */
    public synchronized Array put( QuickSerializable id, Array row )
    {
	return (Array)map.put( id, row );
    }

    /**
     * Removes the row associated with <code>id</code> from this
     * <code>Matrix</code>, and returns the removed row.  The
     * specified <code>id</code> will no longer be associated with
     * any row in this <code>Matrix</code>.
     *
     * @param id        The index to be disassociated from the 
     *                  <code>Matrix</code>.
     * @return the <code>Array</code> previously associated with the index
     *  being removed, or <code>null</code> if the index is unknown.
     */
    public synchronized Array remove( QuickSerializable id )
    {
	return (Array) map.remove( id );
    }

    /**
     * Adds the contents of the specified <code>Matrix</code> to this one.
     * <code>Array</code>s for known IDs are <code>add</code>ed to the
     * current rows.  Unknown IDs are inserted directly (not
     * <code>clone</code>d). 
     *
     * @param addend	The <code>Matrix</code> whose contents to add
     */
    public synchronized void add( Matrix addend )
    {
	Iterator ids = addend.ids();
	QuickSerializable id;
	
	while( ids.hasNext() ) {
	    id = (QuickSerializable)ids.next();
	    add( id, addend.get(id) );
	}
    }

    /**
     * Adds the specified <code>Array</code> to this <code>Matrix</code>.
     * If the ID is known, the <code>Array</code> is <code>add</code>ed to
     * the current row.  Otherwise it is <code>clone</code>d and inserted.
     *
     * @param id	The unique ID for <code>addend</code>
     * @param addend	The <code>Array</code> to add
     */
    public synchronized void add( QuickSerializable id, Array addend )
    {
	if( map.containsKey( id ) ) {
	    get(id).add( addend );
	} else {
	    put( id, (Array)addend.clone() );
	}
    }

    /**
     * Performs a slightly deeper copy than <code>Object.clone</code>.
     * The mapping is cloned, so the new <code>Matrix</code> can be modified
     * independently of this one.  Neither the IDs nor the rows are
     * themselves cloned.
     *
     * @return a new <code>Matrix</code> with the same set of elements as
     * this one.
     */
    public synchronized Object clone(){
	Matrix cloned = new Matrix();
	cloned.map = (HashMap)map.clone();
	return cloned;
    }

    /**
     * Produce a human-readable version of this <code>Matrix</code>.
     * This method consumes <code>O(full_size())</code> time and
     * <code>String</code> length.
     *
     * @return a <code>String</code> of  &lt id:row &gt pairs.
     */
    public synchronized String toString()
    {
	StringBuffer sbuf = new StringBuffer();

	Iterator ids = ids();
	Object id;
	while( ids.hasNext() ) {
	    id = ids.next();
	    sbuf.append( '<' );
	    sbuf.append( id );
	    sbuf.append( ':' );
	    sbuf.append( map.get(id).toString() );
	    sbuf.append( ">\n" );
	}
	return sbuf.toString();
    }
    
    /* Specified by ostore.util.QuickSerializable */
    public void serialize( OutputBuffer buffer )
    {
	buffer.add( map.size() );

	Iterator ids = ids();
	QuickSerializable id;
	
	while( ids.hasNext() ) {
	    id = (QuickSerializable)ids.next();
	    
	    // First store the id
	    buffer.add( id );

	    // Now store the row
	    buffer.add( get(id) );
	}
    }

    /**
     * Stores this <code>Matrix</code> in its <code>QuickSerializable</code>
     * format, much like
     * <code>to_bytes(byte[], int[])</code> using an
     * index over the IDs for compression.  This method uses the specified
     * mappings to replace each ID with an associated integer.
     * This method allows the (possibly large) IDs to be stored
     * only once, saving significant storage space.
     * <p>
     * If this <code>Matrix</code> contains an ID not present in the
     * <code>compression_map</code>s, it adds a new mapping, using the next
     * available number.  Therefore the maps should not be serialized until
     * <b>after</b> this <code>Matrix</code>, to avoid losing the added
     * mappings. 
     *
     * @param data	The array into which to store, or <code>null</code>.
     * @param offset	The index in <code>data</code> at which to start.
     *			On return, contains the first unused offset.
     * @param compression_map	A <code>QuickSerializable->Integer</code>
     * 			mapping for ID compression.
     * @param row_compression_map  A <code>QuickSerializable->Integer</code>
     * 			mapping for ID compression within rows.
     */
//      public synchronized void to_bytes (byte [] data, int [] offset,
//  				       HashMap compression_map,
//  				       HashMap row_compression_map )
//      {
//  	ByteUtils.intToBytes( map.size(), data, offset );

//  	Iterator ids = ids();
//  	QuickSerializable id;
//  	Integer mapped_id;
	
//  	while( ids.hasNext() ) {
//  	    id = (QuickSerializable)ids.next();
		
//  	    // First store the id
//  	    mapped_id = (Integer)compression_map.get( id );
//  	    if( mapped_id == null ) {
//  		// Assume that the mapped indices are 0..size().
//  		mapped_id = new Integer( compression_map.size() );
//  		if( mapped_id.intValue() >= Integer.MAX_VALUE ) {
//  		    throw new IndexOutOfBoundsException( "Matrix.to_bytes:"
//  					 + " compression map too large " );
//  		}
//  		compression_map.put( id, mapped_id );
//  	    }
//  	    ByteUtils.intToBytes( mapped_id.intValue(), data, offset );
		
//  	    // Now store the row
//  	    get( id ).to_bytes( data, offset, row_compression_map );
//  	}
//      }
    /* Specified by ostore.util.QuickSerializable */
    public void serialize( OutputBuffer buffer, HashMap compression_map,
			   HashMap row_compression_map )
    {
	throw new NoSuchMethodError();
	// FIXME
    }    

}
