/**
 * Dealer.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 *
 * This copyright (below) covers the rsa function (also below).
 *
 * Copyright (C) 1995-1999 The Cryptix Foundation Limited.
 * All rights reserved.
 *
 * Use, modification, copying and distribution of this software is subject
 * the terms and conditions of the Cryptix General Licence. You should have
 * received a copy of the Cryptix General Licence along with this library;
 * if not, you can download a copy from http://www.cryptix.org/ .
 */

package ostore.security.thresh;
import ostore.util.Carp;
import java.util.Random;
import java.security.*;
import ostore.security.SafeBigInt;
import cryptix.jce.provider.rsa.RSAPrivateCrtKeyCryptix;
import cryptix.jce.provider.rsa.RSAPublicKeyCryptix;
import cryptix.jce.provider.rsa.RSAPrivateKeyCryptix;
import cryptix.jce.provider.random.DevRandom;

/**
 * A Key Dealer for an RSA based (k,l) Threshold Signature Scheme<BR>
 *
 * Reference: "Practical Threshold Signatures",<br>
 * Victor Shoup (sho@zurich.ibm.com), IBM Research Paper RZ3121, 4/30/99<BR>
 *
 * @author Steve Weis <sweis@cs.berkeley.edu>
 * @version    $Id: Dealer.java,v 1.7 2002/08/02 19:39:56 geels Exp $
 */
public class Dealer {

// Constants and variables
//............................................................................

    private static SecureRandom random;

    private final static SafeBigInt ONE = (SafeBigInt) SafeBigInt.valueOf(1l);
    private final static SafeBigInt TWO = (SafeBigInt) SafeBigInt.valueOf(2l);        

    /** Fermat prime F4. */    
    private final static SafeBigInt F4 = (SafeBigInt) SafeBigInt.valueOf(0x10001L);

    /** An arbitrary security parameter for generating secret shares */
    private final static int L1 = 128;
    
    private KeyPairGenerator kpgen=null;
    private int keysize;
    private KeyShare[] shares=null;
    
    /** Group Verifier */
    private SafeBigInt vk=null;

    /** Group Key */
    private GroupKey gk;
    
    /** (k,l) Parameters. k out of l shares needed for a signature */
    private int k, l;
    
    /** Randomly generated polynomial used to generate shares */
    private Poly poly;
	
    private boolean genInit = false;
    private boolean keyInit = false;

    static {
	final String tag = "Dealer.static";

	byte[] randSeed = new byte[20];
	(new Random()).nextBytes(randSeed);

	Provider cryptixRandom = new cryptix.jce.provider.CryptixRandom ();
	Security.addProvider (cryptixRandom);
	if (!DevRandom.isAvailable ()) Carp.die( tag + ": need DevRandom.");
	try {
	random = SecureRandom.getInstance ("DevRandom", "CryptixRandom");
	if (random == null) Carp.die( tag + ": need DevRandom.");
	}
	catch (Exception e) {
	if (random == null) Carp.die( tag + ": caught " + e + 
		" while trying to init DevRandom.");
	}
    }

// Constructors
//............................................................................

    /**
     * Create a new instance of a key dealer
     *
     * @param provider - the provider to use for RSA KeyPair generator
     * @param keysize - the size of the group key 
     */
    public Dealer(String provider, int keysize) {
	try {
	    String algorithm = "RSA";
	    if (DEBUG)
		debug("Testing " 
		      + algorithm + " ("
		      + provider+ ") with "
		      + keysize + " bit Keys...");
	    
	    this.kpgen = KeyPairGenerator.getInstance(algorithm,provider);
	    
	    // The keypair generator generates primes of size (keysize/2)
	    // We want p = 2p' len(p) = keysize/2, so len(p') = keysize/2 - 1
	    // So it's keysize-2
	    this.kpgen.initialize(keysize-2);
	    this.keysize = keysize;
	}
	catch (Exception e) {
	    System.err.println("Dealer class could not instantiate KeyPairGenerator RSA -" 
			       + provider);
	    e.printStackTrace();
	}
	// Inidicates key generator has been initialized
	this.genInit = true;
    }


// Public Methods
//............................................................................

    /**
     * Generate a group public key and l shares for a (k,l) <BR>
     * threshold signatures scheme<BR>
     *
     * @param k - k valid signatures will verify
     * @param l - l members of the group will receive shares
     *
     * @throws ThresholdSigException 
     */
    public void generateKeys(int k, int l) 
	throws ThresholdSigException {

	this.checkGenInit();

	this.k = k;
	this.l = l;

	KeyPair groupKeyPair = null;
	SafeBigInt pr, qr, d, e, m, n;
	n=m=pr=qr=null;
	
	// Create the group key pair
	try {
	    while (true) {
		if (DEBUG)
		    debug("Attempting to generate group keypair..");
		
		/* Generate a special prime keypair */
		pr = generateSpecialPrime(kpgen);
		qr = generateSpecialPrime(kpgen);		

		/* Creates a KeyPair object from special primes */
		groupKeyPair = generateGroupKeyPair(pr,qr,l);
		
		/* Extracts modulo from the keypair object */
		n = new SafeBigInt (((RSAPublicKeyCryptix) groupKeyPair.getPublic ()).getModulus());
		break;
	    }

	} catch (Exception ex) {
	    ex.printStackTrace();
	    debug("Generator could not create group key pair");
	    throw new 
		ThresholdSigException("Dealer.generateKeys could not "
		+ "create group key pair");	    
	}
	
	// Generate Secret Key KeyShares
	m = pr.multiply(qr);
	e = new SafeBigInt (((RSAPublicKeyCryptix)groupKeyPair.getPublic()).getPublicExponent());
	d = e.modInverse(m);



	// Create Secret KeyShares and KeyShare Verifiers
	shares = this.generateKeyShares(d,m,k,l,n);	

	// Create verification shares
	vk = this.generateVerifiers(n,shares);


	// Create a group key
	this.gk = new GroupKey();
	this.gk.initialize(k,l,keysize,vk,e,n,groupKeyPair);

	this.keyInit = true;
    }


    /** 
     * Returns the group key
     */
    public GroupKey getGroupKey() 
    	throws ThresholdSigException {
    	checkKeyInit();
    	return this.gk; 
    }

    public ThresholdPublicKey getPublicKey () throws ThresholdSigException {
	return getGroupKey ().getPublicKey ();
    }

    /** 
     * Returns the initialized secret key shares
     */
    public KeyShare[] getShares() 	
	throws ThresholdSigException {
	checkKeyInit();
	return shares; 
    }

// Initialization Checks
//............................................................................

    private void checkGenInit()
	throws ThresholdSigException
    {
	if (genInit == false) {
	    if (DEBUG)
		debug("Key pair generator was not initialized for Dealer");
	    throw new ThresholdSigException("Key pair generator not initialize");
	}
	
    }

    private void checkKeyInit()
	throws ThresholdSigException
    {
	if (keyInit == false) {
	    if (DEBUG)
		debug("Key pair has not been initialized by generateKeys()");
	    throw new ThresholdSigException("Key pair has not been initialized by generateKeys()");
	}
	
    }

// Private Methods
//............................................................................

    /**
     * Generates a prime p' such that such that 2p'+1 is prime
     *
     * @param kpgen - KeyPair generator initialized to generate RSA Keys
     * @return - A prime p' such that 2p'+1 is prime
     */
    private SafeBigInt generateSpecialPrime(KeyPairGenerator kpgen) {

	RSAPrivateCrtKeyCryptix privKey;
	SafeBigInt t1, t2, t3, t4;
	KeyPair kp;
	
	// debug
	int i = 0;
	
	while(true) {		
	    // Using RSA Key gen because it is convienent
	    kp = kpgen.generateKeyPair();
	    privKey = (RSAPrivateCrtKeyCryptix)kp.getPrivate();		
	    t1 = new SafeBigInt (privKey.getPrimeP());
	    t2 = new SafeBigInt (privKey.getPrimeQ());
	    
	    t3 = (t1.multiply(TWO)).add(ONE);
	    t4 = (t2.multiply(TWO)).add(ONE);
	    
	    
	    if (t3.isProbablePrime(80)) {
		if (DEBUG)
		    debug(i + " iterations to find special prime");
		return t1;
	    }
	    
	    if (t4.isProbablePrime(80)) {
		if (DEBUG)
		    debug(i + " iterations to find special prime");
		return t2;
	    }
	    
	    // debug
	    i++;
	}
	
    }   

    /**
     * Generates an "RSA" keypair<BR>
     * This is not a traditional RSA keypair and is built with special
     * properties for the Threshold Signature scheme<BR>
     * It cannot be used in place of standard RSA keys<BR>
     *
     * @param pr - A special prime such that 2pr+1 is prime
     * @param qr - A special prime such that 2qr+1 is prime
     * @param l - the size of the group in a (k,l) threshold scheme
     * @return - A "RSA" keypair
     */
    private KeyPair generateGroupKeyPair(SafeBigInt pr, 
					 SafeBigInt qr,
					 int l) {
	SafeBigInt p, q, n, e, m, d;
	
	RSAPublicKeyCryptix PK;
	RSAPrivateKeyCryptix SK;
	
	KeyPair kp;
	SafeBigInt groupSize;
	
	// m = pr*qr
	m = pr.multiply(qr);
	
	// p = 2*pr + 1
	p = (pr.multiply(TWO)).add(ONE);
	
	// q = 2*qr + 1
	q = (qr.multiply(TWO)).add(ONE);	    	    
	
	// n = p*q
	n = p.multiply(q);

	// the RSA public exponent must be a prime bigger than
	// l, the size of the group
	groupSize = SafeBigInt.valueOf((long)l);
	// If group size is less than Fermat's prime, just use it.
	if (groupSize.compareTo(F4) < 0)
	    e = F4;
	// Otherwise pick a prime bigger then groupSize
	else 
	    e = new SafeBigInt(groupSize.bitLength()+1, 80, random);

	
	// Note: This is not a standard RSA Key Pair
	// Usually:
	//	SafeBigInt phi = (p.subtract(ONE)).multiply(q.subtract(ONE));
	//	d = e.modInverse(phi);
	d = e.modInverse(m);

	PK = new RSAPublicKeyCryptix(n.toBigInteger (),e.toBigInteger ());
	SK = new RSAPrivateKeyCryptix(n.toBigInteger (),d.toBigInteger ());
	kp = new KeyPair (PK, SK);
	    
	return kp;
    }



    /**
     * 
     * Generates secret shares for a (k,l) threshold signatures scheme<BR>
     *
     * @param k - k valid signatures will verify
     * @param l - l members of the group will receive shares
     *
     * @return An array of l secret shares
     * @throws ThresholdSigException 
     */
    // TODO: Merge Dealer.generateShares and Dealer.generateVerifiers
    // and generate them simultaneously
    private KeyShare[] generateKeyShares(SafeBigInt d, SafeBigInt m, 
					int k, int l, 
					SafeBigInt n) {       	
    	SafeBigInt[] secrets;
	SafeBigInt rand;
	int randbits;

	this.poly = new Poly(d,k-1,m);
	secrets = new SafeBigInt[l];	
	randbits = n.bitLength()+L1-m.bitLength(); 

	// Generates the valies f(i) for 1<=i<=l
	// and add some large multiple of m to each value
	for (int i=0; i<l; i++) {
	    secrets[i] = poly.eval(i+1);
	    rand = (new SafeBigInt(randbits,random)).multiply(m);
	    secrets[i] = secrets[i].add(rand);
	}       

	SafeBigInt delta = Dealer.factorial(l);

	KeyShare[] s = new KeyShare[l];
	for (int i=0; i<l; i++)
	    s[i] = new KeyShare(i+1,secrets[i],n, delta);

	return s;
    }

    /**
     * Creates verifiers for secret shares to prevent corrupting shares<BR>
     *
     * Computes v[i] = v^^s[i] mod n, where v is an element of the squares of n<BR>
     * Returns the group verifier and sets the verifier in each share<br>
     *
     * @param n - Size of modulo for group key
     * @param secrets - array of shares
     *
     * @return the group verifier
     */
    // TODO: Merge Dealer.generateShares and Dealer.generateVerifiers
    // and generate them simultaneously
    private SafeBigInt generateVerifiers(SafeBigInt n, 
					   KeyShare[] secrets) {

	debug("Generating Verifiers");

	//	SafeBigInt[] v;
	SafeBigInt rand = null;
	
	//	v = new SafeBigInt[secrets.length];
		
 	for (int i=0; i<secrets.length; i++) {
 	    // rand is an element of Q*n (squares of relative primes mod n)
 	    while(true) {
 		rand = new SafeBigInt(n.bitLength(),random);
 		// ensure that rand is relatively prime to n
 		SafeBigInt d = rand.gcd(n);
 		if (d.compareTo(ONE) == 0)
 		    break;
 		// Else d was not relatively prime
 		// Note: This should be very rare 
 		debug("Verifier was not relatively prime");
 	    }
 	    // Rand is an element of QsubN - square mod n
	    // This value is the group verifier
 	    rand = rand.multiply(rand).mod(n);

	    secrets[i].setVerifiers(rand.modPow(secrets[i].getSecret(),n),
				   rand);
 	}
	
	return rand;
    }

    /**
     * Returns the factorial of the given integer as a SafeBigInt 
     *
     * @return l!
     */
    private static SafeBigInt factorial(int l) {
	SafeBigInt x = SafeBigInt.valueOf(1l);
	for (int i=1; i<=l; i++) 
	    x = x.multiply(SafeBigInt.valueOf((long)i));
	
	return x;
    }

    
    
// Debugging 
//............................................................................

    private final static boolean DEBUG = true;

    private static void debug(String s) 
    { 
	System.err.println("Dealer: " + s); 
    }
    
    public static void main(String[] args) {
	int keysize  = 128;	 
	if (args.length>0) 
	    try {
		keysize = Integer.parseInt(args[0]);
	    } catch (Exception e) { }
	
	try {
	    Dealer d = new Dealer("CryptixCrypto",keysize);
	    d.generateKeys(3,5);
	    
	} catch (Exception e) {
	    e.printStackTrace();	    
	}
    }
	
// Self-Testing and Verification Methods
//............................................................................

    // debug
    private void verifyPoly() {
	Poly f = new Poly(ONE,4,TWO.multiply(TWO));
	debug(f.toString());	
	debug(f.eval(0).toString(10));
	debug(f.eval(1).toString(10));
	debug(f.eval(2).toString(10));
    }

    // debug
    // This method does not really apply. The group keypair
    // is not a standard rsa keypair. The private key is built into
    // the secret shares..
    private void verifyKeyPair(KeyPair groupKeyPair) {
	debug("Verifying group raw RSA Encryption:");
	String test = "Shaolin";
	
	SafeBigInt e, d, n;
	RSAPublicKeyCryptix pub = 
	    (RSAPublicKeyCryptix)groupKeyPair.getPublic();
	RSAPrivateKeyCryptix priv = 
	    (RSAPrivateKeyCryptix)groupKeyPair.getPrivate();
	
	e = new SafeBigInt (pub.getPublicExponent());
	n = new SafeBigInt (pub.getModulus());
	d = new SafeBigInt (priv.getPrivateExponent());

	debug("n bit length: "+ n.bitLength());
	
	SafeBigInt plain = new SafeBigInt(test.getBytes());
	debug("Plain: " + plain.toString(36));
	SafeBigInt cipher = rsa(plain,n,e);
	debug("Cipher: "+ cipher.toString(36));
	SafeBigInt decrypted = rsa(cipher,n,d);
	debug("Decrypted: "+ decrypted.toString(36));
	
    }

    // This function is taken from the Cryptix JCE code verbatim.  I would
    // love to just call it from there, but the RSAAlgorithm class is not
    // visible outside the package.  :(

    private SafeBigInt rsa (SafeBigInt X, SafeBigInt n, SafeBigInt exp)
    {
        return X.modPow(exp, n);
    }
    
    // debug
    private void verifyKeyShares(Poly f, SafeBigInt secrets[], SafeBigInt m) {
	for (int i=0; i<secrets.length; i++) {
	    SafeBigInt temp = f.eval(i+1);
	    debug("Secret: " + secrets[i].toString(16));
	    debug("f(i) mod m: " + secrets[i].mod(m).toString(16));
	    debug("Secret mod m: " + secrets[i].mod(m).toString(16));
	}
    }
    

}


