/*
 * RouteEntry.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl;
import java.util.HashMap;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Map;

import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.NodeId;

import java.util.Set;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.Pair;

/**
 * Implements the routing table entry with multiple outgoing routes. This
 * class supports sorted insertion of routes into the entry, and returns
 * the best entry for general usage.
 *
 * @author     Ben Y. Zhao
 * @version    $Id: RouteEntry.java,v 1.31 2003/05/08 02:03:59 strib Exp $
 */
public class RouteEntry implements QuickSerializable {

  private static final boolean DEBUG = DebugFlags.TAPESTRY_ROUTING_TABLE;

  public static final int REDUNDANT_LVL = 3;
    //redundant_lvl of 3 means that there are 2 backups per route entry
  public static final int LOCKED=1;
  public static final int LEAVING=2;

  public static final int LOSS    = 2001;
  public static final int LATENCY = 2002;

  public static final short STATUS_ALIVE = 2999;
  public static final short STATUS_DEAD = 3000;
  public static final short STATUS_REPAIR_ONE = 3001;
  public static final short STATUS_REPAIR_TWO = 3002;
    
  //for fault tolerance routing
  private int frls;
  private NodeEntry [] routes;
  private static final int FRLS_INITIAL = -1;

  public static class NodeEntry {
    public NodeId  address;
    public boolean leaving;
    public boolean locked;

    public short status; //as specified by patchwork NodeStatusMsg
    
    //These three fields should be update by Patchwork.
    public double  latency;
    public double  loss_rate;
    public int     dead_period;
    
    public NodeEntry (NodeId addr, Double lat, boolean leave, boolean lock) {
      address = addr;
      if (lat == null)
	latency = 10000.0;
      else 
	latency = lat.doubleValue();
      loss_rate = 0.0;
      leaving = leave;
      this.locked = lock;

      status = STATUS_ALIVE;
    }

    public String toString() {
      String s = "[";
      if (leaving) s+= "leaving";
      if (locked) s+= "locked";
      s+="]";
      return "( "+address+", "+latency+loss_rate+" )"+s;
    }
  }

  public RouteEntry() {
    routes = new NodeEntry [REDUNDANT_LVL];
    frls = FRLS_INITIAL;
  }

  public RouteEntry(NodeId route, Double latency) {
    routes = new NodeEntry [REDUNDANT_LVL];
    routes[0] = new NodeEntry(route, latency, false, false);

    frls = FRLS_INITIAL;
  }

  public RouteEntry(NodeId route, Double latency, int flag) {
    routes = new NodeEntry [REDUNDANT_LVL];
    if (flag == LOCKED) 
      routes[0] = new NodeEntry(route, latency, false, true);
    else if (flag == LEAVING)
      routes[0] = new NodeEntry(route, latency, true, false);
    else 
      routes[0] = new NodeEntry(route, latency, false, false);

    frls = FRLS_INITIAL;
  }

  /**
   ** Method to recalculate who the FRLS entry should point to.  By
   ** default, the criteria we're judging FRLS on is failure rate, and
   ** the threshold value we use is Router.THRESHOLD_FRLS.
   ** @return Pair of Ints (oldfrls, newfrls), or null if no change made.
   **/
  public Pair recalcFRLS() {
    int newfrls;
    final String tag = "RouteEntry.recalcFRLS";
    if( DEBUG ) {
      Debug.printtagln(tag, "RouteEntry: recalcFRLS calculating new FRLS.");
    }
    for (int i=0; i<REDUNDANT_LVL; i++) {
      if (routes[i] == null) continue;
      if (routes[i].loss_rate < Router.THRESHOLD_FRLS) {
	newfrls = i;
	if (newfrls != frls) {
	  Pair p = new Pair(new Integer(frls), new Integer(newfrls));
	  frls = newfrls;
	  if( DEBUG ) {
	    Debug.printtagln(tag, "RouteEntry: recalcFRLS result: "+frls+"->"+
			     newfrls);
	  }
	  return p;
	} else 
	  return null;
      }
    }
    //nothing satisfies, so just set to the first nonnull entry;
    for (int i=0; i<REDUNDANT_LVL; i++) {
      if (routes[i] != null) {
	frls = i;
	return null;
      }
    }
    frls = FRLS_INITIAL;
    //this entry is empty
    return null;
  }

  public boolean need_multicast(int mcast_threshold) {
    if( frls < 0 ) {
      return false;
    }
    return (routes[frls].loss_rate > mcast_threshold);
  }
  
  public NodeEntry [] getRoutes() {
    return routes;
  }
  
  public int get_frls() {
    return frls;
  }
  
  /**
   * Method to return the frls NodeId address of this RouteEntry.  This
   * is the method we should use for ALL routing lookups.
   * Returns null if no such NodeId exists.
   */
  public NodeId getFRLS() {
    if( frls < 0 ) {
      return null;
    }
    if (frls < REDUNDANT_LVL && routes[frls] != null)
      return routes[frls].address;
    else return null;
  }

  public NodeEntry getEntryAt(int i) {
    if( i < 0 ) {
      return null;
    }
    return routes[i];
  }

  public void setEntryAt(int i, NodeEntry ne) {
    if( i < 0 ) {
	return;
    }
    routes[i] = ne;
  }

  /**
   * Method to return the first (primary) NodeId address of this
   * RouteEntry. This method should be used RARELY and SPARINGLY 
   * to examine the routes, and at all other times, getFRLS should
   * be used.
   * Returns null if no such NodeId exists.
   */
  public NodeId getFirst() {
    if (routes[0] != null)
      return routes[0].address;
    else return null;
  }

  /**
   * Method to return a NodeId at a position in the RouteEntry.
   * Result could be null if no NodeEntry exists at that position.
   */
  public NodeId getAtIndex(int i) {
    if( i < 0 ) {
      return null;
    }
    if (i < REDUNDANT_LVL && routes[i] != null)
      return routes[i].address;
    return null;
  }
  
  /**
   * Method to ask whether a NodeEntry at a particular position is 
   * in the process of leaving or not. 
   */
  public boolean isLeavingAtIndex(int i) {
    if( i < 0 ) {
      return false;
    }
    if (i < REDUNDANT_LVL && routes[i] != null)
      return (routes[i].leaving);
    return false;
  }

  /**
   * Method to ask whether a NodeEntry at a particular position is 
   * marked as locked (in the process of insertion) or not. 
   */
  public boolean isLockedAtIndex(int i) {
    if( i < 0 ) {
      return false;
    }
    if (i < REDUNDANT_LVL && routes[i] != null)
      return (routes[i].locked);
    return false;
  }

  /**
   ** Search for an existing nodeid if it's in this entry, and remove it
   ** @param node The node to be replaced
   ** @return pair of oldfrls, newfrls if FRLS has changed, 
   ** and a republish is necessary
   **/
  public Pair removeEntry(NodeId node) {
    for (int i=0; i<REDUNDANT_LVL; i++) {
      if (routes[i]!=null && routes[i].address.equals(node)) {
	//found it, now remove, move up the other entries, and insert
	//the replacement node
	routes[i] = null;
	for (; i<REDUNDANT_LVL-1; i++) {
	  routes[i] = routes[i+1];
	}
	//BUGFIX
	routes[i] = null;
	//ASSERT i == REDUNDANT_LVL || routes[i] == null 
	break;  //we're done now
      }
    }

    Pair p = recalcFRLS();
    return p;
  }

  /**
   * Function to mark a node as leaving
   * @param leavingnode The node in process of leaving the network
   * @return whether the node was found in this routeEntry
   */
  public boolean markLeaving(NodeId leavingnode) {
    for (int i=0; i<REDUNDANT_LVL; i++) {
      if (routes[i] == null) return false;
      if (routes[i].address.equals(leavingnode)) {
	routes[i].leaving = true;
	return true;
      }
    }
    return false;
  }

  /**
   * Function used to add a new entry to the RouteEntry.  We need to sort this
   * guy compared to our current entries.
   * @param removed_node Parameter hack for returning the node (if any) that
   *                     was kicked out of the table during this writing.
   *                     Like a pointer in C.
   * @return True or false, whether this entry was added to the route entry.
   */
  public boolean add(NodeId newnode, Double latency, NodeId[] removed_node,
		     Pair [] frlschanged) {
    final String tag = "RouteEntry.add";

    //first we check to see if the entry is there, if so just update latency
    for (int i=0; i<REDUNDANT_LVL && (routes[i] != null); i++) {
      if (newnode.equals(routes[i].address)) {
	if (DEBUG) Debug.printtagln(tag, 
			  "Entry already here. Will just update latency.");
	if (latency != null) //only overwrite with a real value
	  routes[i].latency = latency.doubleValue();
	return false;  //we didn't add anything new, so return false
      }
    }
    boolean written = false;
    NodeEntry ne = new NodeEntry(newnode, latency, false, false);
    NodeEntry temp;
    for (int i=0; i < REDUNDANT_LVL; i++) {
      if (DEBUG) {
	System.err.println("RouteEntry.add: Looking to replace entry nodeid="
			   +routes[i]);
      }
      if (routes[i] == null) 
      {
	//if empty entry, then just put it in
	routes[i] = ne;
	written = true;
	break;
      } 
      else if (ne.latency <  routes[i].latency) {
	temp = routes[i];
	routes[i] = ne;
	ne = temp;
	written = true;
	// if we're kicking someone out of the entry, save their node id
	// in the hack-style array
	if( i == (REDUNDANT_LVL-1) ) {
	    removed_node[0] = ne.address;
	}
      }
    }

    //now we recalculate the frls 
    frlschanged[0] = recalcFRLS();
    
    if (DEBUG) {
      Debug.printtagln(tag, "After Adding entry nodeid:"+newnode);
      for (int i=0; i<REDUNDANT_LVL; i++) {
	Debug.printtagln(tag, "RouteEntry: "+routes[i]);
      }
    }
    return written;
  }

  public void switchRoutes(int from, int to) {
    if (from >= REDUNDANT_LVL || to >= REDUNDANT_LVL ||
	from < 0 || to < 0 ||
  	routes[from] == null || routes[to] == null) {
      if (DEBUG) System.err.println("RouteEntry.switch called with "+
				    "invalid indices.");
      return;
    }
    NodeEntry t = routes[to];
    routes[to] = routes[from];
    routes[from] = t;
  }


  /**
   * Sort the entry according the given criteria.
   */
  public void sortRoutes(int criteria) {
      System.err.println("This function has not been implemented yet!");
  }


  /**
   * Search to see if route is actually part of this
   * routeEntry. If so, then update its latency, and
   * then do a sort to reorder the routes by latency
   */
  public void updateLatency(NodeId route, Double latency) {
    for (int i=0; i < REDUNDANT_LVL; i++) {
      if (routes[i] != null && routes[i].address.equals(route)) {
	routes[i].latency = latency.doubleValue();
	break;
      }
    }
  }

  public String toString() {
    return "RE:size="+size()+":First()="+getFirst();
  }

  //returns number of non-null entries, including the primary
  public int size() {
    int i;
    for (i=0; i<REDUNDANT_LVL; i++)
      if (routes[i]==null) break;
    return i;
  }

  public RouteEntry(InputBuffer buffer) throws ostore.util.QSException {
    routes = new NodeEntry [REDUNDANT_LVL];
    int size = buffer.nextInt();
    NodeId addr;
    Double lat;
    boolean leave, locked;
    for (int i = 0; i<size; i++) {
      addr = (NodeId) buffer.nextObject();
      lat = new Double(buffer.nextDouble());
      leave = buffer.nextBoolean();
      locked = buffer.nextBoolean();
      routes[i] = new NodeEntry(addr, lat, leave, locked);
    }
  } 

 public void serialize (OutputBuffer buffer) { 
   buffer.add(size());
   for (int i=0; i< REDUNDANT_LVL; i++) {
     if (routes[i] == null)
       break;
     buffer.add(routes[i].address);
     buffer.add(routes[i].latency); 
     buffer.add(routes[i].leaving);
     buffer.add(routes[i].locked);
   }
 }

}
