/*
 * RoutingTable.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import ostore.util.AssertionViolatedException;
import ostore.util.DebugFlags;
import ostore.util.Debug;
import ostore.util.NodeId;
import ostore.util.SecureHash;
import seda.sandStorm.api.QueueElementIF;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.Pair;
import ostore.util.SHA1Hash;
import ostore.tapestry.api.TapestryNeighborInfo;
import ostore.dispatch.Classifier;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkIF;
import ostore.tapestry.api.TapestryRoutingTableChanged;
import ostore.network.patchwork.MonitorStopMsg;
import ostore.util.Carp;

//includes for Patchwork update functions
import ostore.network.patchwork.PatchworkTable;
import ostore.network.patchwork.Patch;
import ostore.network.patchwork.DeltaMsg;
import ostore.network.patchwork.PatchworkStats;

/**
 * Implements the routing table interface.
 *
 * @author     Ben Y. Zhao
 * @version    $Id: RoutingTable.java,v 1.50 2003/05/19 20:20:55 strib Exp $
 */
public class RoutingTable implements QuickSerializable {
  
  
  public static class Available implements QueueElementIF {
    public Available (RoutingTable t, HashMap rev, HashMap objs) {
	    table = t;
	    revpointers = rev;
	    objlevels = objs;
	}

    public String toString () {
      return "RoutingTable.available: "+table.toShortString()
	+"\n"+revpointers;
    }

    public RoutingTable table;
    public HashMap revpointers;
    public HashMap objlevels;
  }

    private static final String class_tag = "RoutingTable";
    private static final boolean DEBUG = DebugFlags.TAPESTRY_ROUTING_TABLE;
    
    public static final int BITS_PER_DIGIT = 4;
    public static final int DIGITS_PER_GUID = 40;
    public static final int DIGIT_VALUES = 16;
    /*
    public static final int BITS_PER_DIGIT = 2;
    public static final int DIGITS_PER_GUID = 80;
    public static final int DIGIT_VALUES = 4;
    */
    public int node_status;

    //This is the patchwork statistics infomation for _channels, 
    //including link quality, latency, bandwidth, etc.
    //A NodeId is uniquely identify a same channel in both _channels 
    //and _patchwork.
    public PatchworkTable _patchwork;

    public boolean isAttached() {
      if (node_status == DynamicTClient.NOT_ATTACHED ||
	  node_status == DynamicTClient.DETACH_ROUTINGOBJ ||
	  node_status == DynamicTClient.DETACH_FINALACK)
	return true;
      else return false;
    }

  public RoutingTable (NodeId self_node_id, SecureHash guid,
			Classifier classifier) {
	Debug.turnOn (class_tag);

	_node_map = new HashMap ();
	_node_map.put(self_node_id, guid);
	_routing_table = new RouteEntry [DIGITS_PER_GUID][];
	for (int digit = 0; digit < DIGITS_PER_GUID; ++digit) {
	    _routing_table [digit] = new RouteEntry [DIGIT_VALUES];
	}
	_self_node_id = self_node_id;
	_self_digits = digits (guid);
	_nodeid_distance = new HashMap();
	_removed_nodes = new LinkedList();
	_patchwork = new PatchworkTable();
	_classifier = classifier;
    }

  public RoutingTable (NodeId self_node_id, SecureHash guid) {
    this(self_node_id, guid, null);
  }

    public int digit_values    () { return DIGIT_VALUES; } 
    public int digits_per_guid () { return DIGITS_PER_GUID; } 
    public int bits_per_digit  () { return BITS_PER_DIGIT; } 

    synchronized public SecureHash node_id_to_guid (NodeId node_id) {
	return (SecureHash) _node_map.get (node_id);
    }

    synchronized public SecureHash put_node_id_to_guid (NodeId node_id, 
							SecureHash guid) {
	return (SecureHash) _node_map.put (node_id, guid);
    }

    synchronized public Double node_to_latency(NodeId node_id) {
      return (Double) _nodeid_distance.get(node_id);
    }

    synchronized public NodeId read (int which_digit, int digit_value) {
       RouteEntry re = _routing_table [which_digit][digit_value];
       if (re !=null) return re.getFirst();
       return null;
    }

    private void dispatch (QueueElementIF item) {
      final String tag = class_tag + ".dispatch";
      if (_classifier == null) {
	if (DEBUG) Debug.printtagln(tag+".dispatch", "dispatching: "+
				    item + " type=" +
				    item.getClass().getName());
	return;
      }	
      try {
	if (DEBUG) Debug.printtagln(tag+".dispatch", "dispatching: "+
				    item + " type=" +
				    item.getClass().getName());
	_classifier.dispatch(item);
      } catch (SinkException e) {
	Carp.die ("could not dispatch "+item+".");
      }
    }

    /** 
     * Write for dummies.
     */
    synchronized public boolean write (TapestryNeighborInfo ni) {
	int [] d = digits (ni.guid);
	int digit = 0;
	for ( ; digit < DIGITS_PER_GUID; ++digit) {
	    if (d [digit] != _self_digits [digit]) 
		break;
	}
	if (digit >= DIGITS_PER_GUID) 
	    return addSelf (ni.guid);
	else
	    return write (digit, d [digit], ni.node_id, ni.guid, 
		    new Double (ni.rtt_ms));
    }

    /** 
     * Remove for dummies.
     */
    synchronized public boolean remove (TapestryNeighborInfo ni) {
	int [] d = digits (ni.guid);
	int digit = 0;
	for ( ; digit < DIGITS_PER_GUID; ++digit) {
	    if (d [digit] != _self_digits [digit]) 
		break;
	}
	RouteEntry e = _routing_table [digit][d [digit]];
	if (e == null)
	    return false;
	else {
	  TapestryNeighborInfo removed [] = new TapestryNeighborInfo[1];
	  removed[0] = ni;
	  Pair p = e.removeEntry (ni.node_id);
	  if (p == null) 
	    return false;
	  dispatch(new FRLSChangeMsg(digit, d[digit], 
				      ((Integer) p.first()).intValue(),
				      ((Integer) p.second()).intValue()));
	  return true;
	}
    }

  /**
   * Add primary/backup routes to the routing table. This function
   * takes in distance info about a new route, and compares it against
   * the current latencies in the current entry, and places it in order
   * of shortest latency first.
   * @param which_digit digit of entry being added
   * @param digit_value value of entry's Guid at the relevant digit
   * @param node_id NodeId of entry being added
   * @param latency Last tested latency of new entry
   * @param node_guid Guid of new entry
   * @return Whether the new node was actually useful (shorter) and
   *         therefore inserted into the RouteEntry.
   */
    synchronized public boolean write (int which_digit, int digit_value, 
	    NodeId node_id, SecureHash node_guid, Double latency) {
      final String tag = class_tag+".write";
      if (DEBUG) Debug.printtagln(tag, "Writing nodeid="+node_id+
				  " node_guid="+node_guid);

      // System.out.println("Writing nodeid="+node_id+" node_guid="+node_guid
      //	 +" digit="+which_digit+" value="+digit_value+" "
      // +this.toString());
      RouteEntry re = _routing_table [which_digit][digit_value];
      TapestryNeighborInfo [] added = new TapestryNeighborInfo[1];
      TapestryNeighborInfo [] removed = new TapestryNeighborInfo[1];
      removed[0] = null;
      if (re == null) {
	_routing_table[which_digit][digit_value] = 
	  new RouteEntry(node_id, latency);
	Pair p = _routing_table[which_digit][digit_value].recalcFRLS();
	_node_map.put(node_id, node_guid);
	_nodeid_distance.put(node_id, latency);
	if( p != null ) {
	    dispatch(new FRLSChangeMsg(which_digit, digit_value,
				       ((Integer) p.first()).intValue(),
				       ((Integer) p.second()).intValue()));
	}
	if (DEBUG)
	  Debug.printtagln(tag, "Inserted new RouteEntry at ["+which_digit
			   +"]["+digit_value+"] node:"+node_id+", guid:"+
			   node_guid);
      } else {
	if (DEBUG)
	  Debug.printtagln(tag, "Calling add: ["+which_digit+"]["+digit_value
			   +"], node="+node_id+", guid="+node_guid);
	// remember who's been kicked out
	NodeId removed_node[] = new NodeId[1];
	Pair p [] = new Pair[1];
	if (re.add(node_id, latency, removed_node, p)) {
	  _node_map.put(node_id, node_guid);
	  _nodeid_distance.put(node_id, latency);
	  if( removed_node[0] != null ) {
	      _removed_nodes.add(new Pair(removed_node[0], 
					  new Integer(which_digit)));
	      removed[0] = new TapestryNeighborInfo(removed_node[0],
			   (SecureHash) _node_map.get(removed_node[0]),
	      ((Double)_nodeid_distance.get(removed_node[0])).doubleValue());
	      //stop monitoring the entry that was deleted
	      dispatch(new MonitorStopMsg(removed_node[0], 
					  new Triple(which_digit, digit_value,
						     removed_node[0]), _sink));
	  }
	  if (p[0] != null)   //frls changed
	    dispatch(new FRLSChangeMsg(which_digit, digit_value,
					((Integer) p[0].first()).intValue(),
					((Integer) p[0].second()).intValue()));
	} else return false;
      }
      added[0] = new TapestryNeighborInfo(node_id, node_guid, 
					  latency.doubleValue());
      dispatch(new TapestryRoutingTableChanged(added, removed));
      return true;
    }

  /**
   * This function simply calls the more general optimize function with
   * the minlevel set to 0
   * @param newguid The guid of some node 
   * @param nid The nodeid of that node
   * @param delta is the estimated RTT distance from us to that node
   * @return Linked list of levels where node was added, or null if none
   */
  public LinkedList optimize(double delta, NodeId nid, SecureHash newguid) {
    return optimize(delta, nid, newguid, 0);
  }

  /**
   * @param newguid The guid of some node 
   * @param nid The nodeid of that node
   * @param delta is the estimated RTT distance from us to that node
   * @param minlevel is the minimum level that changes can take place
   * @return Linked list of levels where node was added, or null if none
   */
  synchronized public LinkedList optimize(double delta, NodeId nid,
					  SecureHash newguid, int minlevel) {
    int [] selfguidint = digits((SecureHash) _node_map.get(_self_node_id));
    int [] newguidint = digits(newguid);
    LinkedList changed = new LinkedList();

    int i;
    // Do this for every row that selfguidint[i] == newguidint[i], plus
    // one extra row after that where they differ
    for (i = 0; i<DIGITS_PER_GUID && 
	     (i == 0  || selfguidint[i-1] == newguidint[i-1]); 
	 i++) {

	//TODO, make optimize actually optimize, instead of just fill holes!!!
	//now at the level where newguid can be looked at.
	RouteEntry re = _routing_table[i][newguidint[i]];
	if (re != null) {
	    if (DEBUG) System.err.println("Modifying entry at digit="+
					  i+", value="+newguidint[i]);
	    if (write(i, newguidint[i], nid, newguid, new Double(delta))) {
		_nodeid_distance.put(nid, new Double(delta));
		_node_map.put(nid, newguid);
		changed.add(new Integer(i));
	    }
	} else {
	    if (write(i, newguidint[i], nid, newguid, new Double(delta))) {
		_nodeid_distance.put(nid, new Double(delta));
		_node_map.put(nid, newguid);
		if (DEBUG) 
		    System.err.println("Optimize replacing null entry with " + 
				       "nid="+ nid+", time="+delta+
				       ", digit="+i+", v="+newguidint[i]);
		changed.add(new Integer(i));
	    }
	}
    }
    if (changed.isEmpty()) return null;
    return changed;
  }

  /**
   * Add primary/backup routes to the routing table. This function
   * takes in distance info about a new route, and compares it against
   * the current latencies in the current entry, and places it in order
   * of shortest latency first.
   * @param digit digit of entry being added
   * @param value value of entry's Guid at the relevant digit
   * @param nid NodeId of entry being added
   * @param latency Last tested latency of new entry
   * @param guid Guid of new entry
   * @return Whether the new node was actually useful (shorter) and
   *         therefore inserted into the RouteEntry.
   * @deprecated As of 7/24/2002 in favor of RoutingTable.write() 
   */
  synchronized public boolean addBackup(int digit, int value, NodeId nid,
					Double latency, SecureHash guid) {

    // make sure to save the node that could get kicked out in this
    // call
    NodeId removed_node[] = new NodeId[1];
    Pair p[] = new Pair [1];
    if (_routing_table[digit][value].add(nid, latency, removed_node, p)) {
      _nodeid_distance.put(nid, latency);
      _node_map.put(nid, guid);
      if( removed_node[0] != null ) {
	  _removed_nodes.add(new Pair(removed_node[0], new Integer(digit)));
	  //stop monitoring deleted message
	  dispatch(new MonitorStopMsg(removed_node[0], 
				      new Triple(digit, value, removed_node[0]),
				      _sink));
      }
      if (p[0] != null)   //frls changed
	dispatch(new FRLSChangeMsg(digit, value,
				   ((Integer) p[0].first()).intValue(),
				   ((Integer) p[0].second()).intValue()));
      return true;
    } else {
      if (p[0] != null)   //frls changed
	dispatch(new FRLSChangeMsg(digit, value,
				   ((Integer) p[0].first()).intValue(),
				   ((Integer) p[0].second()).intValue()));
      return false;
    }
  }

  synchronized public RouteEntry getEntry(int digit, int value) {
    return _routing_table[digit][value];
  }

  synchronized public void putEntry(int digit, int value,
				    RouteEntry re) {
    _routing_table[digit][value] = re;
  }


  /**
   * Method to add the latency measurement to the data structures
   * @param rtt Estimated roundtrip distance
   * @param nid NodeId of new node
   * @param newguid Guid of new node
   */
  synchronized public void addDistance(double rtt, NodeId nid,
				       SecureHash newguid) {
    if (DEBUG) 
      System.err.println("adding distance for nodeId="+nid);
    _nodeid_distance.put(nid, new Double(rtt));
    if (_node_map.get(nid) == null) _node_map.put(nid, newguid);
  }

    /**
     * Route to the root of the tree for <code>dest_digits</code>, assuming
     * that the first <code>digit [0] - 1</code> digits have already been
     * accounted for, and return the next node in the routing path.  
     * 
     * @param dest_digits the digits of the destination GUID
     * @param exact       whether this should be an exact match; if this
     *                    parameter is true, the function will return
     *                    <code>null</code> if there is no such node.
     * @param digit       the digit to start routing from
     *
     * @return            the next node in the routing path, if any; return
     *                    <code>null</code> if there is no such node and
     *                    return <code>_self_node_id</code> if the current
     *                    node is the deterministic root
     */
    public NodeId route_to_root (int [] dest_digits, boolean exact, 
	    boolean prefix, int [] digit) {

	RouteEntry entry = _route_to_root(dest_digits, exact, prefix, digit);
	return (entry==null?null:entry.getFRLS());
    }

    public RouteEntry _route_to_root (int [] dest_digits, boolean exact, 
					     boolean prefix, int [] digit) {

        String tag = null;
	if(DEBUG) tag = "RoutingTable.[" + _self_node_id.port () + "]";
	if (DEBUG) {
	    String msg = "called.  Our guid: "; 
	    for (int i = 0; i < dest_digits.length; ++i) {
		msg += Integer.toHexString (_self_digits [i]);
	    }
	    msg += ".  Looking for: ";
	    for (int i = 0; i < dest_digits.length; ++i) {
		msg += Integer.toHexString (dest_digits [i]);
	    }
	    msg += (prefix ? "prefix." : (exact ? ", exact." : ", fuzzy."));
	    Debug.printtagln (tag, msg);
	}
	
	// Start with the requested digit.
	for ( ; digit [0] < digits_per_guid (); ++digit [0]) {
	    
	    int val = dest_digits [digit [0]];

	    //	    if (DEBUG) Debug.printtagln (tag, "digit " + digit [0] + 
	    //		    " looking for " + Integer.toHexString (val) + ".");

	    RouteEntry entry = getEntry (digit [0], val);

	    // If there is no routing entry at the given digit (not even
	    // ourselves) find the next best one by adding circularly to
	    // the desired value.  This is then the ``surrogate'' root.

	    //ling huang: done for frls here. need do something for multicast.
	    if (entry == null || entry.getFRLS() == null) {
		if (exact) {
		    if (DEBUG) Debug.printtagln (tag, 
			    "no one matches; failing.");
		    return null;
		}

		if (DEBUG) Debug.printtagln (tag, 
			"no one matches; approximating");
			
		entry = _surrogate (dest_digits, val, digit [0]);
		if (entry == null) {
		    // finish early
		    RouteEntry re = new RouteEntry(_self_node_id, 
						   new Double(0));  
		    re.recalcFRLS();
		    return re;
		}
	    }

	    // Now entry != null.

	    if ( entry.getFRLS() != null && 
		 ! entry.getFRLS().equals (_self_node_id)) {

		// If the surrogate root is not us, we need to forward on
		// to whoever it is.

		if (DEBUG) Debug.printtagln (tag, "Node " + entry +
			", guid " + 
			(entry==null?null:node_id_to_guid(entry.getFRLS())) +
			" matches.");

		return entry;
	    }

	    // If the (surrogate) root is us, continue to the next digit.
	}

	// If we are the surrogate root for the last digit, we are
	// deterministically the root for all nodes.

	if (DEBUG) Debug.printtagln (tag, "we match.");
	if(digit[0] == digits_per_guid())
	    digit[0] -= 1;
	RouteEntry re = new RouteEntry(_self_node_id, new Double(0));  
	re.recalcFRLS();
	return re;
    }

    /**
     * Find the surrogate for an empty routing table entry.
     *
     * @param dest_digits The array of ints representing the destination guid
     * @param dest_digit  The digit value that points to an empty entry
     * @param hop         Which hop are we looking for the surrogate route on
     *
     * @return            <code>null</code> if we are <i>definitely</i> the
     *                    surrogate, and the <code>NodeId</code> of the
     *                    next person to ask if we cannot tell from this
     *                    hop yet.
     **/
    synchronized public NodeId surrogate (int [] dest_digits, 
	    int dest_digit, int hop) {
	final String tag = class_tag + ".surrogate";

	RouteEntry entry = _surrogate(dest_digits, dest_digit, hop);
	return (entry==null?null:entry.getFRLS());
    }

    synchronized public RouteEntry _surrogate (int [] dest_digits, 
	    int dest_digit, int hop) {
	final String tag = class_tag + "._surrogate";

	boolean empty = true;
	if (DEBUG) System.err.println("Surrogate called on table: "+
				      toString());
	for (int next_hop = hop; next_hop < DIGITS_PER_GUID; ++next_hop) {
	  if (DEBUG) Debug.printtagln (tag, "checking level " +
				       next_hop + ".");
	  for (int i = 0; i < DIGIT_VALUES; ++i) {
	    if ((_routing_table [next_hop][i] != null) && 
		((RouteEntry) (_routing_table [next_hop][i])).getFRLS() != null &&
		!((RouteEntry) (_routing_table [next_hop][i])).getFRLS().equals(_self_node_id)) {
	      if (DEBUG) Debug.printtagln (tag, "neighbor " +
			 _routing_table [next_hop][i] + " present.");
	      empty = false;
	      break;
	    }
	  }
	  if (! empty) break;
	}

	if (empty) {
	    if (DEBUG) Debug.printtagln (tag, "we are definitely the root.");
	    return null;
	}

	// Surrogate routing to next higher non-null value (circularly).
	int i = (dest_digit == DIGIT_VALUES - 1) ? 0 : dest_digit;

	while (_routing_table [hop][i] == null  ||
	       _routing_table [hop][i].getFRLS() == null)
	    i = (i == DIGIT_VALUES - 1) ? 0 : i + 1;

	if (DEBUG) Debug.printtagln (tag, "surrogate is " + 
		_routing_table [hop][i].getFRLS());

	return _routing_table [hop][i];
    }

  /**
   * method that calculates the NEXT best surrogate hop other than the current
   * node for ANY key/GUID.  Assumes we're currently the best surrogate, then
   * looks for the next surrogate from the right most digit to the left. The
   * first non-null value in the table found would be the next hop to the
   * next surrogate.  We also make sure that we take in a HashSet, and ignore
   * those nodes when searching for the next surrogate. This is to avoid 
   * binary loops, where A and B think the other is the next best surrogate
   * for an object. For example: A=1234, B=1235 for GUID=1233, where no other
   * nodes match prefix of 123X.
   * @return NodeId of the next best surrogate hop if we were not around
   **/
  synchronized public NodeId nextSurrogate(HashSet ignoredNodes, 
					   SecureHash selfguid) {
    final String tag = class_tag + ".surrogate";
    int [] self_digits = RoutingTable.digits(selfguid);

    for (int hop = DIGITS_PER_GUID - 1; hop >= 0; hop--) {

      int i = (self_digits[hop] == DIGIT_VALUES -1) ? 0 : self_digits[hop]+1;
      while (_routing_table[hop][i] == null ||
	     _routing_table[hop][i].getFirst() == null ||
	     (ignoredNodes != null &&
	      ignoredNodes.contains(_routing_table[hop][i].getFirst()) ))
	i = (i == DIGIT_VALUES - 1) ? 0 : i + 1;
      //note: we can safely ignore any branch off the route path that 
      //has a node that is on the ignorelist. Reason is that that branch
      //of the namespace must have been fully explored for us to have 
      //backtracked to this previous digit.

      if (i != self_digits[hop]) //found a real next surrogate hop
	return _routing_table[hop][i].getFirst();
    }

    return null;  //this should never happen. we should ALWAYS find somebody
  }

  public boolean addSelf (SecureHash self) {
    if (DEBUG) System.err.println("Writing entries for self into map.\n");

    boolean changed = false;
    int [] digits = RoutingTable.digits(self);
    for (int i=0; i<digits_per_guid(); i++) {
      changed = write(i, digits[i], _self_node_id, self, new Double(0)) ||
	  changed;
    }
    return changed;
  }

  public static int [] digits (SecureHash h) {
      final String tag = class_tag + ".digits";
      byte [] buf = h.bytes ();
      
      // assume that BITS_PER_DIGIT = 2^n for some positive
      // integer n <= 3
      if( BITS_PER_DIGIT != 4 && BITS_PER_DIGIT != 2 && 
 	  BITS_PER_DIGIT != 1 && BITS_PER_DIGIT != 8 ) {
 	  throw new AssertionViolatedException(tag + ": Ack!");
      }

      int result[];
      int i;
      int digits_per_byte = 8/BITS_PER_DIGIT;
      if (buf.length * digits_per_byte != DIGITS_PER_GUID)
	  throw new AssertionViolatedException(tag + ": Ack!");

      result = new int [DIGITS_PER_GUID];
      i = result.length - 1;
      while (i >= 0) {
	  for( int j = i; j > (i-digits_per_byte); j-- ) {
	      result [j] = ((int) (buf [i/digits_per_byte] >>> 
				   (i-j)*BITS_PER_DIGIT) & 
			    (DIGIT_VALUES-1));
	  }
	  i -= digits_per_byte;
      }
      
      return result;
  }

  public String toString() {
    String val ="RoutingTable for "+_self_node_id+"\n";
    for (int i=0; i<5; i++) {
      for (int j=0; j<DIGIT_VALUES; j++) {
	  if ((_routing_table[i][j] == null) || 
	      (_routing_table[i][j].getFirst() == null) ) {
	  val+=" -";
	} else {
	  val+=" -["+_routing_table[i][j].size()+"]"+
	    printGuid((SecureHash) 
		      _node_map.get(_routing_table[i][j].getFirst()));
	  val += "{";
	  for(int k = 1; k < RouteEntry.REDUNDANT_LVL; k++)
	  {
	      if(_routing_table[i][j].getAtIndex(k)!=null)
	      {
		  if(k>1)
		      val += ", ";
		  val+= printGuid((SecureHash) 
		      _node_map.get(_routing_table[i][j].getAtIndex(k)));
	      }
	  }	  
	  val += "}";
	}
      }
      val+="\n";
    }
    return val;
  }
    
  public String toShortString() {
      StringBuffer sbuf =
	  new StringBuffer("RoutingTable for "+_self_node_id+"\n");
      // First print out the whole table with indices instead of hashes
      // Makes for a more readable, strictly-aligned table.
      final int spaces_per_entry = 10;
      int next_peer_index = 0;
      HashMap peer_indices = new HashMap();
      LinkedList peers = new LinkedList();
      for (int i=0; i<5; i++) {
	  int line_start_index = sbuf.length();
	  for (int j=0; j<DIGIT_VALUES; j++) {
	      // Start at a rigid offset, even if previous entry was too long.
	      int entry_start_index = ( line_start_index +
					( j * spaces_per_entry ) );
	      sbuf.setLength( entry_start_index );
	      // Now print the entry:
	      if ((_routing_table[i][j] == null) || 
		  (_routing_table[i][j].getFirst() == null) ) {
		  sbuf.append("  -  ");
	      } else {
		  NodeId peer;

		  // Make sure hidden backups are listed in footer:
		  for(int k = 1; k < RouteEntry.REDUNDANT_LVL; k++) {
		      NodeId node_id =
			  (NodeId)_routing_table[i][j].getAtIndex(k);
		      if( node_id != null ) {
			  peer = node_id;
			  if( ! peer_indices.containsKey( peer ) ) {
			      peer_indices.put(peer,
					       new Integer(next_peer_index++));
			      peers.add( peer );
			  }
		      }
		  }
		  peer = _routing_table[i][j].getFirst();

		  if( ! peer_indices.containsKey( peer ) ) {
		      peer_indices.put( peer, new Integer(next_peer_index++) );
		      peers.add( peer );
		  }
		  sbuf.append(" ");
		  sbuf.append(_routing_table[i][j].size()); // hopefully 1 digit.
		  sbuf.append(":");
		  sbuf.append( peer_indices.get( peer ) ); // up to 3 // digits.
		  sbuf.append("-");
		  peer = _routing_table[i][j].getAtIndex(1);
		  if (peer != null) {
		    if (!peer_indices.containsKey(peer))
		      sbuf.append("*");
		    else sbuf.append( peer_indices.get( peer ) );
		  } else sbuf.append("N");
		  sbuf.append("-");
		  peer = _routing_table[i][j].getAtIndex(2);
		  if (peer != null) {
		    if (!peer_indices.containsKey(peer))
		      sbuf.append("*");
		    else sbuf.append( peer_indices.get( peer ) );
		  } else sbuf.append("N");

		  sbuf.append( "  " );
	      }
	  }
	  sbuf.append("\n");
      }
      // Now print out the hashes themselves.
      Iterator peer_iter = peers.iterator();
      while( peer_iter.hasNext() ) {
	  NodeId peer = (NodeId)peer_iter.next();
	  sbuf.append( peer_indices.get( peer ) );
	  sbuf.append(":");
	  sbuf.append( peer );
	  sbuf.append("\n");
      }
      return sbuf.toString();
  }
    
  public static String printGuid(SecureHash x) {
    String out = "";
    if (x ==null) return out;
    int [] digits = digits(x);
    for (int i=0; i<digits.length; i++) {
      out += Integer.toHexString(digits[i]);
    }
    return out+"/"+x;
  }

  public RoutingTable(InputBuffer buffer) throws ostore.util.QSException {
     NodeId nullHash = new NodeId();
     int nodeMapSize = 0;
     int routingTableCol = 0;
     int routingTableRow = 0;
     int nodeDistance = 0;

     _self_node_id = (NodeId) buffer.nextObject(); 

     nodeMapSize = buffer.nextInt();
     
     _node_map = new HashMap();
     for(int i=0; i<nodeMapSize; i++) {
	NodeId key = (NodeId) buffer.nextObject();
	SecureHash element = (SHA1Hash) buffer.nextObject(); 
	_node_map.put(key, element);
     }
     _self_digits = digits (node_id_to_guid(_self_node_id));

     routingTableRow = buffer.nextInt(); 
     routingTableCol = buffer.nextInt(); 

     int NULL_ENTRY = 0;
     int HAS_ENTRY = 1;

     _routing_table = new RouteEntry[routingTableRow][routingTableCol];
     for(int i=0; i< routingTableRow; i++) {
	for(int j=0; j< routingTableCol; j++) {
	   if(buffer.nextInt() != NULL_ENTRY) { 
	      _routing_table[i][j] = (RouteEntry) buffer.nextObject(); 
	   }
	}
     }

     nodeDistance = buffer.nextInt();
     _nodeid_distance = new HashMap();
     for(int i=0; i<nodeDistance; i++) {
	NodeId key = (NodeId) buffer.nextObject(); 
	Double dist = new Double(buffer.nextDouble());
  	_nodeid_distance.put(key, dist);
     }

  }

  public void serialize (OutputBuffer buffer) {
     NodeId nullNode = new NodeId();
     buffer.add(_self_node_id);

     int NULL_ENTRY = 0;
     int HAS_ENTRY = 1;

     Set i = _node_map.keySet();
     Iterator iter = i.iterator();
     buffer.add(_node_map.size());
     while(iter.hasNext()) {
	NodeId key = (NodeId) iter.next();
	SecureHash element = (SecureHash) _node_map.get(key);
	
	buffer.add(key);
	buffer.add(element);
     }

     buffer.add(_routing_table.length);
     buffer.add(_routing_table[0].length); 

     for(int k=0; k < _routing_table.length; k++) {
	for(int j=0; j < _routing_table[0].length; j++) {
	   if(_routing_table[k][j] != null) {
	      buffer.add(HAS_ENTRY);
	      buffer.add(_routing_table[k][j]); 
	   } else {
	      buffer.add(NULL_ENTRY);
	   }
	}
     }

     Set j = _nodeid_distance.keySet();
     Iterator jter = j.iterator();
     buffer.add(_nodeid_distance.size()); 
     while(jter.hasNext()) {
	NodeId key = (NodeId) jter.next();
	Double element = (Double) _nodeid_distance.get(key);
	buffer.add(key); 
	buffer.add(element.doubleValue()); 
     }
  }

  public Map get_node_map() {
      return _node_map;
  }

  public RouteEntry[][] get_table() {
     return _routing_table;
  }

  public NodeId get_self_node_id() {
     return _self_node_id;
  }

  public LinkedList get_removed_nodes() {
     return _removed_nodes;
  }

  /**
   ** Update the Patchwork statistics to RoutingTable, 
   ** given the newest delta ent out by Patchwork model.
   ** At current time, we only update loss_rate to RoutingTable.
   ** @return LinkedList of messages to be sent by whoever calls this
   **/
  public void pw_update( DeltaMsg pdrm ) {
    final String tag = class_tag + ".pw_update()";

    _patchwork.add_delta( pdrm.delta );
    
    //Ling Huang for fault tolerance routing:
    //We should do a lot of smart work here, including:
    //1) update statistics according to recent information
    //2) make a prediction
    //3) make a decision
    //4) ......
    
    //Update link quality from patchwork to routing table.
    updateRouteQuality();

    //update the frls cache in each entry
    if (Router.FRLS)
      updateFRLSRoute();
  }
  
  private void updateRouteQuality() {
    PatchworkStats stats;
    RouteEntry entry;
    RouteEntry.NodeEntry [] routes;
    NodeId     self_node_id;
    double     current_latency, current_loss;
    int digit, value;
    int i, j, k;
    
    final String tag = class_tag + ".updateRouteQuality()";
    
    for (i = 0; i < DIGITS_PER_GUID; i++) {
      for (j = 0; j < DIGIT_VALUES; j++) {
	entry = _routing_table[i][j];
	if (entry == null)
	  continue;
	routes = entry.getRoutes();
	if (routes == null)
	  continue;
	for (k = 0; k < entry.REDUNDANT_LVL; k++) {
	  if (routes[k] == null)
	    continue;
	  if (_self_node_id.equals(routes[k].address)) {
	    //looking at our own entry in our table, should set to no loss
	    routes[k].latency = 0;
	    routes[k].loss_rate = 0;
	    continue;
	  }
	  //You can get both long term link quality and 
	  //current period quality.
	  stats = _patchwork.getStats(routes[k].address, 
				      Patch.TYPE_LATENCY);
	  if (stats != null) {
	    routes[k].latency = stats.getOverallMean();
	    current_latency   = stats.getCurrentMean();
	  } else
	    current_latency = 0.0;
	  
	  stats = _patchwork.getStats(routes[k].address, 
				      Patch.TYPE_LOSS);
	  if (stats != null) {
	    routes[k].loss_rate = stats.getOverallMean();
	    current_loss        = stats.getCurrentMean();
	  } else
	    current_loss = 0.0;
	  
	  //If loss rate in current period is greater than 60%, 
	  //assume the link is dead in this period.
	  if (current_loss > Router.THRESHOLD_DEAD)
	    routes[k].dead_period ++;
	  else
	    routes[k].dead_period = 0;
	  
	  // If a route dies for several period, 
	  //we should find a replacement for it.
	  if (routes[k].dead_period > Router.THRESHOLD_REPLACE)
	    triggerRepair(entry, k);
	}
      }
    }
  }

  private void triggerRepair (RouteEntry entry, int k) {
    /*
      NodeId repair;
      int frls = entry.get_frls();
      if (k == frls) {
        k = frls + 1;
        if (k >= entry.REDUNDANT_LVL)
          k = 0;
        repair = routes[k].address;
      } else
        repair = routes[frls].address;
      msgs.add(new DynamicRepairReqMsg(repair, _router.get_self_guid(), 
                                     i, j, k));
    */
  }

  /**
   ** We should set a threshold in this function when switch route entry.
   ** Trade off: fast switch vs. routes oscillation.
   **/
  private void updateFRLSRoute() {
    RouteEntry entry;
    int i, j;
    boolean changed;

    for (i = 0; i < DIGITS_PER_GUID; i++) {
      for (j = 0; j < DIGIT_VALUES; j++) {
	entry = _routing_table[i][j];
	if (entry == null)
	  continue;
	Pair p = entry.recalcFRLS();
	if (p != null)
	  dispatch(new FRLSChangeMsg(i, j, ((Integer) p.first()).intValue(),
				     ((Integer) p.second()).intValue()));
      }
    }
  }

  protected void setSink( SinkIF sink ) {
    _sink = sink;
  }

  private Classifier _classifier;    //classifier passed in to send msgs
  private SinkIF _sink = null; 
  private Map _node_map;        //NodeId to Guid mapping
  private RouteEntry [] [] _routing_table;
  private NodeId _self_node_id;
  private int [] _self_digits;
  private Map _nodeid_distance; //mapping nodeids to last latency measured
                                //but only for entries in the routingtable

  private LinkedList _removed_nodes; // all of the nodes that got kicked out
                                     // of the routing table.  Note that
                                     // this can (and should) be accessed
                                     // outside of this class, for keeping
                                     // revpointers consistent.
}

