/**
 * WatchList.java
 *
 * The class which encapsulates the watch list.  The watch list is a bitmap
 * passed as part of the acknowledged multicast messages from a new node's
 * surrogate node to all "need-to-know" nodes.  It includes only the
 * "interesting" levels, which are the levels of the routing table not filled
 * by the initial routing table from the surrogate node.  Each bit indicates
 * whether the new node still needs to fill a certain hole in its routing
 * table. A true indicates the position is filled, and a false indicates it's
 * still open.
 *
 * @author     Ben Y. Zhao
 * @version    $Id: WatchList.java,v 1.10 2003/04/26 00:27:45 strib Exp $
 */

package ostore.tapestry.impl;
import ostore.util.NodeId;
import ostore.util.SecureHash;
import java.util.BitSet;
import java.util.StringTokenizer;
import java.util.TreeSet;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;

public class WatchList 
implements QuickSerializable {

  private SecureHash guid;
  private int startlevel;
  private BitSet [] bits;
  public static final int bitSetLength = RoutingTable.DIGIT_VALUES;

  public WatchList(SecureHash guid, int start, int lvls) {
    this.guid = guid;
    this.startlevel = start;
    this.bits = new BitSet[start+lvls];
    for (int i=0; i<bits.length; i++) {
      this.bits[i] = new BitSet(bitSetLength);
    }
  }

  public WatchList (InputBuffer buffer) throws ostore.util.QSException {
    guid = (SecureHash) buffer.nextObject();
    startlevel = buffer.nextInt();
    int size = buffer.nextInt();
    bits = new BitSet [size];

    BitSet bs;
    String s, t;
    StringTokenizer st;
    for (int i=startlevel; i<size; i++) {
      bs = new BitSet(bitSetLength);
      s = buffer.nextString();
      if (ostore.util.DebugFlags.TAPESTRY_DYNAMIC_TCLIENT)
	System.err.println("WatchList(): Watchlist parsing string: "+s);
      if (s.length() > 2) {
	s = s.substring(1, s.length()-1);
	st = new StringTokenizer(s, ", ");

	while (st.hasMoreTokens()) {
	  t = st.nextToken();
	  bs.set(Integer.parseInt(t));
	}
      }
      if (ostore.util.DebugFlags.TAPESTRY_DYNAMIC_TCLIENT)
	System.err.println("WatchList(): Watchlist parsed result: "+bs);
      bits[i] = bs;
    }
  }

  public void serialize (OutputBuffer buffer) {
    buffer.add(guid);
    buffer.add(startlevel);
    buffer.add(bits.length);
    for (int i=startlevel; i<bits.length; i++) {
      //System.err.println("Watchlist output bitset is: "+bits[i].toString());
      buffer.add(bits[i].toString());
    }
  }

  public SecureHash getGuid() {
    return guid;
  }

  public int startLevel() {
    return startlevel;
  }

  public int levels() {
    return (bits.length-startlevel);
  }

  /**
   ** HashSet checkAndSend
   **
   ** This function is called as part of the acknowledged multicast
   ** handler. It checks to see if the routing table of the current node
   ** contains any entries that can fill holes in the current watchlist.  The
   ** function returns a linkedlist of guids which can fill holes, or null if
   ** there is none.  The actual bits are turned ON in the watchlist bitmap
   ** as a side-effect.  So when this same watchlist object is passed on, it
   ** has been updated to reflect the presence of this current node.
   ** <BR>
   **
   ** @param newGuid
   ** @param table
   ** @return HashSet of nodes that need to be notified.
   **/
  public TreeSet checkAndSend (SecureHash newGuid,
			       RoutingTable table) {
    BitSet bs;
    int [] digits = RoutingTable.digits(newGuid);
    TreeSet nodes = new TreeSet();
    for (int i=startlevel; i<bits.length; i++) {
      bs = bits[i];
      if (bs == null) {
	if (ostore.util.DebugFlags.TAPESTRY_DYNAMIC_TCLIENT) {
	  System.err.println("WatchList.checkAndSend: old startlevel="
			     +startlevel+", new startlevel="+(i+1));
	}
	startlevel = i+1;
	continue;
      }
      for (int j=0; j<RoutingTable.DIGIT_VALUES; j++) {
	if (!bs.get(j) && (table.read(i, j) != null)) {
	  bs.set(j);
	  nodes.add(table.read(i,j));
	}
      }
    }
    return nodes;
  }

  /**
   ** void prune
   ** 
   ** Function that checks the lower levels of the watchlist, and removes
   ** levels which have all been filled in, in order to minimize the
   ** size of the watchlist sent along with the multicast
   **
   **/
  public void prune() {
    for (int i=startlevel; i<bits.length; i++) {
      BitSet bs = bits[i];
      if (bs == null) {
	startlevel = i+1;
	if (ostore.util.DebugFlags.TAPESTRY_DYNAMIC_TCLIENT) {
	  System.err.println("WatchList.checkAndSend: incrementing startlevel="
			     +startlevel);
	}
	continue;
      }
      boolean allset = true;
      for (int j=0; j<RoutingTable.DIGIT_VALUES; j++) {
	allset = allset & bs.get(j);
      }
      if (allset) {
	//now remove this level
	bits[i] = null;
	startlevel ++;
      } else {
	break;
      }
    } //end for (int i=startlevel...
  }

}
