/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;
import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.RandomAccessFile;

/**
 * The <code>QSIO</code> class contains general utility methods for
 * writing/reading <code>QuickSerializable</code> object to/from files.
 *
 * @author	Dennis Geels
 * @author      Larry Tung
 * @version	$Id: QSIO.java,v 1.36 2002/09/08 00:20:24 hweather Exp $
 */
public abstract class QSIO
{
  //////////////////// 	CONSTANTS	////////////////////
  private static boolean DEBUG = ostore.util.DebugFlags.QSIO;

  //////////////////// 	VARIABLES	////////////////////
  ////////////////////    STATIC VARIABLES 	////////////////////
  //////////////////// 	 METHODS	////////////////////

  /**
   * Write a <code>QuickSerializable</code> object to the specified
   * <code>OutputStream</code>.
   * An 8B header is written first, containing the size of the object in
   * bytes, followed by its type code.
   *
   * @param object	The <code>QuickSerializable</code> object to write.
   * @param out	The <code>OutputStream</code> into which to write.
   * @exception IOException if an I/O error occurs.
   */
  public static void write( QuickSerializable object, OutputStream out )
      throws IOException
  {
    // TODO: get rid of count buffer?
    CountBuffer cb = new CountBuffer ();
    cb.add ((int) 0);
    cb.add (object);
    
    StreamOutputBuffer sob = new StreamOutputBuffer (out);
    sob.add (cb.size ());
    sob.add (object);
    if(DEBUG) Debug.printtagln("QSIO.write",
	    "just wrote " + (cb.size () + 4) + " bytes to disk");
  }

  public static byte[] qsToBytes( QuickSerializable object )
  {
    CountBuffer cb = new CountBuffer ();
    cb.add ((int) 0);
    cb.add (object);
    byte [] result = new byte [cb.size ()];
    ByteArrayOutputBuffer outb = new ByteArrayOutputBuffer (result);
    outb.add (cb.size ());
    outb.add (object);
    return result;
  }

  /**
   * The inverse of qsToBytes, above.
   */
  public static QuickSerializable bytesToQS ( byte [] bytes) 
      throws QSException
  {
      ByteArrayInputBuffer buffer = new ByteArrayInputBuffer (bytes);
      buffer.nextInt (); // ignore size
      return buffer.nextObject();
  }

  /**
   * Write a <code>QuickSerializable</code> to the named file
   * The file will be overwritten unless an <code>IOException</code> is
   * thrown, in which case the file will remain unmodified.
   *
   * @param object	The <code>QuickSerializable</code> object to write.
   * @param filename	The name of the file into which to write.
   * @exception IOException if an I/O error occurs.
   */ 
  public static void write( QuickSerializable object, String filename )
    throws IOException
  {
      // First write to a temporary file.
      String tmp_name = filename+".tmp";
      write( object, tmp_name, false );
      
      // Write succeeded: move temp file into proper location.
      try {
	  if( ! (new File(tmp_name)).renameTo(new File(filename)) ) {
	      Carp.die( "Could not rename " + tmp_name + " to " + filename);
	  }	      
      }
      catch( java.lang.Exception e ) {
	System.err.println("exception occurred in renaming tmp" + e);
      }
  }

  /**
   * Write a <code>QuickSerializable</code> to the named file
   * This method provides no guarantees against file corruption if an
   * <code>IOException</code> is thrown during the write.
   *
   * @param object	The <code>QuickSerializable</code> object to write.
   * @param filename	The name of the file into which to write.
   * @param append      Whether to write the object to the end of the file
   * @exception IOException if an I/O error occurs.  
   */ 
  public static void write( QuickSerializable object, String filename, 
			    boolean append ) throws IOException 
  {
    try {
      FileOutputStream file = null;
      if (append) { // TODO was FileOutputStream(filename, append)
	file = new FileOutputStream( filename , true );
      }
      else {
	file = new FileOutputStream( filename );
      }
      write( object, file );
      file.close();
    } catch( IOException ioe ) {
      if( DEBUG ) {
	  Debug.printtagln("QSIO.write",  "Could not write <" +
			   object + "> to file \"" + filename + "\"");
      }
      throw ioe;
    }
  }
    
  public static void write( QuickSerializable object, RandomAccessFile file, 
			    long pos ) throws IOException
  {
    try {
	file.seek(pos);
	RandomAccessFileOutputBuffer buf = 
	    new RandomAccessFileOutputBuffer (file);

	// TODO: get rid of count buffer?
	CountBuffer cb = new CountBuffer ();
	cb.add ((int) 0);
	cb.add (object);

	buf.add (cb.size ());
	buf.add (object);

	//TODO - is there a flush to RandomAccessFiles?
    }
    catch( IOException ioe ) {
      if (DEBUG) {
	  Debug.printtagln("QSIO.write",
			   "Could not write to RandomAccessFile at pos: " +
			   pos + "; seek or write failed");
      }
      throw ioe;
    }
  }

   /**
   * Read in the <code>QuickSerializable</code> stored in the specified
   * <code>InputStream</code>.
   *
   * @param in	The <code>InputStream</code> from which to read.
   * @return the <code>QuickSerializable</code> read.
   * @exception IOException if an I/O error occurs.
   * @exception QSException if <code>in</code> did not contain a valid
   * <code>QuickSerializable</code> object.
   */
  public static QuickSerializable read( InputStream in )
      throws IOException, QSException
  {
    StreamInputBuffer sib = new StreamInputBuffer (in);
    sib.nextInt (); // ignore the length
    return sib.nextObject ();
  }

  /**
   * Read in the <code>QuickSerializable</code> stored in the specified
   * <code>RandomAccessFile</code>.
   *
   * @param file	The <code>RandomAccessFile</code> from which to read.
   * @param pos 	amount to seek to
   * @return the <code>QuickSerializable</code> read.
   * @exception IOException if an I/O error occurs.
   * @exception QSException if <code>in</code> did not contain a valid
   * <code>QuickSerializable</code> object.
   */
  public static QuickSerializable read( RandomAccessFile file , long pos )
    throws IOException, QSException
  {
      file.seek(pos);
      RandomAccessFileInputBuffer buf = 
	  new RandomAccessFileInputBuffer (file);
      buf.nextInt (); // ignore the length
      QuickSerializable result = buf.nextObject ();
      return result;
  }

  /** 
   * delete the contents of a QuickSerializable by filling where 
   * the object was with 0's
   * @param pos 
   *      amount to seek to 
   */
  public static void delete( RandomAccessFile file, long pos ) 
    throws IOException
  {
    byte[] data = new byte[4];
    int[] offset = { 0 };
    try {
      file.seek(pos);
      file.read(data);
      int len = ByteUtils.bytesToInt( data, offset );
      System.out.println("seeked: " + pos + " len: " + len);
      if(file.getFilePointer()!= (pos+4) ) {
	  Carp.die( "Moving file pointer" );
      }
      // 0 out the type code and object
      byte[] zeros = new byte[len+4];
      for (int i = 0; i < len+4; ++i) {
	zeros[i] = 0;
      }
      file.write(zeros);
      // (pos+4) == where typecode begins (typecode is 4 bytes long as well)
      if(file.getFilePointer() != (pos+8+len) ) {
	  Carp.die( "delete: write finished at pos " + file.getFilePointer()
		    + " pos: " + pos + " length: " + len);
      }
    }
    catch( IOException ioe ) {
      if (DEBUG) {
	Debug.printtagln("QSIO.delete", "Could not delete object at pos: " + pos);
      }
      throw ioe;
    }
  }
 
  /**
   * Read in the <code>QuickSerializable</code> stored in the named file.
   *
   * @param in	The name of the file from which to read.
   * @return the <code>QuickSerializable</code> read.
   * @exception IOException if an I/O error occurs.
   * @exception QSException if <code>in</code> did not contain a valid
   * <code>QuickSerializable</code> object.
   */
  public static QuickSerializable read( String filename )
      throws IOException, QSException
  {
    QuickSerializable object = null;
    try { 
      FileInputStream file = new FileInputStream( filename );
      object = read( file );
      file.close();
    } catch( IOException ioe ) {
	if( DEBUG ) {
	    Debug.printtagln( "QSIO.read", "Could not read from file \"" +
			      filename + "\"");
	}	    
      throw ioe;
    }
    catch( QSException qse ) { //added for debugging
      if (DEBUG) {
	System.out.println("bad filename?" + filename);
      }
      throw qse;
    }
    return object;
  } 




  ////////////////////     STATIC METHODS 	////////////////////
  ////////////////////      CONSTRUCTORS	////////////////////
}














