/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

/**
***  The <code>StatRecorder</code> class is a utility for recording
***  and analyzing a number of samples for a measurement.

***  @author  Patrick R. Eaton
***  @version $Id: StatRecorder.java,v 1.7 2003/01/24 21:01:39 strib Exp $
**/
public class StatRecorder
    implements ostore.util.QuickSerializable
{
    private String _name;
    private String _tag;
    private Map    _samples;

    //private int skip = 0;

    private int  _num_samples;
    private long _sum_of_samples;
    private long _prod_of_samples;
    private long _square_sum_of_samples;

    private long _max_sample = Long.MIN_VALUE;
    private long _min_sample = Long.MAX_VALUE;

    public StatRecorder(String name)
    {
        this(name, /*tag=*/ "");
        return;
    }

    public StatRecorder(String name, String tag)
    {
        _name = name;
        _tag  = tag;
        reset();

        return;
    }

    public StatRecorder(InputBuffer buffer)
    {
        this(/*name=*/ buffer.nextString(), /*tag=*/ buffer.nextString());

        int num_sample_pairs = buffer.nextInt();
        for (int i = 0; i < num_sample_pairs; ++i)
        {
            long sample = buffer.nextLong();
            long count  = buffer.nextLong();
            add(sample, count);
        }

        return;
    }

    public synchronized void serialize(OutputBuffer buffer)
    {
        buffer.add(_name);
        buffer.add(_tag);

        buffer.add(_samples.size());
        Iterator iter = _samples.keySet().iterator();
        while (iter.hasNext())
        {
            Long sample_l = (Long) iter.next();
            long sample   = sample_l.longValue();
            Long count_l  = (Long) _samples.get((Object) sample_l);
            long count    = count_l.longValue();

            buffer.add(sample);
            buffer.add(count);
        }

        return;
    }

    public synchronized void reset()
    {
        _samples = new TreeMap();

        _num_samples = 0;
        _sum_of_samples = 0;
        _prod_of_samples = 1;
        _square_sum_of_samples = 0;

        _max_sample = Long.MIN_VALUE;
        _min_sample = Long.MAX_VALUE;

        return;
    }

    public synchronized void add(StatRecorder other)
    {
        Iterator iter = other._samples.keySet().iterator();
        while (iter.hasNext())
        {
            Long key_l   = (Long) iter.next();
            long key     = key_l.longValue();
            Long count_l = (Long) other._samples.get((Object) key_l);
            long count   = count_l.longValue();

            add(key, count);
        }

        return;
    }

    public synchronized void add(long sample)
    {
        add(sample, 1);
        return;
    }

    public synchronized void add(long sample, long count)
    {
        /*
        if (_skip > _max_skip)
        {
            _skip++;
            return;
        }
        */

        _num_samples += count;

        //  Compute max.
        if (sample > _max_sample)
        {
            _max_sample = sample;
        }

        //  Compute min.
        if (sample < _min_sample)
        {
            _min_sample = sample;
        }

        //  Compute running total.
        _sum_of_samples += sample * count;
        _square_sum_of_samples += (sample * sample) * count;
        _prod_of_samples *= Math.pow( sample, count );

        //  Record sample.
        Long key_l   = new Long(sample);
        Long count_l = (Long) _samples.get(key_l);
        if (count_l == null)
        {
            count_l = new Long(count);
        }
        else
        {
            count_l = new Long(count_l.longValue() + count);
        }
        _samples.put((Object) key_l, (Object) count_l);

        return;
    }

    public synchronized long getNumSamples()
    {
        return _num_samples;
    }

    public synchronized long getPercentile(double percentile)
    {
      int entryIndex = (int)((double)_num_samples * percentile);
      int currIndex = 0;
      Iterator i = _samples.entrySet().iterator();
      Map.Entry entry = null;
      long key = 0;
      long count = 0;
      boolean found = false;
      long value = 0;
      while(i.hasNext() && !found)
      {
	entry = (Map.Entry)i.next();
	key = ((Long)entry.getKey()).longValue();
	count = ((Long)entry.getValue()).longValue();
	
	if(entryIndex >= currIndex && entryIndex < currIndex+(int)count)
	  found = true;
	else
	  currIndex += (int)count;
      }

      if(!found && _num_samples != 0)
	Carp.die("StatRecorder.getPercentile: something is wrong with" +
		 " obtaining the " + (int)(percentile*100.0) + 
		 "th percentile from count=" + _num_samples + 
		 " samples=" + _samples + ".");

      return key;
    }

    public synchronized long getMedian()
    {
      return getPercentile(0.5);
    }

    public synchronized long getMin()
    {
        return _min_sample;
    }

    public synchronized long getMax()
    {
        return _max_sample;
    }

    public synchronized double getMean()
    {
        double mean = 0.0;

        if (_num_samples != 0)
        {
            mean = (_sum_of_samples * 1.0) / _num_samples;
        }

        return mean;
    }

    public synchronized double getGeometricMean()
    {
        double geo_mean = 0.0;

        if (_num_samples != 0)
        {
            geo_mean = 
		Math.pow((_prod_of_samples * 1.0), 1.0/(_num_samples*1.0));
        }

        return geo_mean;
    }

    public synchronized double getStddev()
    {
      double mean = getMean();
      double stddev = 0.0;
      if(_num_samples != 0) 
      {
	  stddev = _num_samples * mean * mean + 
	      _square_sum_of_samples - 
	      2 * mean * _sum_of_samples;
      }
      return stddev;
    }

    public synchronized String getHistogram()
    {
        return getHistogram(1);
    }

    public synchronized String getHistogram(int bucket_size)
    {
        String histogram = new String(_name + " Histogram: (" +
                                      _num_samples + " samples)\n");

        //  This iterator contains the keys in ascending order.
        int bucket_count = 0;
        Iterator iter = _samples.keySet().iterator();
        if (! iter.hasNext())
        {
            histogram += "  <no samples>\n";
        }
        else
        {
            Long key_l   = (Long) iter.next();
            long key     = key_l.intValue();
            Long count_l = (Long) _samples.get((Object) key_l);
            long count   = count_l.intValue();

            long current_bucket_num = key / bucket_size;

            while (iter.hasNext())
            {
                key_l   = (Long) iter.next();
                key     = key_l.longValue();
                count_l = (Long) _samples.get((Object) key_l);

                long sample_bucket_num = key / bucket_size;
                if (sample_bucket_num == current_bucket_num)
                {
                    count += count_l.longValue();
                }
                else
                {
                    //  Write out the current bucket.
                    String bucket_s =
                        getHistogramBucketString(current_bucket_num,
                                                 bucket_size);
                    histogram +=
                        _tag + "\t" + bucket_s + "\t" +
                        Long.toString(count) + "\n";

                    //  Get ready for the next bucket.
                    current_bucket_num = sample_bucket_num;
                    count = count_l.longValue();
                }
            }

            //  Write out the final bucket.
            String bucket_s =
                getHistogramBucketString(current_bucket_num, bucket_size);
            histogram +=
                _tag + "\t" + bucket_s + "\t" +
                Long.toString(count) + "\n";
        }

        return histogram;
    }

    private String getHistogramBucketString(long current_bucket_num,
                                            long bucket_size)
    {
        String bucket_s =
            Long.toString(current_bucket_num * bucket_size);
        if (bucket_size != 1)
        {
            bucket_s += "-" +
                Long.toString((current_bucket_num + 1) * bucket_size - 1);
        }

        return bucket_s;
    }
}
