/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

import java.util.Hashtable;
import java.util.Date;

/**
 * <code>StopWatch</code> provides timing utilities for profiling code
 * performance.  This class relies on {@link java.util.Date},
 * and therefore can provide only millisecond-granularity timing.
 * <p>
 * All timestamps are retrieved from {@link
 * ostore.util.Clock#current_date}, so this class is simulator-safe.
 * 
 * @author     Dennis Geels
 * @version    $Id: StopWatch.java,v 1.5 2002/01/08 18:26:12 geels Exp $
 */
public class StopWatch
{
    ////////////////////       CONSTANTS	////////////////////
    ////////////////////    STATIC VARIABLES 	////////////////////

    /**
     * A <code>NodeId->StopWatch</code> map of <code>StopWatch</code> objects.
     * Each virtual machine has a single <code>StopWatch</code>.
     */
    private static Hashtable stopwatches;

    ////////////////////       VARIABLES	////////////////////

    /**
     * The tag-><code>Date</code> map which stores pending timers.
     */
    private Hashtable start_times;

    /**
     * The <code>NodeId</code> for this machine (parameterizes virtual
     * SandStorm instances).
     */
    private NodeId node_id;
    
    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	// Initialize the map of signals.
	stopwatches = new Hashtable();
    }
    
    ////////////////////      CONSTRUCTORS	////////////////////

    private StopWatch( NodeId node_id )
    {
	this.node_id = node_id;
	start_times = new Hashtable();
    }
    
    ////////////////////     STATIC METHODS 	////////////////////

    /**
     * Returns the <code>StopWatch</code> instance corresponding to a
     * <code>NodeId</code>. 
     *
     * @param node_id	The <code>NodeId</code> which parameterizes a
     * virtual SandStorm instance.
     * @return the (possibly new) <code>StopWatch</code> instance
     * corresponding to <code>node_id</code>.
     */
    public static StopWatch getInstance( NodeId node_id )
    {
	StopWatch result = null;
	synchronized( stopwatches ) {
	    result = (StopWatch)stopwatches.get( node_id );
	    if (result == null) {
		result = new StopWatch( node_id );
		stopwatches.put( node_id, result );
	    }
	}
	return result;
    }

    //////////////////// 	METHODS		////////////////////

    /**
     * Start a new timer.
     *
     * @param tag	The identifier for this timing.  May be reused.
     */
    public final void start( Object tag )
    {
	start_times.put( tag, Clock.current_date(node_id) );
    }

    /**
     * Return the time elapsed so far, in milliseconds.  This method does not
     * stop or reset a timer; successive calls return the monotonically
     * increasing cumulative time elapsed.
     *
     * @param tag	The identifier for this timing.
     * @return the time since the last call to <code>start</code> with this
     * tag, or 0 if no such call was made.
     */
    public final long split( Object tag )
    {
	long time;
	Date start_time = (Date)start_times.get( tag );
	if( start_time == null ) {
	    time = 0;
	} else {
	    Date now = Clock.current_date(node_id);
	    time = now.getTime() - start_time.getTime();
	}
	return time;
    }

    /**
     * Stop a timer and return the time elapsed, in milliseconds.
     * This method clears the tag's state from the <code>StopWatch</code>,
     * effectively resetting the timer.
     *
     * @param tag 	The identifier for this timing.
     * @return the time since the last call to <code>start</code> with this
     * tag, or 0 if no such call was made.
     */
    public final long stop( Object tag )
    {
	long time;
	Date start_time = (Date)start_times.remove( tag );
	if( start_time == null ) {
	    time = 0;
	} else {
	    Date now = Clock.current_date(node_id);
	    time = now.getTime() - start_time.getTime();
	}
	return time;
    }
    
    ////////////////////     INNER CLASSES	////////////////////
}
