/**
 * VID.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.util;

/**
 * A wrapper class for a <code>DataObject</code> version identifier.  A
 * <code>VID</code> contains a single <code>SecureHash</code>, which is
 * either the VHash of a top-block, the VGUID of a top-block, or the GUID
 * of a log object (LGUID).  <code>VID</code> is also used to refer
 * to the archival for of a <code>DataObject</code>, a disseminatable 
 * (DISS) (e.g. fragment or tombstone).
 *
 * @author	Dennis Geels
 * @version	$Id: VID.java,v 1.16 2003/02/07 20:59:58 hweather Exp $
 */

public class VID implements QuickSerializable, Comparable
{
    ////////////////////       CONSTANTS	////////////////////

    /**
     * The type code for a <code>VID</code> that consists of the VHash of a
     * <code>DataObject</code>'s top block.
     */
    public static final byte VHASH = 1;
    
    /**
     * The type code for a <code>VID</code> that consists of the VGUID of a
     * <code>DataObject</code>'s top block (A checkpoint GUID).
     */
    public static final byte VGUID = 2;

    // TODO:  Drop LGUID?
    // TODO:  allow HASH _and_ GUID?
    /**
     * The type code for a <code>VID</code> that consists of the GUID of
     * the log object for a version.
     */
    public static final byte LGUID = 3;

    /**
     * The type code for a <code>VID</code> that consists of the DISS of
     * the archival form of a <code>DataObject</code> block. 
     */
    public static final byte DISS = 4;
    
    
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * A code identifying which sort of identifier this one is.
     * One of {@link #VHASH}, {@link #VGUID}, {@link #LGUID}, {@link DISS}
     */
    private byte type;
    
    /**
     * The real version id.
     */
    private SecureHash id;
    
    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    /**
     * Constructs a new <code>VID</code>.
     */
    public VID( byte type, SecureHash id )
    {
	this.type = type;
	this.id = id;
    }

    /**
     * Constructs a <code>VID</code> from its
     * <code>QuickSerializable</code> form.
     */
    public VID(InputBuffer buffer)
        throws QSException
    {
        type = buffer.nextByte();
        id = (SecureHash) buffer.nextObject();
        return;
    }

    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Returns the type of identifier for this <code>VID</code>.
     * @return	One of {@link #VHASH}, {@link #VGUID}, {@link #LGUID},
     * {@link #DISS}
     */
    public byte get_type()
    {
	return type;
    }

    /**
     * Returns the VHash, VGUID, or GUID, as appropriate.
     */
    public SecureHash get_id()
    {
	return id;
    }

    /* Specified by ostore.util.QuickSerializable */
    public void serialize(OutputBuffer buffer)
    {
        buffer.add(type);
        buffer.add(id);
        return;
    }

    /**
     * Imposes an order on <code>VID</code>s.  The order is equivalent to
     * the lexicographic order on corresponding
     * <code>VID.get_type(),VID.get_id()</code> pairs. 
     */
    public int compareTo( Object o )
    {
	// Don't worry about <code>ClassCastException</code>:	
	VID ovid = (VID)o;
	int order;
	if( type < ovid.type ) {
	    order = -1;
	} else if( type > ovid.type ) {
	    order = 1;
	} else {	// same type - order by id
	    order = id.compareTo( ovid.id );
	}
	return order;
    }
    
    /**
     * Returns the hashcode of the internal identifier.
     */
    public int hashCode()
    {
	return id.hashCode();
    }

    /**
     * Tests for equality (type and id).
     *
     * @param other	possibly a <code>VID</code>.
     *
     * @return <code>true</code> iff <code>other</code> is a
     * <code>VID</code> with the same identifier (including type).
     */
    public boolean equals( Object other )
    {
	if( other == this ) {
	    return true;
	} else if( ! ( other instanceof VID ) ) {
	    return false;
	} else {
	    VID v = (VID)other;
	    return( (v.type == this.type) &&
		    (v.id.equals( this.id )) );
	}
    }
    
    /**
     * Returns a human-readable representation of this <code>VID</code>.
     */
    public String toString()
    {
	String val;
	switch( type ) {
	case VHASH:
	    val = "(VHASH: ";
	    break;
	case VGUID:
	    val = "(VGUID: ";
	    break;
	case LGUID:
	    val = "(LGUID: ";
	    break;
	case DISS:
	    val = "(DISS: ";
	    break;
	default:
	    val = "(UNKNOWN: ";
	    break;
	}
	val += id.toString() + ")";

	return val;
    }
    ////////////////////     INNER CLASSES	////////////////////
}
