#!/usr/bin/perl -w

use strict;
use Getopt::Std;
use File::Copy;
use Tools;
use vars qw($opt_u $opt_x $opt_h $opt_b $opt_e $opt_n $opt_r $opt_l);

sub usage
{
    select(STDERR);
    print <<'EOUsage';

run-size-test.pl [options]
  Options:
    -h           - Display this help message.
    -b           - Size of btree data block.
    -e           - Erasure code type.  Valid values are cauchy
                        interleaved{encodeBytes}
    -n           - Total number of fragments (per block).
    -r           - Inverse rate of encoding.
    -l           - Use localhost
    -u UPDT_SIZE - Sweep the size of the update submitted by test stages
                   across the range describe by UPDT_SIZE.  The format
                   for notating the range is described below
    -x ROC       - Relocate the test in the X-cluster to start using
                   machine xROC.millennium.berkeley.edu.

  Ranges can be specified in a very flexible manner.  Most generally,
  each range has a 'start' value, an 'end' value, a 'skip operator',
  and a 'skip value'.  The skip operator is one of the four basic
  arithmetic operators (+, -, *, and /).  During execution, when the
  range operator is '+' or '*', the range leads to execution as
    for (x = $start; x <= $end; x = x $skip_op $skip_value) { ; }
  When the operator is '-' or '/', the range leads to execution as
    for (x = $start; x >= $end; x = x $skip_op $skip_value) { ; }

  A range is notated as '$start-$end/$skip_op$skip_value'.  Other
  simpler forms are also accepted.  These forms include '$start' which
  executes only one point, '$start-$end' which uses the '+' operator
  with a skip value of 1, and '$start-$end/$skip_value' which defaults
  to use the '+' operator.  Examples of valid ranges include

    o  28
    o  28-56
    o  2-10/2
    o  128-1024/*2
    o  2048-128//2
EOUsage

    exit(1);
}

#  Constants.
my $BLOWUP_RESULT_LOG = "/tmp/blowup_result_log.$$.log";
open(RESULT_LOG, ">> $BLOWUP_RESULT_LOG") ||
  die("Could not open result log ($BLOWUP_RESULT_LOG).");
print(RESULT_LOG "# <ErasureType> " .
       "<NumFrags> " .
       "<InverseRate >" .
       "<BlockSize> " .
       "<DobjSize> " .
       "<TotalBlockSize> " .
       "<TotalErasureBlockSize> " .
       "<TotalDisseminateSize>\n");
close(RESULT_LOG);

#  Test parameters.
my $SKIP_ARCHIVE = 'true';
my $INLINE_ARCHIVE = 'false';
my $TEST_LENGTH = 10;
my $CLUSTER_MACHINE = 1;
my $RESULT_INPUT_FILE;
my $DOBJ_SIZE = 16384;
my $KEY_SIZE = 512;

my @UPDATE_SIZE_SPEC;
my $UPDATE_SIZE_START = 1024;
my $UPDATE_SIZE_END   = 1024;
my $UPDATE_SIZE_OP    = '+';
my $UPDATE_SIZE_SKIP  = 1;

my $BLOCK_SIZE = 4096;
my $ERASURE_TYPE = "cauchy";
my $NUM_FRAGS = 32;
my $INVERSE_RATE = 2;

my $LOCAL = 0;

#  Get the user-defined parameters.
my $opts_ok = getopts('hlu:b:e:n:r:x:');
if ($opt_h) { &usage; }
if ($opt_b) { $BLOCK_SIZE = $opt_b; }
if ($opt_e) { $ERASURE_TYPE = $opt_e; }
if ($opt_n) { $NUM_FRAGS = $opt_n; }
if ($opt_r) { $INVERSE_RATE = $opt_r; }
if ($opt_x) { $CLUSTER_MACHINE = $opt_x; }
if ($opt_l) { $LOCAL = 1; }
if ($opt_u)
{
    @UPDATE_SIZE_SPEC = &parse_range_spec($opt_u);
    my @spec = @UPDATE_SIZE_SPEC;
    $UPDATE_SIZE_START = shift(@spec);
    $UPDATE_SIZE_END   = shift(@spec);
    $UPDATE_SIZE_OP    = shift(@spec);
    $UPDATE_SIZE_SKIP  = shift(@spec);
}

#  The command line is valid if getopts returned without error and
#  there are no arguments left over.  $#ARGV returns the number of
#  remaining arguments minus one.
if ((! $opts_ok) || ($#ARGV != -1)) { print $#ARGV . " " . $opts_ok . "\n"; &usage; }

#  Check that the set of input parameters is valid.
#if ($UPDATE_SIZE_END * 2 > $DOBJ_SIZE)
#{
#    my $old_fd = select(STDERR);
#    print<<EOErrMsg;
#ERROR:  The largest update size cannot be more than half of the size
#of the data object.  UpdateSize=$UPDATE_SIZE_END, DobjSize=$DOBJ_SIZE.
#EOErrMsg
#    select($old_fd);
#    &usage();
#    exit(1);
#}

if ($KEY_SIZE != 512 && $KEY_SIZE != 1024)
{
    my $old_fd = select(STDERR);
    print<<EOErrMsg;
ERROR:  Invalid key size.
EOErrMsg
    select($old_fd);
    &usage();
    exit(1);
}

#  Create the archiver config files with the right settings.
copy("./archiver.cfg", "/tmp/archiver-$$.cfg");
my %archiver_vals =
  ("LogEvents"   => "true" );
&override("/tmp/archiver-$$.cfg", %archiver_vals);

#  Run experiments to collect data.
my $update_size = $UPDATE_SIZE_START;
my $update_size_comparator = '<=';
if ($UPDATE_SIZE_OP eq '-' || $UPDATE_SIZE_OP eq '/')
{
    $update_size_comparator = '>=';
}
while (eval("$update_size $update_size_comparator $UPDATE_SIZE_END"))
{
    print("Running size test:  " .
          "UpdateSize=$update_size, " .
	  "BlockSize=$BLOCK_SIZE, ErasureType=$ERASURE_TYPE, " .
	  "NumFrags=$NUM_FRAGS, InverseRate=$INVERSE_RATE\n");

    #  Create configuration files.
    my $tester_cfg = &write_tester_config_file($update_size,
					      $BLOCK_SIZE, $ERASURE_TYPE,
					      $NUM_FRAGS, $INVERSE_RATE);
    my $exp_cfg    = &write_experiment_file();

    #  Run the experiment.
    #if ($KEY_SIZE == 512)
    #{
    #    system("KEYSHARE_PREFIX=ir ./run-experiment $exp_cfg");
    #}
    #elsif ($KEY_SIZE == 1024)
    #{
    #    system("KEYSHARE_PREFIX=ir-1024-4-2 ./run-experiment $exp_cfg")
    #}
    #else
    #{
    #    print(STDERR "Script error.");
    #}

    my $log = "/tmp/blowup-log.$$.log";
    print "log=$log\n";
    my $pid = &fork_process( $log, "./run-experiment", $exp_cfg );
    print "pid=$pid\n"; 
    if( &returned_zero( $pid ))
    {
      print "process returned 0\n";
    }
    else
    {
      die "process did NOT return 0, but pid\n";
    }

    #  Read the result log.
    my @stats = &analyze_result_log($pid);
    &write_result_to_log($DOBJ_SIZE, $update_size, $BLOCK_SIZE, $ERASURE_TYPE, $NUM_FRAGS, $INVERSE_RATE, @stats);

    #  Remove the files we created.
    unlink($tester_cfg, $exp_cfg);
    unlink($log);
    $update_size = eval("$update_size $UPDATE_SIZE_OP $UPDATE_SIZE_SKIP");
}

#  Remove the inner ring config file used for this test.
unlink("/tmp/archiver-$$.cfg");

exit(0);

sub parse_range_spec
{
    my $spec = shift(@_);

    my $start = 0;
    my $end   = 0;
    my $op    = '+';
    my $skip  = 1;

    #  (\d+)                  The start of the range...
    #  (-(\d+))?              ...optionally a '-' and the end of the range...
    #  (\/([-\+\*\/])?(\d+))? ...optionally a '/', an optional 
    #                         operation (-, +, *, or /) and the skip value.
    if ($spec =~ /(\d+)(-(\d+))?(\/([-\+\*\/])?(\d+))?/ )
    {
        $start = $1;
        $end   = $start;
        if (defined($3)) { $end = $3; }
        if (defined($5)) { $op = $5; }
        if (defined($6)) { $skip = $6; }
    }

    print("start=$start \n");
    print("end=$end   \n");
    print("op=$op    \n");
    print("skip=$skip  \n");

    my @range;
    push(@range, $start);
    push(@range, $end);
    push(@range, $op);
    push(@range, $skip);

    return @range;
}

sub write_result_to_log
{
    my $dobj_size    = shift(@_);
    my $update_size  = shift(@_);
    my $block_size   = shift(@_);
    my $erasure_type = shift(@_);
    my $num_frags    = shift(@_);
    my $inverse_rate = shift(@_);
    my $totalBlockSize = shift(@_);
    my $totalErasureBlockSize = shift(@_);
    my $totalDisseminateSize = shift(@_);

    my $log =
       "$erasure_type " .
       "$num_frags " .
       "$inverse_rate " .
       "$block_size " .
       "$update_size " .
       "$totalBlockSize " .
       "$totalErasureBlockSize " .
       "$totalDisseminateSize\n";

    open(RESULT_LOG, ">> $BLOWUP_RESULT_LOG") ||
      die("Could not open result log ($BLOWUP_RESULT_LOG).");
    print(RESULT_LOG "$log");
    close(RESULT_LOG);

    return;
}

sub analyze_result_log
{
    my $pid = shift(@_);
    print "analyze_result_log: pid==$pid\n";
    my @internal_logs = glob("/tmp/*-${pid}-*GenerateChkptStage.log");
    (scalar(@internal_logs) == 1) ||
      die("Cannot determine result log name. @internal_logs\n");

    my $result_log = shift(@internal_logs);

    open(RESULT, "$result_log") ||
        die("Could not open result log.\n");

      my $stageId = 0;
      my $vhash = "";
      my $vguid = "";
      my $numFrags = 0;
      my $inverseRate=  0;
      my $erasureType = 0;
      my $startTime = 0;
      my $totalTime = 0;
      my $encodeHandlerTimeSum = 0;
      my $encodeHandlerTimeSquareSum = 0;
      my $encodeMethodTimeSum = 0;
      my $encodeMethodTimeSquareSum = 0;
      my $encodeTimeSum = 0;
      my $encodeTimeSquareSum = 0;
      my $vhashTreeTimeSum = 0;
      my $vhashTreeTimeSquareSum = 0;
      my $totalBlockSize = 0;
      my $totalErasureBlockSize = 0;
      my $totalDisseminateSize = 0;
      my $count = 0;
#     my $modifiedGuids = "";
    my $line = "";
    while ($line = <RESULT>)
    {
      #print $line;
      my @Fields = split /\s/, $line;
      $stageId = $Fields[0];
      $vhash = $Fields[1];
      $vguid = $Fields[2];
      $numFrags = $Fields[3];
      $inverseRate = $Fields[4];
      $erasureType = $Fields[5];
      $startTime = $Fields[6];
      $totalTime = $Fields[7];
      $encodeHandlerTimeSum = $Fields[8];
      $encodeHandlerTimeSquareSum = $Fields[9];
      $encodeMethodTimeSum = $Fields[10];
      $encodeMethodTimeSquareSum = $Fields[11];
      $encodeTimeSum = $Fields[12];
      $encodeTimeSquareSum = $Fields[13];
      $vhashTreeTimeSum = $Fields[14];
      $vhashTreeTimeSquareSum = $Fields[15];
      $totalBlockSize = $Fields[16];
      $totalErasureBlockSize = $Fields[17];
      $totalDisseminateSize = $Fields[18];
      $count = $Fields[19];
#     $modifiedGuids = $Fields[20];

    }
    

    close(RESULT);
    unlink($result_log) || die("Could not remove result log.\n");

    #print("$min $max $mean $median $fiveperc $ninefiveperc\n");
    my @stats;
    push(@stats, $totalBlockSize);
    push(@stats, $totalErasureBlockSize);
    push(@stats, $totalDisseminateSize);

    my @internal2_logs = glob("/tmp/*-${pid}-*internal.log*");
    unlink(@internal2_logs);

    return @stats;
}

sub write_experiment_file
{
    my $exp_cfg = "/tmp/size.$$.exp";
    open(EXP, ">$exp_cfg") ||
      die("Could not create experiment configuration file ($exp_cfg).\n");
    my $old_fd = select(EXP);
    my $hosts = "ROC";
    my $fed = "";
    my $archiver = "";
    my $storage0 = "";
    my $storage1 = "";
    my $tester = "";

    if($LOCAL==1)
    {
      $hosts = "localhost";
    }
    else
    {
       $fed = $CLUSTER_MACHINE - 1;
       $archiver = $CLUSTER_MACHINE + 1 - 1;
       $storage0 = $CLUSTER_MACHINE + 2 - 1;
       $storage1 = $CLUSTER_MACHINE + 3 - 1;
       $tester = $CLUSTER_MACHINE + 4 - 1;
    }

    print <<EOExp;
hosts      $hosts
topology   NONE
Federation \${run_dir}/federation.cfg      fed     daemon $fed
Archiver       /tmp/archiver-$$.cfg        static  daemon $archiver
# StorageServer0 \${run_dir}/storage.cfg     static  daemon $storage0
# StorageServer1 \${run_dir}/storage.cfg     static  daemon $storage1
# StorageServer2 \${run_dir}/storage.cfg     static  daemon $storage0
# StorageServer3 \${run_dir}/storage.cfg     static  daemon $storage1
# StorageServer4 \${run_dir}/storage.cfg     static  daemon $storage0
# StorageServer5 \${run_dir}/storage.cfg     static  daemon $storage1
# StorageServer6 \${run_dir}/storage.cfg     static  daemon $storage0
# StorageServer7 \${run_dir}/storage.cfg     static  daemon $storage1
# StorageServer8 \${run_dir}/storage.cfg     static  daemon $storage0
# StorageServer9 \${run_dir}/storage.cfg     static  daemon $storage1
# StorageServer10 \${run_dir}/storage.cfg    static  daemon $storage0
# StorageServer11 \${run_dir}/storage.cfg    static  daemon $storage1
# StorageServer12 \${run_dir}/storage.cfg    static  daemon $storage0
# StorageServer13 \${run_dir}/storage.cfg    static  daemon $storage1
# StorageServer14 \${run_dir}/storage.cfg    static  daemon $storage0
# StorageServer15 \${run_dir}/storage.cfg    static  daemon $storage1
# StorageServer16 \${run_dir}/storage.cfg    static  daemon $storage0
# StorageServer17 \${run_dir}/storage.cfg    static  daemon $storage1
# StorageServer18 \${run_dir}/storage.cfg    static  daemon $storage0
# StorageServer19 \${run_dir}/storage.cfg    static  daemon $storage1
# StorageServer20 \${run_dir}/storage.cfg    static  daemon $storage0
# StorageServer21 \${run_dir}/storage.cfg    static  daemon $storage1
# StorageServer22 \${run_dir}/storage.cfg    static  daemon $storage0
# StorageServer23 \${run_dir}/storage.cfg    static  daemon $storage1
# StorageServer24 \${run_dir}/storage.cfg    static  daemon $storage0
# StorageServer25 \${run_dir}/storage.cfg    static  daemon $storage1
# StorageServer26 \${run_dir}/storage.cfg    static  daemon $storage0
# StorageServer27 \${run_dir}/storage.cfg    static  daemon $storage1
# StorageServer28 \${run_dir}/storage.cfg    static  daemon $storage0
# StorageServer29 \${run_dir}/storage.cfg    static  daemon $storage1
# StorageServer30 \${run_dir}/storage.cfg    static  daemon $storage0
# StorageServer31 \${run_dir}/storage.cfg    static  daemon $storage1
# Creator         \${run_dir}/creator.cfg    static  daemon $fed
Tester     /tmp/size-tester.$$.cfg     static  mortal $tester
EOExp

    close(EXP);
    select($old_fd);

    return $exp_cfg;
}

sub write_tester_config_file
{
    my $dobj_size   = shift(@_);

    my $block_size = shift(@_);
    my $erasure_type = shift(@_);
    my $num_frags   = shift(@_);
    my $inverse_rate = shift(@_);

    my $tester_cfg = "/tmp/size-tester.$$.cfg";
    open(TESTER, ">$tester_cfg") ||
        die("Could not create tester configuration file ($tester_cfg).\n");

    my $old_fd = select(TESTER);
    &write_tester_header($block_size, $erasure_type, $num_frags, $inverse_rate);
    &write_tester_stage($dobj_size);
    &write_tester_footer();

    close(TESTER);
    select($old_fd);

    return $tester_cfg;
}

sub write_tester_header
{
    my $block_size = shift(@_);
    my $erasure_type = shift(@_);
    my $num_frags   = shift(@_);
    my $inverse_rate = shift(@_);

    print <<EOHeader;
<sandstorm>
  <global>
  defaultThreadManager TPSTM
    <TPSTM>
    governor false
    governorDelay 2000
    governorMaxThreads 10
    </TPSTM>

    <TPPTM>
    numCpus 4
    maxThreads 4
    </TPPTM>

    <profile>
      enable false
      delay 1000
      filename \${SSProfileFile}
      graphfilename \${SSGraphFile}
      graph false
      sockets false
    </profile>
    <initargs>
        node_id \${NodeID}
        Profile false
    </initargs>
  </global>

<stages>
  <Network>
    class ostore.network.Network
    queueThreshold 1000
    <initargs>
    </initargs>
  </Network>

  <Router>
    class ostore.tapestry.impl.Router
    queueThreshold 1000
    <initargs>
      pkey \${NodePublicKey}
      skey \${NodePrivateKey}
      dynamic_route \${TapestryMode}
    </initargs>
  </Router>

  <TClient>
    class ostore.tapestry.impl.StaticTClient
    queueThreshold 1000
    <initargs>
      federation \${FederationID}
      dynamic_route \${TapestryMode}
    </initargs>
  </TClient>

  <FakeClient>
    class ostore.client.FakeClient
    queueThreshold 1000
    <initargs>
      UseDynamicTapestry false
      ProvideApiServer false

      AlarmTimeout  60000
      rp_pkeyfile \${RPPublicKey}
      block_size  $block_size
      num_frags   $num_frags
      inv_rate    $inverse_rate
      erasure_type $erasure_type
      skip_archive false
      inline_archive true

      LogEvents	false
      LogFilename \${InternalLog}
      LogSampleRate	1
    </initargs>
  </FakeClient>

  <UpdateVerifiableBlocksStage>
    class ostore.archive.UpdateVerifiableBlocksStage
    queueThreshold 1000
    <initargs>
      LogEvents	false
      LogFilename \${InternalLog}-UpdateVerifiableBlocksStage.log
      LogSampleRate	1
    </initargs>
  </UpdateVerifiableBlocksStage>

  <UpdateStage>
    class ostore.update.UpdateStage
    <initargs>
      LogEvents	false
      LogFilename \${InternalLog}-UpdateStage.log
      LogSampleRate	1
    </initargs>
  </UpdateStage>

  <ReadStage> 
    class ostore.read.ReadStage
    queueThreshold 1000
    <initargs>
      LogEvents	false
      LogFilename \${InternalLog}-ReadStage.log
      LogSampleRate	1
    </initargs>
  </ReadStage>

  <RecoalesceBlkStage>
    class ostore.archive.RecoalesceBlkStage
    queueThreshold 1000
    <initargs>
      RandomSeed 102
      Verify false
      LogEvents	false
      LogFilename \${InternalLog}-RecoalesceBlkStage.log
      LogSampleRate	1
    </initargs>
  </RecoalesceBlkStage>

  <RequestorStage>
    class ostore.archive.RequestorStage
    queueThreshold 1000
    <initargs>
      RequestCardinality	20
      CooperativeCaching	false
      AlarmTimeout		60000
      LogEvents	false
      LogFilename \${InternalLog}-RequestorStage.log
      LogSampleRate	1
    </initargs>
  </RequestorStage>

  <Cache>
    class ostore.cache.MemoryCache
    queueThreshold 1000
    <initargs>
      CacheSize 128000000
      cachedir \${CacheDir}
      Publish false
    </initargs>
  </Cache>

EOHeader

    return;
}

sub write_tester_stage
{
    my $dobj_size = shift(@_);
    
    print <<EOStage;
  <OSApplicationHowTo>
    class ostore.apps.test.OSApplicationHowTo
    queueThreshold 1000
    <initargs>
      DobjSize             $dobj_size
      rp_pkeyfile        \${RPPublicKey}
      principal_pkeyfile \${PrincipalPublicKey}
      principal_skeyfile \${PrincipalPrivateKey}
    </initargs>
  </OSApplicationHowTo>
EOStage

    return;
}

sub write_tester_footer
{
    print <<EOFooter;
</stages>
</sandstorm>
EOFooter

    return;
}
