#!/usr/bin/perl -w

use strict;
use Getopt::Std;
use File::Copy;
use Tools;
use vars qw($opt_a $opt_c $opt_i $opt_u $opt_k $opt_t $opt_x $opt_f $opt_h $opt_p $opt_r $opt_o $opt_m $opt_w $opt_s);

sub usage
{
    select(STDERR);
    print <<'EOUsage';

run-tput-test.pl [options]
  Options:
    -h           - Display this help message.
    -a           - Enable the archive.  By default, the inner ring does
                   not initiate the inner ring.
    -p           - Publish.  That is, use tapestry.  (leave this out to 
                   use archive.
    -r           - Number of fragments to request from Tapestry.
    -i           - Inline the archive.  This option has no effect if the
                   archive is disabled.  By default, when the archive is
                   enabled, it runs asynchronously with the updates.
    -s SETTLE    - Pause SETTLE ms between the initialization and test
                   phases.
    -c MACHINES  - Use MACHINES machines to apply load.  If the number
                   of machines requested is greater than the number of
                   test stages requested, the extra machines will remain
                   idle through the test.
    -t TESTERS   - Sweep the number of test stages used of the range
                   described by TESTERS.  The format for notating the
                   range is described below.
    -u UPDT_SIZE - Sweep the size of the update submitted by test stages
                   across the range describe by UPDT_SIZE.  The format
                   for notating the range is described below.  Note, the
                   size of the data object is set to twice the length of
                   the update size.
    -w WARM_CFG  - Warm the configuration by performing a number of
                   updates at an interval as specified by the WARM_CFG
                   phase descriptor.  The format for notating the test
                   stage is described below.
    -m MEAS_CFG  - Run each test for LENGTH seconds.
                   Measure the configuration by performing a number of
                   updates at an interval as specified by the MEAS_CFG
                   phase descriptor.  The format for notating the test
                   stage is described below.
    -x ROC       - Relocate the test in the X-cluster to start using
                   machine xROC.millennium.berkeley.edu.
    -f INNER_CFG - Use the file INNER_CFG as the configuration file
                   for members of the inner ring.
    -k KEY_SIZE  - Use keys of of length KEY_SIZE bits.  Valid values
                   are 512 and 1024.  The default value is 512.

  Ranges can be specified in a very flexible manner.  Most generally,
  each range has a 'start' value, an 'end' value, a 'skip operator',
  and a 'skip value'.  The skip operator is one of the four basic
  arithmetic operators (+, -, *, and /).  During execution, when the
  range operator is '+' or '*', the range leads to execution as
    for (x = $start; x <= $end; x = x $skip_op $skip_value) { ; }
  When the operator is '-' or '/', the range leads to execution as
    for (x = $start; x >= $end; x = x $skip_op $skip_value) { ; }

  A range is notated as '$start-$end/$skip_op$skip_value'.  Several
  shortened forms are also accepted.  These forms include '$start' which
  executes only one point, '$start-$end' which uses the '+' operator
  with a skip value of 1, and '$start-$end/$skip_value' which defaults
  to use the '+' operator.  Examples of valid ranges include

    o  28
    o  28-56
    o  2-10/2
    o  128-1024/*2
    o  2048-128//2

  The different stages of the test are described by phase descriptor.
  A phase descriptor specifies the number of updates to perform during
  that stage and the amount of time to pause, in milliseconds, before
  submitting each update.  The general form of the phase descriptor is
  '$number_updates/$pause_time'.  The shortened for '$number_updates'
  is also accepted; the pause time defaults to 0.  Examples of valid
  phase descriptors include

    o  1000
    o  1000/5
EOUsage

    exit(1);
}

#  Constants.
my $TPUT_RESULT_LOG;
my $MASTER_MACHINE_OFFSET = 5;
my $TESTER_MACHINE_OFFSET = 6;

#  Test parameters.
my $SKIP_ARCHIVE = 'true';
my $INLINE_ARCHIVE = 'false';
my $TEST_MACHINES = 1;
my $TEST_LENGTH = 10;
my $CLUSTER_MACHINE = -1;
my $KEY_SIZE = 512;
my $INNER_CFG = "./inner.cfg";

my $SETTLE = 5000;

my @TEST_STAGES_SPEC;
my $TEST_STAGES_START = 1;
my $TEST_STAGES_END   = 1;
my $TEST_STAGES_OP    = '+';
my $TEST_STAGES_SKIP  = 1;

my @WARM_PHASE_DESC;
my $WARM_UPDATES      = 1000;
my $WARM_PAUSE        = 5;

my @MEAS_PHASE_DESC;
my $MEAS_LENGTH      = 1000;
my $MEAS_PAUSE        = 100;

my @UPDATE_SIZE_SPEC;
my $UPDATE_SIZE_START = 1024;
my $UPDATE_SIZE_END   = 1024;
my $UPDATE_SIZE_OP    = '+';
my $UPDATE_SIZE_SKIP  = 1;

my $PUBLISH = "false";
my $REQ_CARDINALITY = 20;
my $LOCAL = 0;

#  Get the user-defined parameters.
my $opts_ok = getopts('haiopc:r:u:m:w:s:k:t:x:g:f:');
if ($opt_h) { &usage; }
if ($opt_a) { $SKIP_ARCHIVE = 'false'; }
if ($opt_i) { $INLINE_ARCHIVE = 'true'; }
if ($opt_p) { $PUBLISH = 'true'; }
if ($opt_r) { $REQ_CARDINALITY = $opt_r; }
if ($opt_o) { $LOCAL = 1; }
if ($opt_c) { $TEST_MACHINES = $opt_c; }
if ($opt_s) { $SETTLE = $opt_s; }
if ($opt_x) { $CLUSTER_MACHINE = $opt_x; }
if ($opt_k) { $KEY_SIZE = $opt_k; }
if ($opt_f) { $INNER_CFG = $opt_f; }
if ($opt_t)
{
    @TEST_STAGES_SPEC = &parse_range_spec($opt_t);
    my @spec = @TEST_STAGES_SPEC;
    $TEST_STAGES_START = shift(@spec);
    $TEST_STAGES_END   = shift(@spec);
    $TEST_STAGES_OP    = shift(@spec);
    $TEST_STAGES_SKIP  = shift(@spec);
}
if ($opt_w)
{
    @WARM_PHASE_DESC = &parse_phase_desc($opt_w);
    my @desc = @WARM_PHASE_DESC;
    $WARM_UPDATES = shift(@desc);
    $WARM_PAUSE   = shift(@desc);
}
if ($opt_m)
{
    @MEAS_PHASE_DESC = &parse_phase_desc($opt_m);
    my @desc = @MEAS_PHASE_DESC;
    $MEAS_LENGTH = shift(@desc);
    $MEAS_PAUSE   = shift(@desc);
}
if ($opt_u)
{
    @UPDATE_SIZE_SPEC = &parse_range_spec($opt_u);
    my @spec = @UPDATE_SIZE_SPEC;
    $UPDATE_SIZE_START = shift(@spec);
    $UPDATE_SIZE_END   = shift(@spec);
    $UPDATE_SIZE_OP    = shift(@spec);
    $UPDATE_SIZE_SKIP  = shift(@spec);
}

#  The command line is valid if getopts returned without error and
#  there are no arguments left over.  $#ARGV returns the number of
#  remaining arguments minus one.
if ((! $opts_ok) || ($#ARGV != -1)) { &usage; }

#  Check that the set of input parameters is valid.
if ($KEY_SIZE != 512 && $KEY_SIZE != 1024)
{
    my $old_fd = select(STDERR);
    print<<EOErrMsg;
ERROR:  Invalid key size.
EOErrMsg
    select($old_fd);
    &usage();
    exit(1);
}
if ($CLUSTER_MACHINE == -1)
{
    my $old_fd = select(STDERR);
    print<<EOErrMsg;
ERROR:  Undefined cluster machine.
EOErrMsg
    select($old_fd);
    &usage();
    exit(1);
}

#  Compute the result log name.
$TPUT_RESULT_LOG = "/tmp/tput_result_log.noar.$$.log";
if ($opt_a && $opt_i)
{
    $TPUT_RESULT_LOG = "/tmp/tput_result_log.inar.$$.log";
}
elsif ($opt_a)
{
    $TPUT_RESULT_LOG = "/tmp/tput_result_log.dear.$$.log";
}

#  Set the environment variable for the keys.
my $key_cmd;
if ($KEY_SIZE == 512)
{
    $ENV{KEYSHARE_PREFIX} = "ir";
    $key_cmd = "KEYSHARE_PREFIX=ir";
}
elsif ($KEY_SIZE == 1024)
{
    $ENV{KEYSHARE_PREFIX} = "ir-1024-4-2";
    $key_cmd = "KEYSHARE_PREFIX=ir-1024-4-2";
}
else
{
    print(STDERR "Script error.");
}

#  Create the inner ring config files with the right archive settings.
copy($INNER_CFG, "/tmp/inner.cfg");
my %inner_vals =
  ("skip_archive"   => $SKIP_ARCHIVE,
   "inline_archive" => $INLINE_ARCHIVE,
   "Publish" => $PUBLISH,
   "RequestCardinality" => $REQ_CARDINALITY
  );
&override("/tmp/inner.cfg", %inner_vals);

#  Run experiments to collect data.
my $update_size = $UPDATE_SIZE_START;
my $update_size_comparator = '<=';
if ($UPDATE_SIZE_OP eq '-' || $UPDATE_SIZE_OP eq '/')
{
    $update_size_comparator = '>=';
}
while (eval("$update_size $update_size_comparator $UPDATE_SIZE_END"))
{
    my $dobj_size = $update_size * 2;
    if ($dobj_size > $update_size + 512 * 1024) {
	$dobj_size = $update_size + 512 * 1024;
    }

    my $test_stages = $TEST_STAGES_START;
    my $test_stages_comparator = '<=';
    if ($TEST_STAGES_OP eq '-' || $TEST_STAGES_OP eq '/')
    {
        $test_stages_comparator = '>=';
    }
    while (eval("$test_stages $test_stages_comparator $TEST_STAGES_END"))
    {
        #  Create configuration files.
        my $master_cfg  =
          &write_master_config_file($test_stages, $MEAS_LENGTH); #, #TEST_LENGTH);
        my @tester_cfgs =
          &write_tester_config_files($TEST_MACHINES, $test_stages,
                                     $dobj_size, $update_size,
				     $WARM_UPDATES, $WARM_PAUSE,
				     $MEAS_LENGTH, $MEAS_PAUSE,
				     $SETTLE);
        my $exp_cfg     =
          &write_experiment_file($TEST_MACHINES);
        my $tput_fork_log = "/tmp/tput_fork_log.$$.log";

        #  Run the experiment.
        print("Running tput test: $key_cmd ./run-experiment $exp_cfg\n" .
              "  TestMachines = $TEST_MACHINES\n" .
              "  TestStages   = $test_stages\n" .
              "  KeySize      = $KEY_SIZE\n" .
              "  DobjSize     = $dobj_size\n" .
              "  ReadSize   = $update_size\n" .
              "  TestLength   = $MEAS_LENGTH\n"); # $TEST_LENGTH\n");

        my $pid = &fork_process($tput_fork_log,
                                "./run-experiment", $exp_cfg);
        if(! &returned_zero($pid))
        {
            die("Experiment returned non-zero.\n");
        }

        #  Read the result log.
        my $total_ops = &analyze_result_log($pid);

        #  Compute the results.
        my $ops_tput  = $total_ops / $MEAS_LENGTH; #/ $TEST_LENGTH;
        my $data_tput = ($total_ops * $update_size) / $MEAS_LENGTH; # / $TEST_LENGTH;
        print("  OpsTput      = $ops_tput\n" .
              "  DataTput     = $data_tput\n");

        #  Write to log.
        &write_result_to_log($SKIP_ARCHIVE, $INLINE_ARCHIVE,
                             $TEST_MACHINES, $test_stages, $MEAS_LENGTH, # $TEST_LENGTH,
                             $KEY_SIZE, $dobj_size, $update_size,
                             $total_ops, $ops_tput, $data_tput);

        #  Remove the files we created.
        unlink($master_cfg, @tester_cfgs, $exp_cfg, $tput_fork_log);

        $test_stages = eval("$test_stages $TEST_STAGES_OP $TEST_STAGES_SKIP");
    }

    $update_size = eval("$update_size $UPDATE_SIZE_OP $UPDATE_SIZE_SKIP");
}

#  Remove the inner ring configuration file.
unlink("/tmp/inner.cfg");

exit(0);


sub parse_phase_desc
{
    my $desc_in = shift(@_);

    my $updates = 0;
    my $pause   = 0;

    if ($desc_in =~ /(\d+)(\/(\d+))?/ )
    {
        $updates = $1;
        if (defined($3)) { $pause = $3; }
    }

    my @desc;
    push(@desc, $updates);
    push(@desc, $pause);

    return @desc;
}

sub parse_range_spec
{
    my $spec = shift(@_);

    my $start = 0;
    my $end   = 0;
    my $op    = '+';
    my $skip  = 1;

    #  (\d+)                  The start of the range...
    #  (-(\d+))?              ...optionally a '-' and the end of the range...
    #  (\/([-\+\*\/])?(\d+))? ...optionally a '/', an optional 
    #                         operation (-, +, *, or /) and the skip value.
    if ($spec =~ /(\d+)(-(\d+))?(\/([-\+\*\/])?(\d+))?/ )
    {
        $start = $1;
        $end   = $start;
        if (defined($3)) { $end = $3; }
        if (defined($5)) { $op = $5; }
        if (defined($6)) { $skip = $6; }
    }

    #print("start=$start \n");
    #print("end=$end   \n");
    #print("op=$op    \n");
    #print("skip=$skip  \n");

    my @range;
    push(@range, $start);
    push(@range, $end);
    push(@range, $op);
    push(@range, $skip);

    return @range;
}

sub write_result_to_log
{
    my $skip          = shift(@_);
    my $inline        = shift(@_);
    my $test_machines = shift(@_);
    my $test_stages   = shift(@_);
    my $test_length   = shift(@_);
    my $key_size      = shift(@_);
    my $dobj_size     = shift(@_);
    my $update_size   = shift(@_);
    my $total_ops     = shift(@_);
    my $ops_tput      = shift(@_);
    my $data_tput     = shift(@_);

    my $log =
      "SkipArchive=${skip} " .
      "InlineArchive=${inline} " .
      "TestMachines=${test_machines} " .
      "TestStages=${test_stages} " .
      "TestLength=${test_length}s " .
      "KeySize=${key_size} " .
      "DobjSize=${dobj_size} " .
      "ReadSize=${update_size} " .
      "TotalOps=${total_ops} " .
      "ReadTput=${ops_tput}ops/s " .
      "DataTput=${data_tput}bytes/s\n";

    open(RESULT_LOG, ">> $TPUT_RESULT_LOG") ||
      die("Could not open result log ($TPUT_RESULT_LOG).");
    print(RESULT_LOG "$log");
    close(RESULT_LOG);

    return;
}

sub analyze_result_log
{
    my $pid = shift(@_);

    my @internal_logs = glob("/tmp/*-${pid}*-Master-internal.log");
    (scalar(@internal_logs) == 1) ||
      die("Cannot determine result log name.\n");

    my $log = shift(@internal_logs);
    open(RESULT_LOG, "$log") || die("Could not open result log.\n");
    my $total_ops = 0;
    while(<RESULT_LOG>)
    {
        if (/NumOps=(.*)/)
        {
            $total_ops += $1;
        }
    }

    close(RESULT_LOG);
    unlink($log) || die("Could not remove result log.\n");

    return $total_ops;
}

sub write_experiment_file
{
    my $test_machines = shift(@_);

    my $exp_cfg = "/tmp/tput.$$.exp";
    open(EXP, ">$exp_cfg") ||
      die("Could not create experiment configuration file ($exp_cfg).\n");
    my $old_fd = select(EXP);
    my $hosts = "ROC";
    my $fed;
    my $ir0;
    my $ir1;
    my $ir2;
    my $ir3;
    my $master;
    my $storage0;
    my $storage1;
    my $storage2;
    my $storage3;
    my $storage4;
    my $storage5;
    my $storage6;
    my $storage7;
    my $storage8;
    my $storage9;
    my $storage10;
    my $storage11;
    my $storage12;
    my $storage13;
    my $storage14;
    my $storage15;
    my $storage16;
    my $storage17;
    my $storage18;
    my $storage19;
    my $storage20;
    my $storage21;
    my $storage22;
    my $storage23;
    my $storage24;
    my $storage25;
    my $storage26;
    my $storage27;
    my $storage28;
    my $storage29;
    my $storage30;
    my $storage31;
    if($LOCAL==1)
    {
      $hosts = "localhost";
    }
    else
    {
      $fed = $CLUSTER_MACHINE - 1;
      $ir0 = $CLUSTER_MACHINE + 1 - 1;
      $ir1 = $CLUSTER_MACHINE + 2 - 1;
      $ir2 = $CLUSTER_MACHINE + 3 - 1;
      $ir3 = $CLUSTER_MACHINE + 4 - 1;
      $master = $CLUSTER_MACHINE + 5 - 1;
      $storage0 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage1 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage2 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage3 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage4 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage5 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage6 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage7 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage8 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage9 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage10 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage11 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage12 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage13 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage14 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage15 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage16 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage17 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage18 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage19 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage20 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage21 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage22 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage23 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage24 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage25 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage26 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage27 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage28 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage29 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
      $storage30 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
      $storage31 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
    }
#     else
#     {
#       $fed = $CLUSTER_MACHINE - 1;
#       $ir0 = $CLUSTER_MACHINE + 1 - 1;
#       $ir1 = $CLUSTER_MACHINE + 2 - 1;
#       $ir2 = $CLUSTER_MACHINE + 3 - 1;
#       $ir3 = $CLUSTER_MACHINE + 4 - 1
#       $master = $CLUSTER_MACHINE + 5 - 1;
#       $storage0 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 0 - 1;
#       $storage1 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 1 - 1;
#       $storage2 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 2 - 1;
#       $storage3 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 3 - 1;
#       $storage4 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 4 - 1;
#       $storage5 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 5 - 1;
#       $storage6 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 6 - 1;
#       $storage7 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 7 - 1;
#       $storage8 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 8 - 1;
#       $storage9 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 9 - 1;
#       $storage10 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 10 - 1;
#       $storage11 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 11 - 1;
#       $storage12 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 12 - 1;
#       $storage13 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 13 - 1;
#       $storage14 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 14 - 1;
#       $storage15 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 15 - 1;
#       $storage16 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 16 - 1;
#       $storage17 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 17 - 1;
#       $storage18 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 18 - 1;
#       $storage19 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 19 - 1;
#       $storage20 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 20 - 1;
#       $storage21 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 21 - 1;
#       $storage22 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 22 - 1;
#       $storage23 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 23 - 1;
#       $storage24 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 24 - 1;
#       $storage25 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 25 - 1;
#       $storage26 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 26 - 1;
#       $storage27 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 27 - 1;
#       $storage28 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 28 - 1;
#       $storage29 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 29 - 1;
#       $storage30 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 30 - 1;
#       $storage31 = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $test_machines + 31 - 1;
#     }

    print <<EOExp;
hosts      $hosts
topology   NONE
Federation \${run_dir}/federation.cfg  fed     daemon $fed
RP         \${run_dir}/rp.cfg          static  daemon $fed
Ring0      /tmp/inner.cfg              static  daemon $ir0
Ring1      /tmp/inner.cfg              static  daemon $ir1
Ring2      /tmp/inner.cfg              static  daemon $ir2
Ring3      /tmp/inner.cfg              static  daemon $ir3
StorageServer0 \${run_dir}/storage.cfg     static  daemon $storage0
StorageServer1 \${run_dir}/storage.cfg     static  daemon $storage1
StorageServer2 \${run_dir}/storage.cfg     static  daemon $storage0
StorageServer3 \${run_dir}/storage.cfg     static  daemon $storage1
StorageServer4 \${run_dir}/storage.cfg     static  daemon $storage0
StorageServer5 \${run_dir}/storage.cfg     static  daemon $storage1
StorageServer6 \${run_dir}/storage.cfg     static  daemon $storage0
StorageServer7 \${run_dir}/storage.cfg     static  daemon $storage1
StorageServer8 \${run_dir}/storage.cfg     static  daemon $storage0
StorageServer9 \${run_dir}/storage.cfg     static  daemon $storage1
StorageServer10 \${run_dir}/storage.cfg    static  daemon $storage0
StorageServer11 \${run_dir}/storage.cfg    static  daemon $storage1
StorageServer12 \${run_dir}/storage.cfg    static  daemon $storage0
StorageServer13 \${run_dir}/storage.cfg    static  daemon $storage1
StorageServer14 \${run_dir}/storage.cfg    static  daemon $storage0
StorageServer15 \${run_dir}/storage.cfg    static  daemon $storage1
StorageServer16 \${run_dir}/storage.cfg    static  daemon $storage0
StorageServer17 \${run_dir}/storage.cfg    static  daemon $storage1
StorageServer18 \${run_dir}/storage.cfg    static  daemon $storage0
StorageServer19 \${run_dir}/storage.cfg    static  daemon $storage1
StorageServer20 \${run_dir}/storage.cfg    static  daemon $storage0
StorageServer21 \${run_dir}/storage.cfg    static  daemon $storage1
StorageServer22 \${run_dir}/storage.cfg    static  daemon $storage0
StorageServer23 \${run_dir}/storage.cfg    static  daemon $storage1
StorageServer24 \${run_dir}/storage.cfg    static  daemon $storage0
StorageServer25 \${run_dir}/storage.cfg    static  daemon $storage1
StorageServer26 \${run_dir}/storage.cfg    static  daemon $storage0
StorageServer27 \${run_dir}/storage.cfg    static  daemon $storage1
StorageServer28 \${run_dir}/storage.cfg    static  daemon $storage0
StorageServer29 \${run_dir}/storage.cfg    static  daemon $storage1
StorageServer30 \${run_dir}/storage.cfg    static  daemon $storage0
StorageServer31 \${run_dir}/storage.cfg    static  daemon $storage1
Creator         \${run_dir}/creator.cfg    static  daemon $fed
Master     /tmp/tput-master.$$.cfg     static  mortal $master
EOExp

    for (my $i = 0; $i < $test_machines; $i++)
    {
        my $tester_cfg = "/tmp/tput-tester-$i.$$.cfg";
        my $machine_num = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $i - 1;
        print("Tester$i $tester_cfg static  daemon $machine_num\n");
    }
    close(EXP);
    select($old_fd);

    return $exp_cfg;
}


sub write_master_config_file
{
    my $test_stages = shift(@_);
    my $test_length = shift(@_);

    my $master_cfg = "/tmp/tput-master.$$.cfg";
    open(MASTER, ">$master_cfg") ||
      die("Could not create master configuration file ($master_cfg).\n");
    my $old_fd = select(MASTER);
    &write_master_header();
    print("      NumTesters $test_stages\n");
    print("      TestLength $test_length\n");
    &write_master_footer();
    close(MASTER);
    select($old_fd);

    return $master_cfg;
}

sub write_tester_config_files
{
    my $test_machines = shift(@_);
    my $test_stages   = shift(@_);
    my $dobj_size     = shift(@_);
    my $update_size   = shift(@_);
    my $warm_updates = shift(@_);
    my $warm_pause   = shift(@_);
    my $meas_updates = shift(@_);
    my $meas_pause   = shift(@_);
    my $settle       = shift(@_);

    my @tester_config_files;
    for (my $i = 0; $i < $test_machines; $i++)
    {
        my $tester_cfg = "/tmp/tput-tester-$i.$$.cfg";
        push(@tester_config_files, $tester_cfg);
        open(TESTER, ">$tester_cfg") ||
          die("Could not create tester configuration file ($tester_cfg).\n");

        my $old_fd = select(TESTER);
        &write_tester_header();

        my $num_stages = int($test_stages / $test_machines);
        if ($i < ($test_stages % $test_machines))
        {
            $num_stages++;
        }

        my $master_machine = $CLUSTER_MACHINE + $MASTER_MACHINE_OFFSET;
        #  The port number is a magic number, we start at port 3630 for now.
        my $master = "x$master_machine.millennium.berkeley.edu:3630";
        for (my $j = 0; $j < $num_stages; $j++)
        {
	    &write_tester_stage($j, $master, $dobj_size, $update_size,
				$warm_updates, $warm_pause,
				$meas_updates, $meas_pause, $settle);
        }

        &write_tester_footer();
        close(TESTER);
        select($old_fd);
      }

    return @tester_config_files;
}

sub write_master_header
{
    print <<EOHeader;
<sandstorm>

<global>
  defaultThreadManager TPSTM   
  <TPSTM>
    governor false
    governorDelay 2000
    governorMaxThreads 10
  </TPSTM>

  <TPPTM>
    numCpus 4
    maxThreads 4
  </TPPTM>

  <profile>
    enable false
    delay 1000
    filename \${SSProfileFile}
    graphfilename \${SSGraphFile}
    graph false
    sockets false
  </profile>

  <initargs>
    node_id \${NodeID}
    profile false
  </initargs>
</global>

<stages>
  <TputMicroBenchmarkMaster>
    class ostore.apps.test.TputMicroBenchmarkMaster
    queueThreshold 1000
    <initargs>
        ResultLog \${InternalLog}
EOHeader
}

sub write_master_footer
{
    print <<EOFooter;
    </initargs>
  </TputMicroBenchmarkMaster>

  #  We need to have Network and Tapestry stages to generate the
  #  proper 'Ready' messages that the run-scripts look for.
  <Network>
    class ostore.network.Network
    queueThreshold 1000
    <initargs>
    </initargs>
  </Network>

  <Router>
    class ostore.tapestry.impl.Router
    queueThreshold 100000

    <initargs>
      pkey \${NodePublicKey}
      skey \${NodePrivateKey}
      dynamic_route \${TapestryMode}
    </initargs>
  </Router>

  <TClient>
    class ostore.tapestry.impl.StaticTClient
    queueThreshold 100000
    <initargs>
      federation \${FederationID}
      dynamic_route \${TapestryMode}
    </initargs>
  </TClient>
</stages>

</sandstorm>
EOFooter
}

sub write_tester_header
{
    print <<EOHeader;
<sandstorm>
  <global>
  defaultThreadManager TPSTM
    <TPSTM>
    governor false
    governorDelay 2000
    governorMaxThreads 10
    </TPSTM>

    <TPPTM>
    numCpus 4
    maxThreads 4
    </TPPTM>

    <profile>
      enable false
      delay 1000
      filename \${SSProfileFile}
      graphfilename \${SSGraphFile}
      graph false
      sockets false
    </profile>

    <initargs>
        node_id \${NodeID}
        Profile false
    </initargs>

    <threadPool>
        initialThreads 2
        minThreads 2
        maxThreads 2
        <sizeController>
            enable false
            autoMaxDetect false
            delay 1000
            threshold 10
            idleTimeThreshold 1000
        </sizeController>
    </threadPool>
  </global>

<stages>
  <Network>
    class ostore.network.Network
    queueThreshold 100000
    <initargs>
    </initargs>
  </Network>

  <Router>
    class ostore.tapestry.impl.Router
    queueThreshold 100000

    <initargs>
      pkey \${NodePublicKey}
      skey \${NodePrivateKey}
      dynamic_route \${TapestryMode}
    </initargs>
  </Router>

  <TClient>
    class ostore.tapestry.impl.StaticTClient
    queueThreshold 100000
    <initargs>
      federation \${FederationID}
      dynamic_route \${TapestryMode}
    </initargs>
  </TClient>

  <ClientStage>
    class ostore.client.ClientStage
    queueThreshold 100000
    <initargs>
      UseDynamicTapestry false
      ProvideApiServer   false
      LogEvents          false
      LogFilename        /tmp/log
      LogSampleRate      1
    </initargs>
  </ClientStage>

  <ReplicaStage>
    class ostore.replica.ReplicaStage
    queueThreshold 100000
    <initargs>
    </initargs>
  </ReplicaStage>

  <ReadStage>
    class ostore.read.ReadStage
    queueThreshold 100000
    <initargs>
    </initargs>
  </ReadStage>

  <DTree>
    class ostore.dtree.DTreeNode
    queueThreshold 100000
    <initargs>
    </initargs>
  </DTree>

  <RecoalesceBlkStage>
    class ostore.archive.RecoalesceBlkStage
    queueThreshold 100000
    <initargs>
      RandomSeed 102
      Verify false
      LogEvents	false
      LogFilename \${InternalLog}-RecoalesceBlkStage.log
      LogSampleRate	1
    </initargs>
  </RecoalesceBlkStage>

  <RequestorStage>
    class ostore.archive.RequestorStage
    queueThreshold 100000
    <initargs>
      RequestCardinality        $REQ_CARDINALITY
      CooperativeCaching        false
    </initargs>
  </RequestorStage>

  <Cache>
    class ostore.cache.MemoryCache
    queueThreshold 100000
    <initargs>
      CacheSize 314572800 # 300 MB 125829120 # 120 MB
      cachedir \${CacheDir}
      Publish false
    </initargs>
  </Cache>

EOHeader
}

sub write_tester_stage
{
    my $tester_num  = shift(@_);
    my $master      = shift(@_);
    my $dobj_size   = shift(@_);
    my $update_size = shift(@_);
    my $warm_updates = shift(@_);
    my $warm_pause   = shift(@_);
    my $meas_updates = shift(@_);
    my $meas_pause   = shift(@_);
    my $settle       = shift(@_);
    my $init_len = int(500 / (int ($update_size/8192) + 1));
    if ($init_len < 20) { 
	$init_len = 20;
    }

    print <<EOStage;
  <ReadTputMicroBenchmark$tester_num>
    class ostore.apps.test.ReadTputMicroBenchmark
    queueThreshold 1000
    <initargs>
      Master              $master
      TesterNumber        $tester_num
      InitLength          $init_len # $warm_updates
      InitPauseTime       $warm_pause
      SettleTime          $settle
      TestLength          $meas_updates
      TestPauseTime       $meas_pause
      DobjSize            $dobj_size
      ReplaceSize         $update_size

      rp_pkeyfile        \${RPPublicKey}
      principal_pkeyfile \${PrincipalPublicKey}
      principal_skeyfile \${PrincipalPrivateKey}
    </initargs>
  </ReadTputMicroBenchmark$tester_num>
EOStage
}

sub write_tester_footer
{
    print <<EOFooter;
</stages>
</sandstorm>
EOFooter
}
#  LocalWords:  cfg
