#!/usr/bin/perl -w

# stats.pl by Matt Welsh <mdw@cs.berkeley.edu>
# 
# Reads in a histogram of <val>, <count> values and prints some
# basic statistics about it. Also generates histograms.

use Getopt::Std;
use POSIX;
use stats;
use strict;
use vars qw($opt_s $opt_h $opt_c $opt_k $opt_f $opt_g);
sub usage {
  print STDERR "Usage: stats.pl [options] <datacolumn>\n";
  print STDERR "  Prints statistics on data from the given file.\n";
  print STDERR "  <datacolumn> is the column of data values.\n";
  print STDERR "     (First column is 1)\n\n";
  print STDERR " Options:\n";
  print STDERR "\t-f <file>\t Read data from given file\n";
  print STDERR "\t-c <column>\t Specify column with counts\n";
  print STDERR "\t-k <key>\t Only consider lines containing regexp <key>\n";
  print STDERR "\t-h <bucketsize>\t Produce histogram of data\n";
  print STDERR "\t-g\t Graph histogram (requires -h)\n";
  print STDERR "\t-s\t Line number to Start measure data.\n";
  print STDERR "\n";
  exit -1;
}

my $START = -1;
my $FNAME = "";
my $COUNTCOLUMN = 0;
my $KEY = "";
my $BUCKETSIZE = -1;
my $GRAPH = 0;

getopt('f:c:k:h:s');
if($opt_s) {
  $START = $opt_s;
} else {
  $START = 1;
}

if ($opt_f) {
  $FNAME = $opt_f; 
}

if ($opt_c) {
  $COUNTCOLUMN = $opt_c;
} else {
  $COUNTCOLUMN = 0;
}

if ($opt_k) {
  $KEY = $opt_k;
}

if ($opt_h) {
  $BUCKETSIZE = $opt_h;
}

if ($opt_g) {
  $GRAPH = 1;
}

if ($#ARGV == -1) {
  &usage;
}

my $DATACOLUMN = $ARGV[0];

if ($COUNTCOLUMN == 0) {
  print "Using data column $DATACOLUMN, no count column\n";
} else {
  print "Using data column $DATACOLUMN, count column $COUNTCOLUMN\n";
}
$DATACOLUMN--; $COUNTCOLUMN--;

my $filename = "";
my $filehandle;
if ($FNAME) {
  open(IN, $FNAME) || die "Can't open $FNAME\n";
  $filehandle = \*IN;
} else {
  $filehandle = \*STDIN;
}

my $mode = -1;
my $first = 1;
my $line = 0;
my $i = 0;
my $count = -1 ;
my $maxcount = -1;
my @data;
while (<$filehandle>) {
  $line++;
  next if /^\#/; chop;

  if ($KEY) {
    next if !(/$KEY/);
  }
  if ($first) {
    print "First line of data:\n\n$_\n\n";
    $first = 0;
  }
  next if($line <$START);
    
  my @vals = split;

  if ($COUNTCOLUMN == -1) {
    # 1 count per sample
    $data[$i++] = $vals[$DATACOLUMN];
  } else {
    $count = $vals[$COUNTCOLUMN];
    if ($count > $maxcount) {
      $maxcount = $count;
      $mode = $vals[$DATACOLUMN];
    }
    # Number of counts for each sample in countcolumn
    for (my $j = 0; $j < $count; $j++) {
      $data[$i++] = $vals[$DATACOLUMN];
    }
  }
}

close $filehandle;

my ($num, $min, $max, $mean, $median, $stddev, $fiveperc,
    $ninefiveperc) = &calc_stats(@data);

print "$num data points\n";
print "min: $min\n";
print "max: $max\n";
print "mean: $mean\n";
print "stddev: $stddev\n";
print "median: $median\n";
print "5percentile: $fiveperc\n";
print "95percentile: $ninefiveperc\n";
if ($COUNTCOLUMN != -1) {
  print "mode: $mode with count $maxcount\n";
}

if ($BUCKETSIZE>0) {
  my $print = 1;
  my $filename = "/tmp/stats-hist.$$";
  my $title = "Histogram of $FNAME column $DATACOLUMN count $COUNTCOLUMN";
  my $save = 0;

  my @histogram = &calc_histogram($BUCKETSIZE, @data);
  &graph_histogram($title, $filename, $GRAPH, $save, $print, 
		   $BUCKETSIZE, @histogram);
}

