/**
 * TabbedDialog.java 
 *
 * Copyright (c) 2002 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package Sonar.Dialog;

import javax.swing.*;
import java.awt.*;
import Sonar.util.*;
import Sonar.Dialog.*;

/**
 * This is a standard dialog with a tabbed pane on it which can display
 * as many ActivePanels as you like. It is used for displaying both Node 
 * and Edge information.
 *
 * It is created by DisplayManager when a node or edge is clicked.
 * Each of the PacketAnalyzers provides an ActivePanel which is displayed
 * on a unique tab on the dialog
 *
 * When the dialog is closed, it applys all the changes in the ActivePanels.
 * When it is cancelled, it just closes.
 *
 * You must set the modal properties and title of the dialog manually.
 * Otherwise, just use the same way as standard dialog.
 *
 * @author	Kamin Whitehouse <kamin@cs.berkeley.edu>
 */
public class TabbedDialog extends javax.swing.JDialog
{
    //-----------------------------------------------------------------------
    //CONSTRUCTORS
    public TabbedDialog(Frame parent)
    {
	super(parent);

	// This code is automatically generated by Visual Cafe when you add
	// components to the visual environment. It instantiates and initializes
	// the components. To modify the code, only use code syntax that matches
	// what Visual Cafe can generate, or Visual Cafe may be unable to back
	// parse your Java file into its visual environment.
	//{{INIT_CONTROLS
	setDefaultCloseOperation(javax.swing.JFrame.DISPOSE_ON_CLOSE);
	setModal(true);
	setTitle("DialogTitle");
	getContentPane().setLayout(null);
	setSize(312,359);
	setVisible(false);
	applyButton.setNextFocusableComponent(CancelButton);
	applyButton.setText("apply");
	applyButton.setActionCommand("apply");
	getContentPane().add(applyButton);
	applyButton.setBounds(36,324,105,30);
	CancelButton.setText("Close");
	CancelButton.setActionCommand("Close");
	getContentPane().add(CancelButton);
	CancelButton.setBounds(168,324,102,30);
	JTabbedPane1.setOpaque(true);
	getContentPane().add(JTabbedPane1);
	JTabbedPane1.setBounds(0,0,312,324);
	//}}

	//{{REGISTER_LISTENERS
	SymAction lSymAction = new SymAction();
	applyButton.addActionListener(lSymAction);
	CancelButton.addActionListener(lSymAction);
	//}}
    }

    public TabbedDialog()
    {
	this((Frame)null);
    }

    public TabbedDialog(String sTitle)
    {
	this();
	setTitle(sTitle);
    }
    //CONSTRUCTORS
    //-----------------------------------------------------------------------

    public void setVisible(boolean b)
    {
	if (b)
	    setLocation(50, 50);
	super.setVisible(b);
    }

    //-----------------------------------------------------------------------
    //APPLY CHANGES
    //this is what happens when "OK" is clicked
    public void applyChanges()
    {
	ActivePanel currentPanel;
	int numTabs = JTabbedPane1.getTabCount();
	for(int count = 0; count < numTabs; count++)
	{
	    currentPanel = (ActivePanel)JTabbedPane1.getComponentAt(count);
	    currentPanel.applyChanges();
	}
    }
    //APPLY CHANGES
    //-----------------------------------------------------------------------


    //-----------------------------------------------------------------------
    //ADD ACTIVE PANEL
    public void addActivePanel(String name, ActivePanel pPanel)
    {
	if( (pPanel == null) ||
		(pPanel.getCancelInfoDialog() == true))	
	{
	    return;
	}
	pPanel.initializeDisplayValues();
	if(name == null)
	{
	    JTabbedPane1.add((Component)pPanel);
	}
	else
	{
	    JTabbedPane1.add(name, (Component)pPanel);
	}
    }
    //-----------------------------------------------------------------------
    //ADD ACTIVE PANEL

    public void addNotify()
    {
	// Record the size of the window prior to calling parents addNotify.
	Dimension size = getSize();

	super.addNotify();

	if (frameSizeAdjusted)
	    return;
	frameSizeAdjusted = true;

	// Adjust size of frame according to the insets
	Insets insets = getInsets();
	setSize(insets.left + insets.right + size.width, insets.top + insets.bottom + size.height);
    }

    // Used by addNotify
    boolean frameSizeAdjusted = false;

    //{{DECLARE_CONTROLS
    javax.swing.JButton applyButton = new javax.swing.JButton();
    javax.swing.JButton CancelButton = new javax.swing.JButton();
    javax.swing.JTabbedPane JTabbedPane1 = new javax.swing.JTabbedPane();
    //}}


    class SymAction implements java.awt.event.ActionListener
    {
	public void actionPerformed(java.awt.event.ActionEvent event)
	{
	    Object object = event.getSource();
	    if (object == applyButton)
		applyButton_actionPerformed(event);
	    else if (object == CancelButton)
		CancelButton_actionPerformed(event);
	}
    }

    void applyButton_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.

	applyButton_actionPerformed_Interaction1(event);
    }

    void applyButton_actionPerformed_Interaction1(java.awt.event.ActionEvent event)
    {
	try {
	    this.applyChanges();
	    //			this.dispose();		don't dispose the window so that we can keep working
	} catch (java.lang.Exception e) {
	}
    }

    void CancelButton_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.

	CancelButton_actionPerformed_Interaction1(event);
    }

    void CancelButton_actionPerformed_Interaction1(java.awt.event.ActionEvent event)
    {
	try {
	    this.dispose();
	} catch (java.lang.Exception e) {
	}
    }


}
