/**
 * GridLocationAnalyzer.java
 *
 * Copyright (c) 2002 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package Sonar.PacketAnalyzers;

import Sonar.*; 
import Sonar.event.*; 
import Sonar.util.*; 
import Sonar.Dialog.*; 

import java.util.*; 
import java.lang.*; 
import javax.swing.*; 
import java.awt.*; 
import java.awt.event.*; 
import java.io.*; 
import java.util.zip.*;

import ostore.util.NodeId;
import ostore.util.QSVector;
import ostore.util.QSString;


/**
 * GridLocationAnalyzer determines the placement of the nodes, based on
 * identification, simple matrix layout or a centralized viewpoint.
 *
 * ********************************************************************************************
 * This file is a template that shows the conventions on how to write a PacketAnalyzer subclass.
 * It shows you how to do the following things 
 * 1. receive and process a new data packet 
 * 2. recieve and process new node or edge clicks in the interface  
 * 3. recieve and process the creation or deletion of nodes or edges 
 * 4. Display a panel on the edge or node properties panels when they are clicked 
 * 5. add graphical output to the node, edge or the screen 
 * 6. Display a properties panel to edit parameters of this PacketAnalyzer 
 * 7. run this process in the background as a seperate thread 
 * 
 * If you want to write a new class to do packet analysis, you should 
 * not change any files besides this one, the MainFrame class (where you add menus) 
 * And the MainClass (where this class is instantiated).  This modularity 
 * will help maintain future compatibilty with other people's analyzers. 
 ************************************************************************************************
 * The main thing to understand here is that the class is highly mutltithreaded.
 * 1. It can spawn a thread in the constructor to run in the background (as shown below).
 * 2. It can spawn new threads every time an event is received (to free the eventGenerating thread)
 * 3. Every "EventReceived()" function is initially run on the thread of the object that generated
 *    the event. In this case, the following functions are initially run on the threads of the
 *    following objects: 
 *    a. packetReceived()			--> PacketListener
 *    b. nodeCreated/Deleted()			--> ObjectMaintainer
 *    c. nodeClicked(), edgeClicked()		--> DisplayManager 
 *    d. paintNode()/Edge()/Screen()		--> DisplayManager
 *    e. getProrietaryNode/EdgeDisplayPanel()	--> DisplayManager
 *
 * 4. All get/set functions may be called by the GUI thread.
 * 5. The constructor is called on the main() thread of the entire program.
 *************************************************************************************************
 * As a general rule for avoiding synchronization problems:
 * 1. All functions should be synchronized (except run() and the constructor) to eliminate
 *    problems with member variables (i.e. only one thread can be running functions of this
 *    object at a time).
 * 2. All threads in the constructor should be started at the end of the constructor to
 *    eliminate problems between that thread and the code in the constructor.
 * 3. Be careful about synchronizing over methods that call functions in other classes. It
 *    could cause problems such as starvation/deadlock (look in your OperatingSystem course
 *    book).
 * 4. Try not to call more than one synchronized method within the same call stack. (don't 
 *    have one thread synchronized on more than one object at a time).
 *
 * 
 * @author	Jason Lee <jlee81@cs.berkeley.edu> 
 * @author	Kamin Whitehouse <kamin@cs.berkeley.edu>
 */

public class GridLocationAnalyzer extends PacketAnalyzer implements LocationAnalyzer, GroupDialogContributor//, java.io.Serializable 
{ 
    //Define your member variables (try not to have publics) 
    protected Hashtable proprietaryNodeInfo; 
    protected TwoKeyHashtable proprietaryEdgeInfo; 
    protected Thread thread; 
    protected MenuManager menuManager; 
    //determines whether NodeInfo objects are deleted when nodes expire 
    protected boolean persistantLocations; 

    protected Vector group_leaders = new Vector();


    /* Layout of Nodes, upon a New Node or Identification Event
     * 0 - Matrix for New Nodes, then sort by Identification
     * 1 - Only matrix of new nodes, no sorting based on ID's
     * 2 - Tree for new nodes, (id disregarded), set at max number of groups
     * 3 - Circular chooses a central node, and then shows the number of hops to reach another node
     */
    public static final int ID_GROUPING = 0; //Planetlab nodes arranged on map?
    public static final int MATRIX = 1;
    public static final int TREE = 2;
    public static final int CIRCULAR = 3;
    protected int layoutMode = MATRIX;  //default setting 

    protected NodeInfo centered_node = null; //For CIRCULAR arrangement
    protected double center_X;
    protected double center_Y;
    protected boolean showGrids;   //Display grid or concentric circles
    protected boolean stickyGrids; //Always center the grid around the centralized node

    protected final double DEFAULT_X = 2.0;
    protected final double DEFAULT_Y = 1.7;

    protected PlanetLabAnalyzer plAnalyzer;
    protected DisplayManager displayManager;

    public int getLayoutMode() { return layoutMode; }

    public NodeId getCenteredNodeId() { return centered_node.getNodeId(); }

    public void updateGraph() {
	if(layoutMode == CIRCULAR) {
	    makeConcentric();
	} else if(layoutMode == TREE) {
	    makeTree();
	}
    }
    
    //------------------------------------------------------------------------ 
    //*****---CONSTRUCTOR---******//  
    //the constructor should be called by the MainClass constructor when  
    //it instantiates all the packetAnalyzers that we want 
    //Make sure you edit that constructor to do so 
    public GridLocationAnalyzer (DisplayManager dM)  
    { 
	plAnalyzer = new PlanetLabAnalyzer();
	displayManager = dM;
	
	persistantLocations = true;
	//this variable determines whether NodeInfo objects 
	//are deleted when nodes expire, i.e. whether their location's are discarded
	showGrids = true;
	stickyGrids = true;

	//create new hashtables for your proprietary data 
	proprietaryNodeInfo = new Hashtable(); 
	proprietaryEdgeInfo = new TwoKeyHashtable(); 

	//register to be notified of nodes and edges being created or deleted 
	//MainClass.objectMaintainer.addNodeEventListener(this);//listen to edge event 
	//register myself to be able to contribute to the node/edge properties panel 
	displayManager.addNodeDialogContributor(this); 

	displayManager.addGroupDialogContributor(this);
	//displayManager.addEdgeDialogContributor(this);

	//register myself to recieve NodeClickedEvents and EdgeClickedEvents 
	displayManager.addNodeClickedEventListener(this); 

	//register myself to paint nodes and edges and display info panels 
	displayManager.addNodePainter(this); //paint the nodes 
	//remember to change the default checkbox setting in the menuManager if you delete this line 
	displayManager.addEdgePainter(this); //likewise, but for painting edges
	//that need specific information about placement of nodes, such as CIRCULAR

	//register to contribute a panel to the SonarOptions panel 
	displayManager.mainFrame.addOptionsPanelContributor(this); 

	menuManager = new MenuManager();//this creates the menu, adds it to the MainMenuBar,
	//and creates and object to handle all menu events 

    } 
    //*****---CONSTRUCTOR---******// 
    //------------------------------------------------------------------------ 



    //************************************************************************ 
    //************************************************************************ 
    //the following four functions correspond with the  
    //NodeEventListener and EdgeEventListener interfaces and only work if you  
    //register as a listener for these events as show in the constructor. 
    //Note that becuase of the order the packetAnalyzers were created 
    //in the MainClass cosntructor, the ObjectMaintainer will always get 
    //a packet before all other PacketAnalyzers.  
    //************************************************************************ 
    //************************************************************************ 


    //------------------------------------------------------------------------ 
    //*****---Node Created---******// 
    //this function defines what you do when a new node is created 
    //It is called by Sonar.PacketAnalyzers.ObjectMaintainer[OStore] 
    public synchronized void nodeCreated(NodeEvent e) 
    { 
	NodeId node_id = e.getNodeId();
	NodeInfo node_info = e.getNodeInfo(); //new NodeInfo(node_id);

	//unless it already exists (it might exist if you don't delete it in NodeDeleted() 
	if(!proprietaryNodeInfo.containsKey(node_id))
	{ 
	    proprietaryNodeInfo.put(node_id, node_info); 
	}

	//layoutMode
	if(layoutMode == ID_GROUPING) {
	    groupNodesById();
	} else if(layoutMode == MATRIX) {
	    localizeNodes(); 
	} else if(layoutMode == TREE) {
	    makeTree();
	} else if(layoutMode == CIRCULAR) {
	    makeConcentric();
	}

	//Arbitrarily decided by the first node
	if(proprietaryNodeInfo.size() == 1) {
	    centered_node = node_info; 
	    center_X = node_info.getX();
	    center_Y = node_info.getY();
	}
    } 
    //*****---Node Created---******// 
    //------------------------------------------------------------------------ 


    //------------------------------------------------------------------------ 
    //*****---Node Deleted---******// 
    //this function defines what you do when a new node is deleted 
    //It is called by Sonar.PacketAnalyzers.ObjectMainter 
    public synchronized void nodeDeleted(NodeEvent e) 
    { 
	if(!persistantLocations)//if we don't delete the NodeInfo objects, they will be stored and used again when the node appears again.  Otherwise, all old data will be thrown away 
	{
	    NodeId deletedNode = e.getNodeId();
	    //Integer deletedNode = e.getNodeId();//you probably want to delete the info pbject to track the data of this new node 
	    proprietaryNodeInfo.remove(deletedNode);//but you might also want to leave it there but disable it, unless this node reappears and you want to use the same info 
	}

	if(proprietaryNodeInfo.size() == 0) {
	    centered_node = null;
	}

	//LocalizeNodes(); 
    } 
    //*****---Node Deleted---******// 
    //------------------------------------------------------------------------ 


    //************************************************************************ 
    //************************************************************************ 
    //the following two functions correspond to the  
    //NodeClickedEventListener, EdgeClickedEventListener interfaces and will 
    //only work if you register as a listener for these events 
    //************************************************************************ 
    //************************************************************************ 


    //------------------------------------------------------------------------ 
    //*****---Edge Clicked---******// 
    //this function defines what you do when a node is clicked 
    //It is called by Sonar.DisplayManager 
    /*    public synchronized void NodeClicked(NodeClickedEvent e) 
	  { 
	  Integer nodeClicked = e.getNodeId(); 
    //and maybe do some other processing 
    }
    */ 
    //*****---Node Clicked---******// 
    //------------------------------------------------------------------------ 

    //------------------------------------------------------------------------ 
    //*****---NODE DRAGGED---******// 
    //this function defines what you do when a node is clicked 
    //It is called by Sonar.DisplayManager 
    public void nodeDragged(NodeDraggedEvent e) 
	//public synchronized void nodeDragged(NodeDraggedEvent e) 
    { 
	NodeId nodeDragged = e.getNodeId(); 
	NodeInfo selectedNode = null;

	if(nodeDragged != null) {
	    selectedNode = (NodeInfo)proprietaryNodeInfo.get(nodeDragged); 
	}

	if(selectedNode!=null) 
	{     //this function acts on a custom mouse drag, and moves the node with the drag  
	    selectedNode.setX(e.getDraggedToX()); 
	    selectedNode.setY(e.getDraggedToY()); 
	    selectedNode.setFixed(true); 
	    displayManager.refreshScreenNow(); 
	} 
	//and maybe do some other processing 
    } 

    /*
     * Selects the first node found as the group leader. Obviously, each node
     * can specify its particular group leader, regardless of locality.
     *
     * This is only used for testing right now. displayGroupPropertyDialog
     * will handle the group icon and the election of a leader.
     */
    public void nodesSelected(NodesSelectedEvent e) {

	/*
	 // Handled by Visualization via the gateway node
	NodeInfo group_leader = null;
	NodeInfo node_info;
	//MainClass.out("nodes selected");

	for(Enumeration enumerator = e.getNodes().elements(); enumerator.hasMoreElements();) {
	    node_info = (NodeInfo) enumerator.nextElement();
	    //MainClass.out("nodes selected : "+ node_info.getNodeId().port());
	    if(group_leader == null) group_leader = node_info;
	    node_info.setGroupLeader(group_leader);
	}
	*/

    }

    //*****---Node Clicked---******// 
    //------------------------------------------------------------------------ 

    //************************************************************************ 
    //************************************************************************ 
    //the following two functions correspond to the  
    //NodeDialogContributor, EdgeDialogContributor interfaces and will 
    //only work if you register as a contributor as shown in the constructor. 
    //You must define the ProprietaryNodeInfoPanel and ProprietaryEdgeInfoPanel 
    //classes and they will automaticallyshow up when a node/edge is clicked 
    //by using the following two functions. 
    //************************************************************************ 
    //************************************************************************ 


    //------------------------------------------------------------------------ 
    //GET PROPRIETARY NODE INFO PANEL 
    //This function returns the Panel that you define it to retunr 
    //which will then automatically appear ina dialog when a node is clicked. 
    //this function is called by DisplayManager 
    public ActivePanel getProprietaryNodeInfoPanel(NodeId pNode) 
    { 
	NodeInfo nodeInfo = (NodeInfo)proprietaryNodeInfo.get(pNode); 
	if(nodeInfo==null)  
	    return null; 
	ProprietaryNodeInfoPanel panel = new ProprietaryNodeInfoPanel(nodeInfo); 
	return (ActivePanel)panel; 
    } 
    //GET PROPRIETARY NODE INFO PANEL 
    //------------------------------------------------------------------------ 

    public ActivePanel getProprietaryGroupInfoPanel(Vector group) {
	ProprietaryGroupInfoPanel panel = new ProprietaryGroupInfoPanel(group);
	if(panel != null) {
	    return panel;
	}
	return null;
    }

    public ActivePanel getProprietaryEdgeInfoPanel(NodeId source, NodeId dest) {
	return null;
    }


    //************************************************************************ 
    //************************************************************************ 
    //the following three functions correspond to the  
    //NodePainter, EdgePainter, and ScreenPainter interfaces and will 
    //only work if you register as a Painter as shown in the constructor. 
    //Whatever  painting function you implement here will be called every 
    //time a node or edge is painted, and after all the nodes/edges are painted 
    //the paintScreen functions are called.  You are called in the order that 
    //you register as a painter. 
    //************************************************************************ 
    //************************************************************************ 


    //------------------------------------------------------------------------ 
    //NODE PAINTER 
    //Put some function here to paint whatever you want over the node. 
    //The x1,y1 coordinates are the top left corner within which the node will be drawn 
    //The x2,y2 coordinates are the bottom right corner 
    //Paint everything on the graphics object 
    //this function is called by DisplayManager 
    public void paintNode(NodeId pNode, int x1, int y1, int x2, int y2, Graphics g)  
    { 
	NodeInfo nodeInfo = (NodeInfo)proprietaryNodeInfo.get(pNode); 
	if(nodeInfo==null) return; 

	if(nodeInfo.getDisplayCoords()) 
	{ 
	    String temp = String.valueOf(nodeInfo.getX()); 
	    String text = temp.substring(0,Math.min(4, temp.length())); 
	    text = text.concat(","); 
	    temp = String.valueOf(nodeInfo.getY()); 
	    text = text.concat(temp.substring(0,Math.min(4, temp.length()))); 
	    g.setColor(Color.black);	
	    g.drawString(text,  x1, y1); 
	} 
    } 
    //NODE PAINTER 
    //------------------------------------------------------------------------ 

    public void paintEdge(NodeId source, NodeId dest, int x1, int y1, int x2, int y2, Graphics g) {
	/*
	//DEBUGGING
	if(layoutMode == CIRCULAR) {
	    NodeInfo source_info = (NodeInfo) proprietaryNodeInfo.get(source);
	    NodeInfo dest_info = (NodeInfo) proprietaryNodeInfo.get(dest);

	    g.setColor(Color.black);

	    if((source_info.stratum + 1 == dest_info.stratum) || 
		    ((source_info.stratum == dest_info.stratum +1) && 
		     displayManager.getEdgeInfo(source, dest).isBidirectional())) { 
		g.drawLine(x1, y1, x2, y2);
	    }
	}
	*/

    }

    //Returns a random double to avoid a null pointer exception
    public double getCenterX() {
	if(centered_node != null) {
	    if(!stickyGrids) return center_X;
	    return centered_node.getX();
	} else {
	    return Math.random();
	}
    }

    public double getCenterY() {
	if(centered_node != null) {
	    if(!stickyGrids) return center_Y;
	    return centered_node.getY();
	} else {
	    return Math.random();
	}
    }

    public void resetAllNodesPosition() {

    }

    public void makeTree() {
	if(centered_node == null) return;

	//Arbitrarily set values
	centered_node.setX(DEFAULT_X); //(2.0, 0.0)
	centered_node.setY(0.0);
	centered_node.setPartitionNum(1.0); 

	NodeInfo info = null;
	Vector processed_nodes = new Vector(); //nodes, neighbors processed
	Vector stratum = new Vector(); //nodes are fixed, but neighbors are not
	Vector next_stratum;
	Vector list_neighbors;

	int current_stratum = 0;
	int iter = 0;
	NodeInfo next_info;
	//double num_per = (double) list_neighbors.size();
	stratum.addElement(centered_node);

	while(stratum.size() > 0 && processed_nodes.size() < proprietaryNodeInfo.size()) { 
	    iter = 0;
	    next_stratum = new Vector();
	    for(Enumeration e = stratum.elements(); e.hasMoreElements();) {
		info = (NodeInfo) e.nextElement();
		info.setX((double)iter / 4);
		info.setY((double)current_stratum / 4);

		list_neighbors = info.get_neighbors();
		//list_neighbors = MainClass.routingAnalyzer.getNeighbors(info.getNodeId());
		if(list_neighbors != null) {
		    for(Enumeration elm = list_neighbors.elements(); elm.hasMoreElements();) {
			next_info = (NodeInfo) proprietaryNodeInfo.get(elm.nextElement());
			if(next_info != null && !stratum.contains(next_info) && !next_stratum.contains(next_info) && !processed_nodes.contains(next_info)) {
			    //next_info.stratum = current_stratum + 1;
			    //next_info.num_on_partition = per_stratum;
			    next_stratum.addElement(next_info);
			}
			/*
			   if(current_stratum == 0) {
			   MainClass.out("GLA> (1):"+next_info.getNodeId().port());
			   }
			   */
		    }
		}
		processed_nodes.addElement(info);
		iter++;
	    }
	    stratum.clear();
	    stratum = (Vector) next_stratum.clone();
	    current_stratum++;
	}
	//MainClass.out("GLA> finished processing a dag");
    }

    //DisplayManager calls this function to display the underlying circles
    //The values are set according to the image width's - they should be scaled
    //accordingly
    public void paintGrids(int x, int y, int radius, Graphics g) {
	if(layoutMode != CIRCULAR || centered_node == null || !showGrids) return; 

	g.setColor(Color.black);
	//g.drawOval(x - radius/2, y-radius/2, radius, radius);
	g.drawOval(x - radius, y - radius, 2*radius, 2*radius);
	g.drawOval(x - radius * 2, y - radius * 2, 4 * radius, 4 * radius);
	g.drawOval(x - radius * 3, y - radius * 3, 6 * radius, 6 * radius);
	g.drawOval(x - radius * 4, y - radius * 4, 8 * radius, 8 * radius);
	//g.drawOval(x - radius * 5, y - radius * 5, 10 * radius, 10 * radius);
	//g.drawOval(x - radius * 6, y - radius * 6, 12 * radius, 12 * radius);
	//g.drawOval(x - radius * 7, y - radius * 7, 14 * radius, 14 * radius);

    }

    //Similar to Localize Nodes, but places the nodes in concentric circles
    //around the centered node
    public void makeConcentric() {
	if(centered_node == null) return;

	//Arbitrarily set values
	centered_node.setX(DEFAULT_X);
	centered_node.setY(DEFAULT_Y);
	centered_node.setStratum(0);
	centered_node.setPartitionNum(1.0);

	NodeInfo info = null;
	Vector processed_nodes = new Vector(); //neighbors already processed
	Vector stratum = new Vector(); //nodes fixed but neighbors are not yet
	Vector list_neighbors = centered_node.get_neighbors();
	//Vector list_neighbors = MainClass.routingAnalyzer.getNeighbors(centered_node.getNodeId());
	if(list_neighbors == null) return;

	int iter = 0;
	int current_stratum = 1;
	double num_per = (double) list_neighbors.size();
	double degree = 0.0; //in radians
	stratum.addElement(centered_node);

	Vector next_stratum;  
	int stratum_iter = 1;
	int per_stratum = 0;
	NodeInfo pred_info;
	while(stratum.size() > 0 && processed_nodes.size() < proprietaryNodeInfo.size()) {
	    next_stratum = new Vector();

	    for(Enumeration e = stratum.elements(); e.hasMoreElements();) {
		pred_info = (NodeInfo) e.nextElement(); 
		list_neighbors = pred_info.get_neighbors();
		if(list_neighbors != null) {
		    num_per = (double) list_neighbors.size();
		    iter = 1;
		    for(Enumeration elm = list_neighbors.elements(); elm.hasMoreElements();) {
			info = (NodeInfo) proprietaryNodeInfo.get(elm.nextElement());
			if(info!= null && !stratum.contains(info)
				&& !next_stratum.contains(info) && !processed_nodes.contains(info)) {
			    if(current_stratum == 1) {
				degree = 2 * Math.PI * ((double)iter) / num_per;
				info.setX(getCenterX() + (Math.cos(degree) * current_stratum *0.5 ));
				info.setY(getCenterY() + (Math.sin(degree) * current_stratum *0.5));
				info.setDegree(degree); 
				info.setPartitionNum(num_per);
				info.setStratum(current_stratum);

				//double-checking the values
				//degree = ((double)iter) / num_per * 360; 
				//MainClass.out(info.getNodeId().port()+" ("+info.getX()+","+info.getY()+
				//") at "+degree); 
			    } else {
				degree = (double)iter * 2 * Math.PI / (num_per * stratum.size())
				    - Math.PI / (double)stratum.size() + pred_info.getDegree();
				if(degree >= 2 * Math.PI) { degree -= 2 * Math.PI; }
				info.setX(getCenterX() + (Math.cos(degree)
					    * current_stratum * 0.5 ));
				info.setY(getCenterY() + (Math.sin(degree)
					    * current_stratum * 0.5));
				info.setDegree(degree);
				info.setPartitionNum(per_stratum); //num_per;
				info.setStratum(current_stratum); //For painting edges to layers one below
			    }
			    next_stratum.addElement(info);
			    //degree = degree * 180 / Math.PI;
			    //MainClass.out(pred_info.getNodeId().port()+"/"+info.getNodeId().port()
			    //	+"/"+iter+"/"+current_stratum+" ("+info.getX()+","+info.getY()
			    //	+") at "+degree);
			}
			iter++;
		    }
		}
		processed_nodes.addElement(pred_info);
		stratum_iter++;
	    }
	    stratum.clear();
	    stratum = (Vector) next_stratum.clone();
	    current_stratum++;
	}
	//MainClass.out("GLA> finished processing circular arrangement");

    }

    //Used with ID_GROUPING scheme, to place nodes by their purpose  
    public void identifyNode(NodeId node, int type) {
	if(!displayManager.rolePlaying) return;

	NodeInfo currentNodeInfo;
	if((currentNodeInfo= (NodeInfo)proprietaryNodeInfo.get(node)) != null) {
	    //If there is a type that actually changes then:
	    if((type > 0) && (currentNodeInfo.getNodeType() != DisplayManager.BOTH_IR_FS)) {
		if(layoutMode == ID_GROUPING) { 
		    groupNodesById();
		}
	    }
	}
    }

    public void groupNodesById() {
	
	//MainClass.out("GLA> grouping nodes by id");
	Enumeration nodes = proprietaryNodeInfo.elements();
	NodeInfo currentNodeInfo;

	Vector innerRings = new Vector(); //Inner Rings
	Vector storageServers = new Vector(); //Storage Servers	      
	Vector others = new Vector(); //Clients
	Vector clientAndRp = new Vector();
	// federation? rp?
	
	for(;nodes.hasMoreElements();) { 
	    currentNodeInfo = (NodeInfo)nodes.nextElement(); 

	    if((currentNodeInfo.getNodeType() == DisplayManager.INNER_RING) || (currentNodeInfo.getNodeType() == DisplayManager.BOTH_IR_FS)) {
		innerRings.addElement(currentNodeInfo);
	    } else if(currentNodeInfo.getNodeType() == DisplayManager.FILE_SERVER) {
		storageServers.addElement(currentNodeInfo);
	    } else if(currentNodeInfo.getNodeType() == DisplayManager.CLIENT || currentNodeInfo.getNodeType()== DisplayManager.RESP_PARTY) {
		clientAndRp.addElement(currentNodeInfo);
	    } else {
		others.addElement(currentNodeInfo);
	    }
	}

	double heightD = Math.sqrt(proprietaryNodeInfo.size());
	int heightI = 2 * (int)Math.ceil(heightD);

	if(heightI <= 5) { // Other space IR IR space Other
	    heightI = 6;
	}

	Enumeration nodesSS = storageServers.elements();
	Enumeration nodesIR = innerRings.elements();
	Enumeration otherNodes = others.elements();
	Enumeration nodesCRP = clientAndRp.elements();

	for(int r = 1; r <= heightI; r++) {
	    for(int c = 1; c <= heightI; c++) {
		if(nodesSS.hasMoreElements()) {
		    currentNodeInfo = (NodeInfo) nodesSS.nextElement();
		    currentNodeInfo.setX((double)c * 0.25); //(double) c/ (double) heightI); 
		    currentNodeInfo.setY((double)r * 0.25); //(double) r/ (double) heightI); 

		    if(!nodesSS.hasMoreElements()) {
			c = heightI;
		    }
		} else if(nodesIR.hasMoreElements()) {
		    if(c==1 || c == heightI) {
			if(otherNodes.hasMoreElements()) {
			    currentNodeInfo = (NodeInfo) otherNodes.nextElement();
			    currentNodeInfo.setX((double)c * 0.25); //(double) c/ (double) heightI); 
			    currentNodeInfo.setY((double)r * 0.25); //(double) r/ (double) heightI); 
			}	  
		    } else if(c != 2 && c != (heightI - 1)) {
			currentNodeInfo = (NodeInfo) nodesIR.nextElement();
			currentNodeInfo.setX((double)c * 0.25); //(double) c/ (double) heightI); 
			currentNodeInfo.setY((double)r * 0.25); //(double) r/ (double) heightI); 
			if(!nodesIR.hasMoreElements()) {
			    if(otherNodes.hasMoreElements()) {
				c = heightI - 1;
			    } else {
				c = heightI;
			    }
			}
		    }
		} else if(nodesCRP.hasMoreElements()) {
		    currentNodeInfo = (NodeInfo) nodesCRP.nextElement();
		    currentNodeInfo.setX((double)c * 0.25);//(double) c/ (double) heightI); 
		    currentNodeInfo.setY((double)r * 0.25);//(double) r/ (double) heightI); 
		    if(!otherNodes.hasMoreElements()) {
			c = heightI - 1;
		    }
		} else if(otherNodes.hasMoreElements()) {
		    currentNodeInfo = (NodeInfo) otherNodes.nextElement();
		    currentNodeInfo.setX((double)c * 0.25); //(double) c/ (double) heightI); 
		    currentNodeInfo.setY((double)r * 0.25); //(double) r/ (double) heightI); 
		}
	    }
	}
    }


    //------------------------------------------------------------------------ 
    //*****---LOCALIZE NODES---******// 
    public void localizeNodes() 
    { 
	//MainClass.out("GLA> localizing nodes");
	Enumeration nodes = proprietaryNodeInfo.elements(); 
	NodeInfo currentNodeInfo; 

	int numNodes = proprietaryNodeInfo.size(); 
	int numFixed = 0; 

	if(displayManager.mainFrame.getGraphDisplayPanel().planetlab ) {
	    for(;nodes.hasMoreElements();) { 
		currentNodeInfo = (NodeInfo)nodes.nextElement(); 
		plAnalyzer.setNodePosition(currentNodeInfo);
	    }
	    return;
	}

	/* TODO fixed nodes
	   for(;nodes.hasMoreElements();)
	//if we are not including the fixed nodes then subtract them from the total 
	{ 
	currentNodeInfo = (NodeInfo)nodes.nextElement(); 
	if(currentNodeInfo != null && currentNodeInfo.getFixed()) numFixed++; 
	} 
	numNodes -= numFixed; 
	*/

	double heightD = Math.sqrt(numNodes); 
	int heightI = (int)Math.ceil(heightD); 
	//MainClass.out("GLA> height"+heightI+" for nodes "+numNodes);
	for(int r = 1; r <= heightI; r++) 
	{ 
	    for(int c = 1; c <= heightI; c++)  
	    { 
		if(nodes.hasMoreElements()) 
		{	 
		    currentNodeInfo = (NodeInfo)nodes.nextElement(); 
		    //if we are not including the fixed nodes, then keep going until we find one that is not 
		    while(currentNodeInfo.getFixed() == true && nodes.hasMoreElements())
		    { 
			currentNodeInfo = (NodeInfo)nodes.nextElement(); 
		    }
		    if(!currentNodeInfo.getFixed()) {
			currentNodeInfo.setX((double) c/ (double) heightI); 
			currentNodeInfo.setY((double) r/ (double) heightI); 
		    }
		    //MainClass.out("GLA>"+ (c/heightI) + " " + (r/heightI));
		} 
	    } 
	} 
    } 
    //*****---LOCALIZE NODES---******// 
    //------------------------------------------------------------------------ 


    //public void setX(Integer node, double pX){}; 
    //public void setY(Integer node, double pY){};
    //public void setDistance(Integer sourceNode, Integer destinationNode, double pDistance){}; 

    public void setX (NodeId node_id, double pX) {};
    public void setY (NodeId node_id, double pY) {};
    public void setDistance(NodeId source_node_id, NodeId dest_node_id, double pDistance) {};


    //------------------------------------------------------------------------ 
    //*****---FIND NEAREST NODE---******// 
    public NodeId findNearestNode(double x, double y) 
    { 
	NodeInfo selectedNode = null; 
	NodeInfo tempNode = null; 
	double dist = Double.MAX_VALUE; 
	double bestdist = Double.MAX_VALUE; 
	double xDist, yDist; 
	for(Enumeration nodes = proprietaryNodeInfo.elements(); nodes.hasMoreElements();)  
	{ 
	    tempNode = (NodeInfo)nodes.nextElement(); 
	    xDist = Math.pow(tempNode.getX() - x,2.0); 
	    yDist = Math.pow(tempNode.getY() - y,2.0); 
	    dist = Math.sqrt(xDist + yDist); 
	    if (dist < bestdist) { 
		selectedNode = tempNode; 
		bestdist = dist; 
	    } 
	} 
	if(selectedNode == null) 
	{ 
	    return null; 
	} 
	else 
	{	 
	    return selectedNode.getNodeId(); 
	} 
    } 
    //*****---FIND NEAREST NODE---******// 
    //------------------------------------------------------------------------ 

    public Vector findAllNodesWithinRectangle (double x1, double y1, double x2, double y2) {
	Vector selectedNodes = null;
	NodeInfo tempNode = null;

	if(x1 > x2) {
	    double xtemp = x2;
	    x2 = x1;
	    x1 = xtemp;
	} 
	if(y1 > y2) {
	    double ytemp = y2;
	    y2 = y1;
	    y1 = ytemp;
	}

	for(Enumeration nodes = proprietaryNodeInfo.elements(); nodes.hasMoreElements();) {
	    tempNode = (NodeInfo) nodes.nextElement();
	    if(tempNode.getX() >= x1 && tempNode.getX() <= x2 &&
		    tempNode.getY() >= y1 && tempNode.getY() <= y2) {
		if(selectedNodes == null) 
		    selectedNodes = new Vector();

		//MainClass.out("*gla> included node: "+tempNode.getNodeId().port());
		selectedNodes.add(tempNode); //.getNodeId());
	    }
	}

	return selectedNodes;
    }


    //------------------------------------------------------------------------ 
    //*****---FIND NEAREST EDGE---******// 
    public Vector findNearestEdge(double x, double y) 
    { 
	EdgeInfo selectedEdge = null; 
	EdgeInfo tempEdge = null; 
	double dist = Double.MAX_VALUE; 
	double bestdist = Double.MAX_VALUE; 
	for(Enumeration edges = proprietaryEdgeInfo.elements(); edges.hasMoreElements();)  
	{ 
	    double x1, y1, x2, y2, xCenter, yCenter; 
	    tempEdge = (EdgeInfo)edges.nextElement(); 
	    NodeId sourceNode = tempEdge.getSource(); 
	    NodeId destinationNode = tempEdge.getDest(); 
	    x1 = getX(sourceNode); 
	    y1 = getY(sourceNode); 
	    x2 = getX(destinationNode); 
	    y2 = getY(destinationNode); 
	    if((x1==Double.NaN) || (y1==Double.NaN) || (x2==Double.NaN) || (y2==Double.NaN)) continue; 
	    xCenter= (x1 + x2)/2; 
	    yCenter= (y1 + y2)/2; 
	    dist = Math.sqrt(Math.pow(xCenter-x,2) + Math.pow(yCenter-y,2)); 
	    if (dist < bestdist)  
	    { 
		selectedEdge = tempEdge; 
		bestdist = dist; 
	    } 
	} 
	if(selectedEdge == null) 
	{ 
	    return null; 
	} 
	else 
	{	 
	    Vector edgeID = new Vector(); 
	    edgeID.add(selectedEdge.getSource()); 
	    edgeID.add(selectedEdge.getDest()); 
	    return edgeID; 
	} 
    } 	
    //*****---FIND NEAREST EDGE---******// 
    //------------------------------------------------------------------------ 


    //------------------------------------------------------------------------ 
    //INTERFACE TO PROPRIETARY DATA 
    //write some functions here that will let other Analyzers find and user your data 
    public NodeInfo getNodeInfo(NodeId node_id){return (NodeInfo)proprietaryNodeInfo.get(node_id);} 
    public EdgeInfo getEdgeInfo(NodeId source, NodeId dest){return (EdgeInfo)proprietaryEdgeInfo.get(source,dest);} 
    public Enumeration getNodeInfo(){return proprietaryNodeInfo.elements();} 
    public Enumeration getEdgeInfo(){return proprietaryEdgeInfo.elements();} 
    public double getDistance(NodeId source, NodeId dest ) 
    { 
	/*	    	EdgeInfo edgeInfo = (EdgeInfo)proprietaryEdgeInfo.get(sourceNode, destinationNode); 
			if(edgeInfo != null) 
			{ 	
			return edgeInfo.getDistance(); 
			} 
			else 
			{*/ 
	return Double.NaN; 
	//  	  	} 
    } 

    public double getX(NodeId node_id ) 
    { 
	NodeInfo nodeInfo = (NodeInfo)proprietaryNodeInfo.get(node_id); 
	if(nodeInfo !=null) { 
	    return nodeInfo.getX(); 
	} 
	else { 
	    return Double.NaN; 
	} 
    } 	
    public double getY(NodeId node_id) 
    { 
	NodeInfo nodeInfo = (NodeInfo)proprietaryNodeInfo.get(node_id); 
	if(nodeInfo !=null) { 
	    return nodeInfo.getY(); 
	} 
	else { 
	    return -1; 
	} 
    } 	
    //INTERFACE TO PROPRIETARY DATA 
    //------------------------------------------------------------------------ 





    //------------------------------------------------------------------------ 
    //*****---SHOW PROPERTIES DIALOG---******// 
    //this function can be called by MainFrame (by the menus, in particular) 
    //and should simply show the dialog as shown here. 
    //You need to define the class "GridLocationAnalyzerPropertiesPanel" 
    //in order for this to do anything.  it is useful for setting parameters 
    //on your analyzer. 
    public void ShowOptionsDialog()  
    { 
	StandardDialog newDialog = new StandardDialog(new OptionsPanel()); 
	newDialog.show(); 
    } 

    public ActivePanel getOptionsPanel()  
    { 
	return new OptionsPanel(); 
    } 
    //*****---SHOW PROPERTIES DIALOG---******// 
    //------------------------------------------------------------------------ 

    public class ProprietaryNodeInfoPanel extends Sonar.Dialog.ActivePanel 
    { 
	protected NodeInfo nodeInfo;

	public ProprietaryNodeInfoPanel(NodeInfo pNodeInfo) 
	{ 
	    tabTitle = "Location"; 
	    nodeInfo = pNodeInfo; 
	    //{{INIT_CONTROLS 
	    setLayout(null); 
	    Insets ins = getInsets(); 
	    setSize(247,168); 
	    JLabel3.setText("X Coordinate"); 
	    add(JLabel3); 
	    JLabel3.setBounds(36,48,84,24); 
	    JLabel4.setText("Y Coordinate"); 
	    add(JLabel4); 
	    JLabel4.setBounds(36,72,75,24); 
	    JTextField1.setNextFocusableComponent(JTextField2); 
	    JTextField1.setToolTipText("The scale of the coordinate system is determined by the user, and scaled automatically by the system to fit to the screen"); 
	    JTextField1.setText("1.5"); 
	    add(JTextField1); 
	    JTextField1.setBounds(120,48,87,18); 
	    JTextField2.setToolTipText("The scale of the coordinate system is determined by the user, and scaled automatically by the system to fit to the screen"); 
	    JTextField2.setText("3.2"); 
	    add(JTextField2); 
	    JTextField2.setBounds(120,72,87,18); 
	    JCheckBox1.setToolTipText("Check this is you don\'t want the node to move around"); 
	    JCheckBox1.setText("Fixed x/y Coordinates"); 
	    add(JCheckBox1); 
	    JCheckBox1.setBounds(36,96,168,24); 
	    JCheckBox2.setToolTipText("Check this is you want the coordinates to be displayed on the screen"); 
	    JCheckBox2.setText("Display x/y Coordinates"); 
	    add(JCheckBox2); 
	    JCheckBox2.setBounds(36,120,168,24); 

	    //Can not un-"center" a node!
	    if(layoutMode == CIRCULAR || layoutMode == TREE) {
		if(centered_node != nodeInfo) {
		    JCheckBox3.setToolTipText("Check this to center all other nodes around this one");
		    JCheckBox3.setText("Central node");
		    add(JCheckBox3);
		    JCheckBox3.setBounds(36, 144, 168, 24);
		} else {
		    JLabel5.setText("This is the central-ized node");
		    JLabel5.setBounds(36, 220, 168, 24);
		    add(JLabel5);
		}
	    }


	    //}} 

	    //{{REGISTER_LISTENERS 
	    //}} 
	} 

	//{{DECLARE_CONTROLS 
	javax.swing.JLabel JLabel3 = new javax.swing.JLabel(); 
	javax.swing.JLabel JLabel4 = new javax.swing.JLabel(); 
	javax.swing.JTextField JTextField1 = new javax.swing.JTextField(); 
	javax.swing.JTextField JTextField2 = new javax.swing.JTextField(); 
	javax.swing.JCheckBox JCheckBox1 = new javax.swing.JCheckBox(); 
	javax.swing.JCheckBox JCheckBox2 = new javax.swing.JCheckBox(); 
	javax.swing.JCheckBox JCheckBox3 = new javax.swing.JCheckBox();
	javax.swing.JLabel JLabel5 = new javax.swing.JLabel();
	//}} 

	public void applyChanges() 
	{ 
	    nodeInfo.setX(Double.valueOf(JTextField1.getText()).doubleValue()); 
	    nodeInfo.setY(Double.valueOf(JTextField2.getText()).doubleValue()); 
	    nodeInfo.setFixed(JCheckBox1.isSelected()); 
	    nodeInfo.setDisplayCoords(JCheckBox2.isSelected());
	    if(JCheckBox3.isSelected()) {
		centered_node = nodeInfo;
		if(layoutMode == CIRCULAR) {
		    makeConcentric();
		} else if(layoutMode == TREE) {
		    makeTree();
		}
	    }
	} 

	public void initializeDisplayValues() 
	{ 
	    JTextField1.setText(String.valueOf(nodeInfo.getX())); 
	    JTextField2.setText(String.valueOf(nodeInfo.getY())); 
	    JCheckBox1.setSelected(nodeInfo.getFixed()); 
	    JCheckBox2.setSelected(nodeInfo.getDisplayCoords()); 

	    JCheckBox3.setSelected(nodeInfo==centered_node);
	} 
    } 

    public class ProprietaryGroupInfoPanel extends Sonar.Dialog.ActivePanel 
    {
	Vector group;
	Hashtable ports_to_nodes;

	javax.swing.JComboBox combobox1 = new javax.swing.JComboBox();
	javax.swing.JLabel label1 = new javax.swing.JLabel();

	public ProprietaryGroupInfoPanel(Vector nodes) {
	    group = nodes;
	    //MainClass.out("PropGroupInfoPanel: "+nodes.size());
	    ports_to_nodes = new Hashtable();

	    tabTitle = "Location";
	    setLayout(null);
	    Insets ins = getInsets();
	    setSize(259, 279);

	    NodeInfo node;
	    for(Enumeration e = group.elements(); e.hasMoreElements();) {
		node = (NodeInfo) e.nextElement();
		combobox1.addItem(Integer.toString(node.getNodeId().port()));
		ports_to_nodes.put(Integer.toString(node.getNodeId().port()), node); 
	    }
	    add(combobox1);
	    combobox1.setBounds(20, 34, 210, 24);
	}

	public void initializeDisplayValue() {

	}

	public void applyChanges () {
	    //Find the matching nodeinfo to the port
	    NodeInfo leader = (NodeInfo) ports_to_nodes.get(combobox1.getSelectedItem());
	    QSVector children = new QSVector();

	    if(!group_leaders.contains(leader)) 
		group_leaders.add(leader);

	    //MainClass.out("children size is "+children.size());

	    if(leader != null) {
		for(Enumeration e = group.elements(); e.hasMoreElements();) {
		    NodeInfo node = (NodeInfo) e.nextElement();
		    //MainClass.out("GLA (1065):"+node.getNodeId());
		    node.setGroupLeader(leader);

		    if(!leader.getNodeId().equals(node.getNodeId())) {
			//MainClass.out("child node :"+node.getNodeId().toString());
			children.add(node.getNodeId());
		    }
		}
	    }
	    //MainClass.out("sending a group message with :"+children.size());
	    Thread.dumpStack();
	    ostore.util.Carp.die("deprecated");
	    
	    //Old code to manually group together nodes
	    //MainClass.packetSender.sendMessage(new SonarGroupMessage(leader.getNodeId(), new ostore.util.SHA1Hash(), new ostore.util.SHA1Hash(), SonarGroupMessage.OPEN, children));
	    
	    //temp
	    /*
	       QSVector rout_msgs = new QSVector();
	       rout_msgs.add(new QSString("ostore.rp.CreateObjectReqMsg"));
	       rout_msgs.add(new QSString("ostore.rp.InnerRingCreateReqMsg"));
	       rout_msgs.add(new QSString("ostore.rp.ObjectCreateReqMsg"));
	       rout_msgs.add(new QSString("ostore.agree.ForwardMsg"));
	       rout_msgs.add(new QSString("ostore.agree.ResultMsg"));
	       rout_msgs.add(new QSString("ostore.archive.dissemination.DisseminateFragMsg"));
	       */

	} 

    }

    public class ProprietaryEdgeInfoPanel extends Sonar.Dialog.ActivePanel 
    { 
	EdgeInfo edgeInfo; 

	public ProprietaryEdgeInfoPanel(EdgeInfo pEdgeInfo) 
	{ 
	    tabTitle = "Location"; 
	    edgeInfo = pEdgeInfo; 
	    //{{INIT_CONTROLS 
	    setLayout(null); 
	    Insets ins = getInsets(); 
	    setSize(247,168); 
	    JLabel3.setText("Distance"); 
	    add(JLabel3); 
	    JLabel3.setBounds(36,48,84,24); 
	    JTextField1.setToolTipText("The scale of the coordinate system is determined by the user, and scaled automatically by the system to fit to the screen"); 
	    JTextField1.setText("1.5"); 
	    add(JTextField1); 
	    JTextField1.setBounds(120,48,87,18); 
	    JCheckBox1.setToolTipText("Check this is you want the length of the edge to be displayed"); 
	    JCheckBox1.setText("Display Length"); 
	    add(JCheckBox1); 
	    JCheckBox1.setBounds(36,66,168,24); 

	    //}} 

	    //{{REGISTER_LISTENERS 
	    //}} 
	} 

	//{{DECLARE_CONTROLS 
	javax.swing.JLabel JLabel3 = new javax.swing.JLabel(); 
	javax.swing.JTextField JTextField1 = new javax.swing.JTextField(); 
	javax.swing.JCheckBox JCheckBox1 = new javax.swing.JCheckBox(); 

	//}} 

	public void applyChanges() 
	{ 
	    edgeInfo.setDistance(Double.parseDouble(JTextField1.getText())); 
	    edgeInfo.setDisplayLength(JCheckBox1.isSelected()); 
	} 

	public void initializeDisplayValues() 
	{ 
	    JTextField1.setText(String.valueOf(edgeInfo.getDistance())); 
	    JCheckBox1.setSelected(edgeInfo.getDisplayLength()); 
	} 

    } 


    //GridLocationAnalyzerPropertiesPanel 
    //This class is an ActivePanel and should have all the information 
    //in GUI form that this class stores with respect to EDGES 
    //It will be displayed automatically with ShowOptionsDialog 
    
    public class OptionsPanel extends Sonar.Dialog.ActivePanel 
    {		 
	public OptionsPanel() 
	{ 
	    tabTitle = "Location";//this will be the title of the tab 
	    //{{INIT_CONTROLS 
	    setLayout(null); 
	    //	Insets ins = getInsets(); 
	    setSize(307,220); 

	    JCheckBox1.setToolTipText("Check this is you want to remember the positions of the nodes even after they expire (in case they reappear)"); 
	    JCheckBox1.setText("Persistant Locations"); 
	    JCheckBox1.setBounds(36,20,168,24); 
	    add(JCheckBox1); 

	    /*
	       JCheckBox2.setToolTipText("Check this is you want a blank space in the grid for all nodes which you placed by hand"); 
	       JCheckBox2.setText("Include Manual Nodes"); 
	       JCheckBox2.setBounds(36,50,168,24); 
	       add(JCheckBox2); 
	       */

	    JLabel4.setText("Grouping");
	    JLabel4.setToolTipText("This selects the desired layout of the nodes");
	    JLabel4.setBounds(24, 100, 200, 24);
	    add(JLabel4);

	    JComboBox1.addItem(new String("ID Grouping")); //identify
	    JComboBox1.addItem(new String("Matrix only")); //Localize, into a matrix
	    JComboBox1.addItem(new String("Tree")); //yet to be implemented
	    JComboBox1.addItem(new String("Circular")); //based on tap. edges
	    add(JComboBox1);
	    JComboBox1.setBounds(24, 130, 128, 24);

	    JCheckBox3.setToolTipText("Check this if you want to show the underlying grids or concentric circles");
	    JCheckBox3.setText("Show grids/circles");
	    JCheckBox3.setBounds(24, 154, 144, 24);
	    add(JCheckBox3);

	    JCheckBox4.setToolTipText("Check this if you want to make the grid 'stick'y to the centralized node");
	    JCheckBox4.setText("Attach to center");
	    JCheckBox4.setBounds(24, 174, 144, 24);
	    add(JCheckBox4);
	    //}} 

	    //{{REGISTER_LISTENERS 
	    //}} 
	} 

	//{{DECLARE_CONTROLS 
	javax.swing.JCheckBox JCheckBox1 = new javax.swing.JCheckBox(); 
	javax.swing.JLabel JLabel4 = new javax.swing.JLabel();
	javax.swing.JComboBox JComboBox1 = new javax.swing.JComboBox();
	javax.swing.JCheckBox JCheckBox3 = new javax.swing.JCheckBox();
	javax.swing.JCheckBox JCheckBox4 = new javax.swing.JCheckBox();
	//}} 
    
	public void applyChanges()//this function will be called when the apply button is hit 
	{ 
	    persistantLocations = JCheckBox1.isSelected(); 
	    showGrids = JCheckBox3.isSelected();
	    stickyGrids = JCheckBox4.isSelected();

	    if(JComboBox1.getSelectedItem().equals("ID Grouping")) {
		layoutMode = ID_GROUPING;
		groupNodesById();
	    } else if(JComboBox1.getSelectedItem().equals("Matrix only")) {
		layoutMode = MATRIX;
		localizeNodes();
	    } else if(JComboBox1.getSelectedItem().equals("Tree")) {
		layoutMode = TREE;
		makeTree();
	    } else if(JComboBox1.getSelectedItem().equals("Circular")) {
		layoutMode = CIRCULAR;
		makeConcentric();
	    }
	} 

	public void initializeDisplayValues()//this function will be called when the panel is first shown 
	{ 
	    JCheckBox1.setSelected(persistantLocations); 
	    JCheckBox3.setSelected(showGrids);
	    JCheckBox4.setSelected(stickyGrids);

	    switch(layoutMode) {
		case ID_GROUPING:
		    JComboBox1.setSelectedItem(new String("ID Grouping"));
		    break;
		case MATRIX:
		    JComboBox1.setSelectedItem(new String("Matrix only"));
		    break;
		case TREE:
		    JComboBox1.setSelectedItem(new String("Tree"));
		    break;
		case CIRCULAR:
		    JComboBox1.setSelectedItem(new String("Circular"));
		    break;
	    }
	} 
    } 	           
    //GridLocationAnalyzerPropertiesPanel 
    //********************************************************* 
    //********************************************************* 
    //********************************************************* 

    //********************************************************* 
    //********************************************************* 
    //********************************************************* 
    //MENU MANAGER 
    //This class creates and holds the menu that controls this 
    //PacketAnalyzer.  It returns the menu to whoever wants 
    //to display it and it also handles all events on the menu 
    protected class MenuManager implements Serializable, ActionListener, ItemListener 
    { 
	//{{DECLARE_CONTROLS 
	JMenu mainMenu = new JMenu(); 
	JCheckBoxMenuItem packetProcessingCheckBox = new JCheckBoxMenuItem(); 
	JCheckBoxMenuItem backgroundProcessingCheckBox = new JCheckBoxMenuItem(); 
	JSeparator separator1 = new JSeparator(); 
	JMenuItem propertiesItem = new JMenuItem(); 
	JSeparator separator2 = new JSeparator(); 

	JMenuItem circular = new JMenuItem();
	JMenuItem tree = new JMenuItem();
	JMenuItem group = new JMenuItem();
	JMenuItem matrix = new JMenuItem();
	JSeparator separator4 = new JSeparator();

	JMenu serializeMenu = new JMenu(); 
	JMenuItem saveNodesItem = new JMenuItem(); 
	JMenuItem loadNodesItem = new JMenuItem(); 
	JMenuItem saveEdgesItem = new JMenuItem(); 
	JMenuItem loadEdgesItem = new JMenuItem(); 
	JSeparator separator3 = new JSeparator(); 
	JMenu paintMenu = new JMenu(); 
	JCheckBoxMenuItem paintNodesItem = new JCheckBoxMenuItem(); 
	JCheckBoxMenuItem paintEdgesItem = new JCheckBoxMenuItem(); 
	JCheckBoxMenuItem paintScreenItem = new JCheckBoxMenuItem(); 
	//}} 

	public MenuManager() 
	{ 
	    //{{INIT_CONTROLS 
	    mainMenu.setText("Grid Location Analyzer"); 
	    mainMenu.setActionCommand("Grid Location Analyzer"); 
	    packetProcessingCheckBox.setSelected(true); 
	    packetProcessingCheckBox.setText("Packet Processing"); 
	    packetProcessingCheckBox.setActionCommand("Packet Processing"); 
	    mainMenu.add(packetProcessingCheckBox); 

	    backgroundProcessingCheckBox.setSelected(true); 
	    backgroundProcessingCheckBox.setText("Background Processing"); 
	    backgroundProcessingCheckBox.setActionCommand("Background Processing"); 
	    mainMenu.add(backgroundProcessingCheckBox); 
	    mainMenu.add(separator1); 

	    circular.setText("Centralized View");
	    circular.setActionCommand("Centralized View");
	    mainMenu.add(circular);
	    tree.setText("Tree");
	    tree.setActionCommand("Tree");
	    mainMenu.add(tree);
	    group.setText("Group by Ids");
	    group.setActionCommand("Group by Ids");
	    mainMenu.add(group);
	    matrix.setText("Matrix");
	    matrix.setActionCommand("Matrix");
	    mainMenu.add(matrix);
	    mainMenu.add(separator4);

	    propertiesItem.setText("Options"); 
	    propertiesItem.setActionCommand("Options"); 
	    mainMenu.add(propertiesItem); 
	    mainMenu.add(separator2); 
	    serializeMenu.setText("Serialize"); 
	    serializeMenu.setActionCommand("Serialize"); 
	    saveNodesItem.setText("Save Node Locations"); 
	    saveNodesItem.setActionCommand("Save Node Locations"); 
	    serializeMenu.add(saveNodesItem); 
	    loadNodesItem.setText("Load Node Locations"); 
	    loadNodesItem.setActionCommand("Load Node Locations"); 
	    serializeMenu.add(loadNodesItem); 
	    saveEdgesItem.setText("Save Edge Lengths"); 
	    saveEdgesItem.setActionCommand("Save Edge Lengths"); 
	    serializeMenu.add(saveEdgesItem); 
	    loadEdgesItem.setText("Load Edge Lengths"); 
	    loadEdgesItem.setActionCommand("Load Edge Lengths"); 
	    serializeMenu.add(loadEdgesItem); 
	    mainMenu.add(serializeMenu); 
	    mainMenu.add(separator3); 
	    paintMenu.setText("Painting"); 
	    paintMenu.setActionCommand("Painting"); 
	    paintNodesItem.setSelected(true); 
	    paintNodesItem.setText("Show Node Location Info"); 
	    paintNodesItem.setActionCommand("Show Node Location Info"); 
	    paintMenu.add(paintNodesItem); 
	    paintEdgesItem.setSelected(true); 
	    paintEdgesItem.setText("Show Edge Location Info"); 
	    paintEdgesItem.setActionCommand("Show Edge Location Info"); 
	    paintMenu.add(paintEdgesItem); 
	    paintScreenItem.setSelected(true); 
	    paintScreenItem.setText("Paint on Screen"); 
	    paintScreenItem.setActionCommand("Paint on Screen"); 
	    paintMenu.add(paintScreenItem); 
	    mainMenu.add(paintMenu); 
	    
	    //HACK
	    displayManager.mainFrame.PacketAnalyzersMenu.add(mainMenu);//this last command adds this entire menu to the main PacketAnalyzers menu 
	    //}} 

	    //{{REGISTER_LISTENERS 
	    circular.addActionListener(this);
	    group.addActionListener(this);
	    tree.addActionListener(this);
	    matrix.addActionListener(this);

	    packetProcessingCheckBox.addItemListener(this); 
	    backgroundProcessingCheckBox.addItemListener(this); 
	    propertiesItem.addActionListener(this); 
	    saveNodesItem.addActionListener(this); 
	    loadNodesItem.addActionListener(this); 
	    saveEdgesItem.addActionListener(this); 
	    loadEdgesItem.addActionListener(this); 
	    paintNodesItem.addItemListener(this); 
	    paintEdgesItem.addItemListener(this); 
	    paintScreenItem.addItemListener(this); 
	    //}} 
	} 

	//---------------------------------------------------------------------- 
	//EVENT HANDLERS 
	//The following two functions handle menu events 
	//The functions following this are the event handling functions 
	public void actionPerformed(ActionEvent e) 
	{ 
	    Object object = e.getSource(); 
	    if (object == circular) {
		layoutMode = CIRCULAR;
		makeConcentric();
	    } else if(object == tree) {
		layoutMode = TREE;
		makeTree();
	    } else if (object == group) {
		layoutMode = ID_GROUPING;
		groupNodesById();
	    } else if (object == matrix) {
		layoutMode = MATRIX;
		localizeNodes();
	    }
	    else if (object == saveNodesItem) 
		SaveNodes(); 
	    else if (object == loadNodesItem) 
		LoadNodes(); 
	    else if (object == saveEdgesItem) 
		SaveEdges(); 
	    else if (object == loadEdgesItem) 
		LoadEdges(); 
	    else if (object == propertiesItem) 
		ShowOptionsDialog(); 
	} 

	public void itemStateChanged(ItemEvent e) 
	{ 
	    Object object = e.getSource(); 
	    if (object == packetProcessingCheckBox) 
		togglePacketProcessing(); 
	    else if (object == backgroundProcessingCheckBox) 
		toggleBackgroundProcessing(); 
	    else if (object == paintNodesItem) 
		toggleNodePainting(); 
	    else if (object == paintEdgesItem) 
		toggleEdgePainting(); 
	    else if (object == paintScreenItem) 
		toggleScreenPainting(); 
	}		 
	//EVENT HANDLERS 
	//---------------------------------------------------------------------- 

	//------------------------------------------------------------------------ 
	//****---SAVE NODES---**** 
	//takes the node hashtable and saves it to a file 
	public void SaveNodes() 
	{ 
	    FileDialog dialog = new FileDialog(displayManager.mainFrame, "Save Nodes", FileDialog.SAVE); 
	    dialog.show(); 
	    String filename = dialog.getFile(); 
	    if( (filename != null) && (proprietaryNodeInfo!=null)) 
	    { 
		try 
		{ 	
		    FileOutputStream fos = new FileOutputStream(filename); 
		    GZIPOutputStream gos = new GZIPOutputStream(fos); 
		    ObjectOutputStream out = new ObjectOutputStream(gos); 
		    out.writeObject(proprietaryNodeInfo); 
		    out.flush(); 
		    out.close(); 
		} 
		catch(Exception e){e.printStackTrace();} 
	    } 
	} 
	//****---SAVE NODES---**** 
	//------------------------------------------------------------------------ 


	//------------------------------------------------------------------------ 
	//****---LOAD NODES---**** 
	//takes a file and loads the nodes into proprietaryNodeInfo hashtable 
	public void LoadNodes() 
	{        //in the future, it should prompt the user for which node should be kept 
	    FileDialog dialog = new FileDialog(displayManager.mainFrame, "Load Nodes", FileDialog.LOAD ); 
	    dialog.show(); 
	    String filename = dialog.getFile(); 
	    if(filename != null) 
	    { 
		NodeInfo currentNodeInfo; 
		Hashtable newNodes; 
		try 
		{ 
		    FileInputStream fis = new FileInputStream(filename); 
		    GZIPInputStream gis = new GZIPInputStream(fis); 
		    ObjectInputStream in = new ObjectInputStream(gis); 
		    newNodes = (Hashtable)in.readObject(); 
		    in.close(); 
		} 
		catch(Exception e){e.printStackTrace(); return;} 

		if((proprietaryNodeInfo == null) || (proprietaryNodeInfo.isEmpty()))//if there are no nodes yet, just assign the new nodes to the entire vector 
		{ 
		    proprietaryNodeInfo = newNodes; 
		} 
		else//otherwise take the new nodes and add them to the vector (first eliminating repeat nodes)  ...in the future, we should ask the user which node to keep, in the case of repeat nodes 
		{ 
		    for(Enumeration e = newNodes.elements();e.hasMoreElements();) 
		    { 
			currentNodeInfo = (NodeInfo)e.nextElement(); 
			proprietaryNodeInfo.remove(currentNodeInfo.getNodeId()); 
			proprietaryNodeInfo.put(currentNodeInfo.getNodeId(), currentNodeInfo); 
		    } 
		} 

	    } 
	} 
	//****---LOAD NODES---**** 
	//------------------------------------------------------------------------ 


	//------------------------------------------------------------------------ 
	//****---SAVE EDGES---**** 
	//takes the node hashtable and saves it to a file 
	public void SaveEdges() 
	{ 
	    FileDialog dialog = new FileDialog(displayManager.mainFrame, "Save Edges", FileDialog.SAVE); 
	    dialog.show(); 
	    String filename = dialog.getFile(); 
	    if( (filename != null) && (proprietaryEdgeInfo!=null)) 
	    { 
		try 
		{ 
		    FileOutputStream fos = new FileOutputStream(filename); 
		    GZIPOutputStream gos = new GZIPOutputStream(fos); 
		    ObjectOutputStream out = new ObjectOutputStream(gos); 
		    out.writeObject(proprietaryEdgeInfo); 
		    out.flush(); 
		    out.close(); 
		} 	
		catch(Exception e){e.printStackTrace();} 
	    } 
	} 
	//****---SAVE EDGES---**** 
	//------------------------------------------------------------------------ 


	//------------------------------------------------------------------------ 
	//****---LOAD EDGES---**** 
	//takes a file and loads the nodes into proprietaryNodeInfo hashtable 
	public void LoadEdges() 
	{        //in the future, it should prompt the user for which node should be kept 
	    FileDialog dialog = new FileDialog(displayManager.mainFrame, "Save Nodes", FileDialog.SAVE); 
	    dialog.show(); 
	    String filename = dialog.getFile(); 
	    if(filename != null) 
	    { 
		EdgeInfo currentEdgeInfo; 
		TwoKeyHashtable newEdges; 
		try 
		{ 
		    FileInputStream fis = new FileInputStream(filename); 
		    GZIPInputStream gis = new GZIPInputStream(fis); 
		    ObjectInputStream in = new ObjectInputStream(gis); 
		    newEdges = (TwoKeyHashtable)in.readObject(); 
		    in.close(); 
		} 
		catch(Exception e){e.printStackTrace(); return;} 

		if((proprietaryEdgeInfo == null) || (proprietaryEdgeInfo.isEmpty()))//if there are no Edges yet, just assign the new Edges to the entire vector 
		{ 
		    proprietaryEdgeInfo = newEdges; 
		} 
		else//otherwise take the new Edges and add them to the vector (first eliminating repeat Edges)  ...in the future, we should ask the user which node to keep, in the case of repeat Edges 
		{ 
		    for(Enumeration e = newEdges.elements();e.hasMoreElements();) 
		    { 
			currentEdgeInfo = (EdgeInfo)e.nextElement(); 
			proprietaryEdgeInfo.remove(currentEdgeInfo.getSource(),currentEdgeInfo.getDest()); 
			proprietaryEdgeInfo.put(currentEdgeInfo.getSource(),currentEdgeInfo.getDest(), currentEdgeInfo); 
		    } 
		} 	

	    } 
	} 
	//****---LOAD EDGES---**** 
	//------------------------------------------------------------------------ 


	//------------------------------------------------------------------------ 
	//****---TOGGLE PACKET PROCESSING 
	//This function will either register or de-register this PacketAnalyzer 
	//as a PacketEventListener.   
	public void togglePacketProcessing() 
	{ 
	    if(packetProcessingCheckBox.isSelected()) 
	    { //Note:  the following syntax "GridLocationAnalyzer.this" allows an inner class to refer to its enclosing class 
		//MainClass.addPacketEventListener(GridLocationAnalyzer.this);//start the background thread of the enclosing packetAnalyzer 
	    } 
	    else 
	    { 
		//MainClass.removePacketEventListener(GridLocationAnalyzer.this);//stop the background thread of the enclosing packetAnalyzer 
	    } 
	} 
	//****---TOGGLE PACKET PROCESSING 
	//------------------------------------------------------------------------ 

	//------------------------------------------------------------------------ 
	//****---TOGGLE BACKGROUND PROCESSING 
	public void toggleBackgroundProcessing() 
	{ 
	    /*
	       if(backgroundProcessingCheckBox.isSelected()) 
	       { 
	       start();//start the background thread of the enclosing packetAnalyzer 
	       } 	
	       else 
	       { 
	       stop();//stop the background thread of the enclosing packetAnalyzer 
	       }*/ 
	} 
	//****---TOGGLE BACKGROUND PROCESSING 
	//------------------------------------------------------------------------ 

	//------------------------------------------------------------------------ 
	//****---TOGGLE NODE PAINTING 
	//This function will either register or de-register this PacketAnalyzer 
	//as a NodePainter.   
	public void toggleNodePainting() 
	{ 
	    if(paintNodesItem.isSelected()) 
	    { //Note:  the following syntax "GridLocationAnalyzer.this" allows an inner class to refer to its enclosing class 
		//MainClass.displayManager.addNodePainter(GridLocationAnalyzer.this);//paint the nodes 
	    } 
	    else 
	    { 
		//MainClass.displayManager.removeNodePainter(GridLocationAnalyzer.this);//paint the nodes 
	    } 
	} 
	//****---TOGGLE NODE PAINTING 
	//------------------------------------------------------------------------ 


	//------------------------------------------------------------------------ 
	//****---TOGGLE EDGE PAINTING 
	//This function will either register or de-register this PacketAnalyzer 
	//as an EdgePainter.   
	public void toggleEdgePainting() 
	{ 
	    if(paintEdgesItem.isSelected()) 
	    { //Note:  the following syntax "GridLocationAnalyzer.this" allows an inner class to refer to its enclosing class 
		displayManager.addEdgePainter(GridLocationAnalyzer.this);//paint the edges 
	    } 	
	    else 
	    {	 	
		displayManager.removeEdgePainter(GridLocationAnalyzer.this);//paint the edges 
	    } 
	} 
	//****---TOGGLE EDGE PAINTING 
	//------------------------------------------------------------------------ 

	//------------------------------------------------------------------------ 
	//****---TOGGLE SCREEN PAINTING 
	//This function will either register or de-register this PacketAnalyzer 
	//as a Screen Painter.   
	public void toggleScreenPainting() 
	{ 
	    if(paintScreenItem.isSelected()) 
	    { //Note:  the followocationAnalyzer = new RoutingTableAnalyzer();ing syntax "GridLocationAnalyzer.this" allows an inner class to refer to its enclosing class 
		displayManager.addScreenPainter(GridLocationAnalyzer.this);// 
	    } 
	    else 
	    { 
		displayManager.removeScreenPainter(GridLocationAnalyzer.this);// 
	    } 
	} 
	//****---TOGGLE SCREEN PAINTING 
	//------------------------------------------------------------------------ 


    }	           
    //MENU MANAGER 
    //********************************************************* 
    //********************************************************* 
    //********************************************************* 

} 
