/*
 * EventQueue.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package soss.core;
import ostore.util.Carp;
import ostore.util.NodeId;
import ostore.util.Pair;
import ostore.util.PriorityQueue;
import ostore.dispatch.Classifier;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkException;

/**
 * The simulator's event queue.
 *
 * @author Sean C. Rhea
 * @version $Id: EventQueue.java,v 1.13 2003/04/26 00:29:45 strib Exp $
 */
public class EventQueue {

    private static final String class_tag = "EventQueue";
    private static final long TIME_SCALE = 1000 * 1000; // microseconds
    private long event_counter = 0;

    public static EventQueue instance = null;

    public EventQueue () {
	_pq = new PriorityQueue (1000 * 1000);
	instance = this;
    }

    /**
     * Schedule the <code>event</code> to be dispatched on the given
     * <code>queue</code> at the given <code>time</code>.
     *
     * @return a token that can be used to cancel the event later
     */
    public Object schedule (QueueElementIF event, double time, 
	    Classifier queue) {
	final String tag = class_tag + ".schedule";

	if (time < _current_time ) {
	    Carp.die( tag + ": time (" + time + ") must be >= current_time (" 
		      + _current_time + ")");
	}

	long ltime = Math.round(time * (double)TIME_SCALE);
	/*System.out.println ("EventQueue.schedule: scheduling " +
		event.getClass ().getName () + " event " +
		"for node " + queue.node_id () + " at time " + 
		time + ", modified = " + ltime + ".");*/

	Pair token = new Pair( new Pair (event, queue), 
			       new Long( event_counter ));
	event_counter++;
	_pq.insert (token, ltime, false);
	return token;
    }

    public void run () {
	final String tag = class_tag + ".run";

	while (! _pq.empty ()) {
	    long priority = _pq.topPriority (false);
	    Pair p = (Pair) ((Pair) _pq.extract (false)).first;
	    double _current_time_tmp = 
		((double) priority) / ((double)TIME_SCALE);

	    QueueElementIF event = (QueueElementIF) p.first;
	    Classifier classifier = (Classifier) p.second;

	    if( _current_time_tmp < _current_time ) {
		Carp.error( "GOING BACK IN TIME!!!!" );
		System.out.println ("EventQueue.run: dispatching " +
		     event + " to node [" + classifier.node_id ().port () + 
		    "] at old time " + _current_time + " and new time " +
		    _current_time_tmp + ".");
	    }
	    _current_time = _current_time_tmp;

	    /*System.out.println ("EventQueue.run: dispatching " +
		    event + " to node [" + classifier.node_id ().port () + 
		    "] at time " + _current_time + ".");*/

	    try {
		classifier.dispatch (event);
	    } catch (SinkException e) {
		Carp.die( tag + ": could not dispatch " + event);
	    }
	}
    }

    public double current_time () { return _current_time; }

    /**
     * Cancel a previously scheduled event, given that event's token.
     */
    public void cancel( Object token ) {
	_pq.remove( token );
    }

    /**
     * Remove all events pertaining to the given node
     * from this queue.
     *
     * @param dead    remove all events for destined for this node
     */
    public void removeNode ( NodeId dead ) {

	// we have to do this the really dumb way, since the
	// PriorityQueue doesn't support a better way.
	// but this should be an infrequent operation, so 
	// hopefully it's ok

	PriorityQueue tempPQ = new PriorityQueue( _pq.capacity() );
	while( !_pq.empty() ) {

	    long prio = _pq.topPriority(false);
	    Pair pair = (Pair) _pq.extract(false);
	    Pair pair2 = (Pair) pair.first;
	    Classifier c = (Classifier) pair2.second;

	    // insert it in the new one if it's node for the dead node
	    if( !c.node_id().equals( dead ) ) {
		tempPQ.insert( pair, prio, false );
	    }

	}

	_pq = tempPQ;

    }

    private PriorityQueue _pq;
    private double _current_time;
}
